/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include "util.h"
#include "draw.h"
#include "MathLib.h"
#include "resource.h"
//#include "main.h" // just for TRACE
#include "astro.h"

#define RAD M_PI/180.
#define DEG 180./M_PI

#ifndef M_TWOPI
#define M_TWOPI (2*M_PI)
#endif

#define NORM2PI(x) NORM((x), M_TWOPI)
#define NORM360(x) NORM((x), 360.)

#define CHAR_SUN  "S"
#define CHAR_MOON "M"

static void DrawSun(Int16 ix, Int16 iy, Int16 radius,
					double UT, double d, nmea_type *nmea) CODE3_SECTION;
static void DrawMoon(Int16 ix, Int16 iy, Int16 radius,
					double UT, double d, nmea_type *nmea) CODE3_SECTION;
static void getTime(nmea_type *nmea, double *UT, double *d) CODE3_SECTION;
static double NORM(double x, double y) CODE3_SECTION;

static double NORM(double x, double y)
{
	return((x) >= 0 ? fmod((x), y) : fmod((x), y) + y);
}

void AstroDraw(Int16 ix, Int16 iy, Int16 iradius, nmea_type *nmea)
{
	double UT, d;

	getTime(nmea, &UT, &d);
	DrawSun(ix, iy, iradius, UT, d, nmea);
	DrawMoon(ix, iy, iradius, UT, d, nmea);
}

static void getTime(nmea_type *nmea, double *UT, double *d)
{
	Int32 year, month, day, hour, minute, second;

#ifdef ASTROTEST
	DateTimeType dateTime;
	Int16 timeZone, daylightSavingAdjustment;
	UInt32 secs;

	timeZone = PrefGetPreference(prefTimeZone);
	daylightSavingAdjustment = PrefGetPreference(prefDaylightSavingAdjustment);
	secs = TimTimeZoneToUTC(TimGetSeconds(), timeZone, daylightSavingAdjustment);
	TimSecondsToDateTime(secs, &dateTime);

	nmea->date_of_fix_yr  = (Int32)dateTime.year;
	nmea->date_of_fix_mo  = (Int32)dateTime.month;
	nmea->date_of_fix_dy  = (Int32)dateTime.day;
	nmea->fix_taken_hour  = (Int32)dateTime.hour;
	nmea->fix_taken_min   = (Int32)dateTime.minute;
	nmea->fix_taken_sec   = (Int32)dateTime.second;
	nmea->lat = 50.881 * RAD;
	nmea->lon = -6.884 * RAD;
#endif
	year   = nmea->date_of_fix_yr + 2000;
	month  = nmea->date_of_fix_mo;
	day    = nmea->date_of_fix_dy;
	hour   = nmea->fix_taken_hour;
	minute = nmea->fix_taken_min;
	second = nmea->fix_taken_sec;

	*UT = (double)hour + (double)minute/60. + (double)second/3600.;
	*d =	367*year - 7*(year + (month+9)/12) / 4 +
		275*month/9 + day - 730530 + (*UT)/24.;
	//TRACE("UT: %ld.%ld", (Int32)(*UT), getdec(*UT));
	//TRACE("jd: %ld.%ld", (Int32)(*d), getdec(*d));
}

// see http://www.stjarnhimlen.se/comp/ppcomp.html
static void DrawSun(Int16 ix, Int16 iy, Int16 iradius,
						double UT, double d, nmea_type *nmea)
{
	double lat, lon;
	double M, Mr, e, E, w, xv, yv, v, r, lonsun, xs, ys, ecl;
	double xe, ye, ze, RA, Dec;
	double L, GMST0, LST, H;
	SkyObjectData data;

	data.x = ix;
	data.y = iy;
	data.r = iradius;
	data.inv    = true;
	data.name   = CHAR_SUN;
	data.bitmap = SunUpBitmapID;

	lat = nmea->lat;
	lon = nmea->lon * DEG;

	M = 356.0470 + 0.9856002585 * d;
	Mr = NORM2PI(RAD*M);
	w = 282.9404 + 4.70935E-5 * d;
	L = NORM360(M + w);
	e = 0.016709 - 1.151E-9 * d;
	ecl = RAD*(23.4393 - 3.563E-7 * d);
	E = M + DEG * e * sin(Mr) * (1. + e * cos(Mr));
	E = NORM2PI(RAD*E);
	xv = cos(E) - e;
	yv = sqrt(1. - e*e) * sin(E);
	r = sqrt(xv*xv + yv*yv);
	v = DEG*atan2(yv, xv);
	lonsun = RAD*NORM360(v+w);
	xs = r * cos(lonsun);
	ys = r * sin(lonsun);
	xe = xs;
	ye = ys * cos(ecl);
	ze = ys * sin(ecl);
	RA  = DEG*atan2(ye, xe);
	Dec = atan2(ze, sqrt(xe*xe+ye*ye));
	GMST0 = L/15. + 12.;
	LST = GMST0 + UT - lon/15.;
	H = NORM2PI(RAD*(15.*LST - RA));

	data.azimuth  = DEG*NORM2PI(atan2(-sin(H)*cos(Dec), cos(lat)*sin(Dec) -
						sin(lat)*cos(Dec)*cos(H)));
	data.elevation = DEG*asin(sin(lat) * sin(Dec) + cos(lat)* cos(Dec) * cos(H));

	//TRACE("sun azimuth: %ld.%ld", (Int32)data.azimuth, getdec(data.azimuth));
	//TRACE("sun altitude: %ld.%ld", (Int32)data.elevation, getdec(data.elevation));

	if(data.elevation < 0)
	{
		data.inv = false;
		data.bitmap = SunDnBitmapID;
		data.elevation = -data.elevation;
	}

	DrawSkyObject(&data);
}

static void DrawMoon(Int16 ix, Int16 iy, Int16 iradius,
						double UT, double d, nmea_type *nmea)
{
	double lat, lon;
	double M, Ms, Mr, N, w, ws, L, e, ecl, E, i;
	double xv, yv, v, vw, r, xh, yh, zh;
	double lonecl, latecl;
	double xg, yg, zg, xe, ye, ze, RA, Dec;
	double GMST0, LST, H;
	SkyObjectData data;

	data.x = ix;
	data.y = iy;
	data.r = iradius;
	data.inv    = true;
	data.name   = CHAR_MOON;
	data.bitmap = MoonUpBitmapID;

	lat = nmea->lat;
	lon = nmea->lon * DEG;

	i = RAD*5.1454;
	M = 115.3654 + 13.0649929509 * d;
	N = RAD*(125.1228 - 0.0529538083 * d);
	Mr = NORM2PI(RAD*M);
	w = RAD*(318.0634 + 0.1643573223 * d);
	e = 0.054900;
	ecl = RAD*(23.4393 - 3.563E-7 * d);
	E = M + DEG * e * sin(Mr) * (1. + e * cos(Mr));
	E = NORM2PI(RAD*E);

	xv = 60.2666 * (cos(E) - e);
	yv = 60.2666 * (sqrt(1. - e*e) * sin(E));
	r  = sqrt(xv*xv + yv*yv);
	v  = NORM2PI(atan2(RAD*yv, RAD*xv));
	vw = v+w;

	xh = r * (cos(N) * cos(vw) - sin(N) * sin(vw) * cos(i));
	yh = r * (sin(N) * cos(vw) + cos(N) * sin(vw) * cos(i));
	zh = r * (sin(vw) * sin(i));

	lonecl = NORM2PI(atan2(yh, xh));
	latecl = atan2(zh, sqrt(xh*xh+yh*yh));

	xg = r * cos(lonecl) * cos(latecl);
	yg = r * sin(lonecl) * cos(latecl);
	zg = r               * sin(latecl);

	xe = xg;
	ye = yg * cos(ecl) - zg * sin(ecl);
	ze = yg * sin(ecl) + zg * cos(ecl);
	RA  = DEG*NORM2PI(atan2(ye, xe));
	Dec = atan2(ze, sqrt(xe*xe+ye*ye));

	ws = 282.9404 + 4.70935E-5 * d;
	Ms = 356.0470 + 0.9856002585 * d;
	L = NORM360(Ms + ws);

	GMST0 = L/15. + 12.;
	LST = GMST0 + UT - lon/15.;
	H = NORM2PI(RAD*(15.*LST - RA));

	data.azimuth  = DEG*NORM2PI(atan2(-sin(H)*cos(Dec), cos(lat)*sin(Dec) -
					sin(lat)*cos(Dec)*cos(H)));
	data.elevation = DEG*asin(sin(lat) * sin(Dec) + cos(lat)* cos(Dec) * cos(H));

	//TRACE("moon azimuth: %ld.%ld", (Int32)data.azimuth, getdec(data.azimuth));
	//TRACE("moon altitude: %ld.%ld", (Int32)data.elevation, getdec(data.elevation));

	if(data.elevation < 0)
	{
		data.inv = false;
		data.bitmap = MoonDnBitmapID;
		data.elevation = -data.elevation;
	}

	DrawSkyObject(&data);
}

/*
	this is used in GPL'ed sunangle:

void CalcSunangle (sunangleInputPtr ins, 
                   sunangleOutputType* now, 
                   Boolean fullCalc)
{

double latitude = (double)ins->latitude;
double longitude = (double)ins->longitude;
double elevation = (double)ins->elevation;
Boolean metric = (double)ins->metric;
double month = (double)ins->dateTime.month;
double date = (double)ins->dateTime.day;
int year = ins->dateTime.year;
double hour = (double)ins->dateTime.hour;
double minute = (double)ins->dateTime.minute;
double tz = ConvertTzone(ins->timeZone);
int ds = ins->dst;
double meridian;
double minutesAfterMidnight;
double solarMinutesAfterMidnight;
double daylightAdj = -60*ds;
int correctedYear;
int correctedMonth;
double t;
double G;
double C;
double L;
double alpha;
double obliquity;
double eotAdj;
double clockTimeToLSOTAdj;
double sunRiseSetLSoTMinutes;
double hourAngle;
double sunrise;            // output
double sunset;             // output
double declination;        // output
double altitude;           // output
double azimuthAngle;       // output


if (!metric) elevation *= METERFEET;

meridian = tz * 15;

minutesAfterMidnight = 60*hour + minute;
solarMinutesAfterMidnight = minutesAfterMidnight 
                      + (4 * (longitude - meridian))
                      + daylightAdj ;
if ( month > 2 ) 
{
correctedYear = year;
correctedMonth = month - 3;
} else 
{
correctedYear = year - 1;
correctedMonth = month + 9;
}

t = ((solarMinutesAfterMidnight / 60.0 / 24.0) + date 
           + floor( 30.6 * correctedMonth + 0.5) 
           + floor( 365.25 * (correctedYear - 1976)) - 8707.5) 
           / 36525.0;

G = 357.528 + 35999.05 * t;
G = NormalizeTo360(G);

C = (1.915 * sin(G*RADDEG)) + (0.020 * sin(2.0 * G * RADDEG));

L = 280.460 + (36000.770 * t) + C;
L = NormalizeTo360(L);

alpha = L - 2.466 * sin(2.0 * L * RADDEG) 
   + 0.053 * sin(4.0 * L * RADDEG);

obliquity = 23.4393 - 0.013 * t;

declination = DEGRAD * 
atan (tan (obliquity * RADDEG) * sin (alpha * RADDEG));

eotAdj = ((L - C - alpha) / 15.0) * 60.0;

clockTimeToLSOTAdj = (4 * (longitude - meridian)) + eotAdj + daylightAdj ;
solarMinutesAfterMidnight = minutesAfterMidnight + clockTimeToLSOTAdj;

hourAngle = -1 * (solarMinutesAfterMidnight - 12 * 60) / 4 ;
altitude = DEGRAD * asin (
      (cos (latitude * RADDEG) *
       cos (declination * RADDEG) *
       cos (hourAngle * RADDEG)) +
      (sin (latitude * RADDEG) *
       sin (declination * RADDEG)));

(*now).declination = declination;
(*now).altitude = altitude;
if (!fullCalc) return;

if (hourAngle == 0) { azimuthAngle = 0.0; }
else
{
azimuthAngle = DEGRAD * acos ((
            (sin (altitude * RADDEG) * 
             sin (latitude * RADDEG)) -
             sin (declination * RADDEG)) /
            (cos (altitude * RADDEG) *
             cos (latitude * RADDEG)));
}    
if ( azimuthAngle * hourAngle < 0 ) { azimuthAngle *= -1 ;}

sunRiseSetLSoTMinutes = DEGRAD * acos ( -1.0 *
                   (sin(latitude * RADDEG) *
                    sin(declination * RADDEG) -
                    sin(RADDEG*(-0.8333 - 0.0347 * sqrt(elevation)))) /
                    cos(latitude * RADDEG) /
                    cos(declination * RADDEG)) * 4;

sunrise = 12*60 - sunRiseSetLSoTMinutes - (4 * (longitude - meridian))
     - eotAdj - daylightAdj;
sunset = 12*60 + sunRiseSetLSoTMinutes - (4 * (longitude - meridian))
    - eotAdj - daylightAdj;


(*now).azimuth = azimuthAngle;
(*now).sunRise.hours = (int)sunrise/60;
(*now).sunRise.minutes = sunrise - ((int)sunrise/60)*60;  // use mod
(*now).sunSet.hours = (int)sunset/60;
(*now).sunSet.minutes = sunset - ((int)sunset/60)*60;  // use mod
(*now).bearing = 180-azimuthAngle;

return;
}
*/

