/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include "beam.h"
#include "util.h"
#include "prefs.h"
#include "marker.h"
#include "track.h"

static void BeamReceiveMarker(ExgSocketPtr sock, Boolean app_is_active);
static void BeamReceiveTrack(ExgSocketPtr sock, Boolean app_is_active);
static Err BeamDBRecord(ExgSocketType *sock, DmOpenRef db, UInt16 index);
static Err WriteDBData(const void* data, UInt32* size, void* userData);
static Err ReadDBData(void* data, UInt32* size, void* udata);
static Boolean DeleteDB(	const char* name,
					UInt16 version, UInt16 cardNo, LocalID dbID,
					void* udata);
static void BeamDoSend(ExgSocketPtr sock, void *buf, UInt32 buf_len, Err *err);

void BeamReceive(ExgSocketPtr sock, Boolean app_is_active)
{
	Err err;

	err = ExgAccept(sock);
	if(err == errNone)
	{
		UInt16 len = StrLen(sock->description);

		if(StrNCompare(sock->description, BEAM_MARKER, MIN(len, StrLen(BEAM_MARKER))) == 0)
			BeamReceiveMarker(sock, app_is_active);
		else if(StrNCompare(sock->description, BEAM_TRACK, MIN(len, StrLen(BEAM_TRACK))) == 0)
			BeamReceiveTrack(sock, app_is_active);
		else UtilError(BeamERRORTYPE);
	}
	ExgDisconnect(sock, err);
}

void BeamDoSend(ExgSocketPtr sock, void *buf, UInt32 buf_len, Err *err)
{
	UInt32 offset = 0, bytes_sent;

	*err = errNone;
	while((offset < buf_len) && (*err == errNone))
	{
		bytes_sent = ExgSend(sock, buf+offset, buf_len - offset, err);
		offset += bytes_sent;
	}
}

Err BeamDB(Char *dbname, Char *type)
{
	Err err;
	ExgSocketType sock;
	LocalID db_id;

	if(dbname == NULL || StrLen(dbname) == 0)
		return(errNone);

	if(UtilIsOS(3,0) == false)
		UtilError(MainWRONGOS);

	MemSet(&sock, sizeof(sock), 0);
	sock.type	= "application/vnd.palm";
	sock.description = UtilMalloc(StrLen(type) + StrLen(dbname) + 3);
	StrPrintF(sock.description, "%s: %s", type, dbname);
	sock.name	= UtilMalloc(StrLen(dbname) + StrLen(exgSendBeamPrefix) + 5);
	StrPrintF(sock.name, "%s%s.pdb", exgSendBeamPrefix, dbname);

	db_id  = DmFindDatabase(0, dbname);
	if(db_id != 0)
	{
		err = ExgPut(&sock);
		if(err == errNone)
		{
			if(err == errNone)
				ExgDBWrite(WriteDBData, &sock, NULL, db_id, 0);

			ExgDisconnect(&sock, err);
		}
	}
	else
		err = DmGetLastErr();

	MemPtrFree(sock.name);
	MemPtrFree(sock.description);
	return(err);
}

Err BeamRecord(Char *dbname, UInt32 dbtype, Char *type, Char *name, UInt16 *index)
{
	ExgSocketType sock;
	DmOpenRef db;
	const Char *app_name;
	Err err;

	if(UtilIsOS(3,0) == false)
	{
		UtilError(MainWRONGOS);
		return(serErrBadParam);/*???*/
	}

	db = UtilDBOpen(dbname, CGCreator, dbtype, dmModeReadOnly);
	if(db == NULL)
		return(DmGetLastErr());

	MemSet(&sock, sizeof(ExgSocketType), 0);
	app_name	= UtilGetAppName();
	sock.target	= CGCreator;
	sock.description = UtilMalloc(StrLen(type) + StrLen(name) + 3);
	StrPrintF(sock.description, "%s: %s", type, name);
	sock.name	= UtilMalloc(StrLen(app_name) + StrLen(exgSendBeamPrefix) + 1);
	StrPrintF(sock.name, "%s%s", exgSendBeamPrefix, app_name);

	sock.count = 1;
	if(index == NULL)
		sock.count = UtilDBGetCountByName(dbname);
	sock.appData = sock.count;

	err	= ExgPut(&sock);
	if(err == errNone)
	{
		/* send record */
		if(index == NULL)
		{
			UInt16 i;

			for(i = 0; (i < sock.count) && (err == errNone); i++)
				err = BeamDBRecord(&sock, db, i);
		}
		else
				err = BeamDBRecord(&sock, db, *index);

		ExgDisconnect(&sock, err);
	}

	MemPtrFree(sock.description);
	MemPtrFree(sock.name);
	DmCloseDatabase(db);

	return(err);
}

static Err BeamDBRecord(ExgSocketType *sock, DmOpenRef db, UInt16 index)
{
	MemHandle handle;
	Err err;

	err    = errNone;
	handle = DmQueryRecord(db, index);
	if(handle != NULL)
	{
		UInt32 size;
		MemPtr p;

		/* send size */
		size = MemHandleSize(handle);
		BeamDoSend(sock, &size, sizeof(UInt32), &err);
		if(err == errNone)
		{
			/* send record */
			p = MemHandleLock(handle);
			BeamDoSend(sock, p, size, &err);
			MemHandleUnlock(handle);
		}
	}
	return(err);
}

static void BeamReceiveMarker(ExgSocketPtr sock, Boolean app_is_active)
{
	DmOpenRef db;

	/* get data */
	db = UtilDBOpen(CGDBMarker, CGCreator, CGDBMarkerType, dmModeWrite);
	if(db != NULL)
	{
		UInt32 i;
		UInt16 first_rec = 0;
		Boolean got_record = false;
		Err err;

		// marker count is in sock->appData
		for(i = 0; i < sock->appData; i++)
		{
			MemHandle handle;
			UInt16 rec_index;
			UInt32 size;

			/* get size */
			ExgReceive(sock, &size, sizeof(UInt32), &err);

			rec_index = dmMaxRecordIndex;
			handle = DmNewRecord(db, &rec_index, size);
			if(handle != 0)
			{
				Char buf[128];
				UInt32 brec, bgot, exp;
				Boolean failed;
				MemPtr p;

				failed	= false;
				bgot	= 0;
				p		= MemHandleLock(handle);
				exp		= size - bgot > 128L ? 128L : size-bgot;
				while(	(exp > 0) &&
						(brec = ExgReceive(sock, buf, exp, &err)) &&
						(err == errNone))
				{
					if(DmWrite(p, bgot, buf, brec) != errNone)
					{
						failed = true;
						break;
					}
					bgot += brec;
					exp = size - bgot > 128 ? 128 : size-bgot;
				}
				MemHandleUnlock(handle);

				/* remove in case of failure */
				if((err != errNone) || (failed == true))
				{
					DmReleaseRecord(db, rec_index, false);
					DmRemoveRecord(db, rec_index);
					break;
				}
				else
				{
					DmReleaseRecord(db, rec_index, true);
					MemSet(&(sock->goToParams), sizeof(sock->goToParams), 0);
					if(got_record == false)
					{
						first_rec = rec_index;
						got_record = true;
					}
				}
			}
		}
		if(got_record)
		{
			if(app_is_active)
			{
				FormPtr form;

				PrefsGetPrefs()->marker_idx = first_rec;
				form = FrmGetFormPtr(MarkerForm);
				if(form == NULL)
				{
					UtilPushFifo(PrefsGetPrefs()->start_form);
					FrmGotoForm(MarkerForm);
				}
				else
					FrmUpdateForm(MarkerForm, MARKERFORM_LISTADDED);
			}
			else
			{
				sock->noGoTo      = false;
				sock->goToCreator = CGCreator;
				sock->goToParams.recordNum = first_rec;
			}
		}
		DmCloseDatabase(db);
	}
}

static void BeamReceiveTrack(ExgSocketPtr sock, Boolean app_is_active)
{
	Err err;
	LocalID loc_id;
	Boolean need_reset;

	need_reset = false;
	err = ExgDBRead(ReadDBData, DeleteDB, sock, &loc_id, 0, &need_reset, true);

	if(app_is_active)
		TrackSendEvent();
}

static Err WriteDBData(const void* data, UInt32* size, void* userData)
{
	Err err;

	*size = ExgSend((ExgSocketPtr)userData, (void*)data, *size, &err);
	return(err);
}

static Err ReadDBData(void* data, UInt32* size, void* udata)
{
	Err err;
	UInt32 count;

	err		= errNone;
	count	= *size;
	count	= ExgReceive((ExgSocketType*)udata, data, count, &err);
	*size	= count;

	return(err);
}

static Boolean DeleteDB(	const char* name,
					UInt16 version, UInt16 cardNo, LocalID dbID,
					void* udata)
{
	Err err;

	UtilDBRemove(name, &err);

	return(err == errNone);
}

