/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include "dist.h"
#include "util.h"
#include "MathLib.h"

void DistInit(dist_type *dist, double lat, double lon, UInt32 cnt)
{
	dist->lat       = lat;
	dist->lon       = lon;
	dist->is_first  = (cnt == 0);
}

void DistFinish(dist_type *dist)
{
	DistInit(dist, 0., 0., 0);
}

// returns NM/1000
double DistGetDist(dist_type *dist, double lat, double lon)
{
	double d;

	d = 2.*asin(sqrt(pow((sin((dist->lat-lat)/2.)), 2.) + 
					cos(dist->lat)*cos(lat)*pow((sin((dist->lon-lon)/2.)), 2.)));

	return(d*1000.*180.*60./M_PI);
}

UInt32 DistGetDistInt(dist_type *dist, double lat, double lon)
{
	return((UInt32)DistGetDist(dist, lat, lon));
}

double DistGetDistUpd(dist_type *dist, double lat, double lon)
{
	double d;

	if(dist->is_first == true)
	{
		dist->is_first = false;
		d = 0.;
	}
	else
	{
		d = DistGetDist(dist, lat, lon);
	}

	dist->lat = lat;
	dist->lon = lon;
	return(d);
}

UInt32 DistGetDistUpdInt(dist_type *dist, double lat, double lon)
{
	return((UInt32)DistGetDistUpd(dist, lat, lon));
}

double DistGetDistAppr(dist_type *dist, double lat, double lon)
{
	double d, dnorth, deast, dlat, dlon, R1, R2;

	dlat = lat - dist->lat;
	dlon = lon - dist->lon;
	R1 = R2 = 60.*180./M_PI;
	// double a  = 6378.137;
	// double f = 1./298.257223563;
	// double e2 = f*(2-f);
	//R1=a(1-e2)/(1-e2*(sin(lat0))^2)^(3/2)
	//R2=a/sqrt(1-e2*(sin(lat0))^2)
  
	dnorth = R1*dlat;
	deast  = R2*cos(lat)*dlon;
	d = sqrt(dnorth*dnorth + deast*deast);
	return(d);
}

double DistGetDistValue(dist_type *dist, double lat, double lon)
{
	double deast, dlat, dlon;

	dlat = lat - dist->lat;
	dlon = lon - dist->lon;

	deast = cos(lat)*dlon;
	return(dlat*dlat + deast*deast);
}

/*
  True course is defined as usual, as the angle between the
  course line and the local meridian measured clockwise. 
*/
#define EPS 1E-6
double DistGetCourse(dist_type *dist, double lat, double lon)
{
	double tc1;

	if(cos(dist->lat) < EPS)	// EPS a small number ~ machine precision
	{
		if(dist->lat > 0)
			tc1 = M_PI;			// starting from N pole
		else
			tc1 = 2*M_PI;		// starting from S pole
	}
	else
	{
   		tc1 = fmod(atan2(sin(dist->lon-lon)*cos(lat),
           cos(dist->lat)*sin(lat)-sin(dist->lat)*cos(lat)*cos(dist->lon-lon)), 2*M_PI);
	}
	return(M_PI + tc1); // in radian
}

Int32 DistGetCourseDeg(dist_type *dist, double lat, double lon)
{
	double tc1;
	Int32 tc;

	tc1 = DistGetCourse(dist, lat, lon);
	tc = (Int32)(tc1 * 180./M_PI);

	return(tc); // in degree
	//return(tc + 540) % 360;
}

/* altitude in feet, returns nm */
UInt32 DistGetHorizon(UInt32 altitude)
{
	return(1.17*sqrt(altitude));
}

