/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include "meter.h"
#include "util.h"
#include "prefs.h"
#include "draw.h"

#define COMP_ANG 30
#define TACH_ANG 90

static const Char CHAR_NSEW[] = "NESW";
static Int16 ANG_NSEW[] = { 270, 0, 90, 180 };
#define NORTH 0
#define EAST  1
#define SOUTH 2
#define WEST  3

static void DrawNeedle(meter_type *meter, Int32 offset,
						UInt16 needle_ang, UInt16 part,
						draw_area_type *area) CODE2_SECTION;
static void DrawCircle(Int16 xCenter, Int16 yCenter, Int16 radius) CODE2_SECTION;
static void DrawPixel(Coord x, Coord y) CODE2_SECTION;
static void circlePoints(Int16 cx, Int16 cy, Int16 x, Int16 y) CODE2_SECTION;

void DrawCompassBack(meter_type *meter, draw_area_type *area)
{
	WinEraseRectangle(&area->rect, 0);
	DrawPixel(area->center_x, area->center_y);
	DrawCircle(area->center_x, area->center_y, meter->radius-1);
}

void DrawCompassNeedle(meter_type *meter, draw_area_type *area)
{
	DrawNeedle(meter, 0, COMP_ANG, 0, area);
}

void DrawCompassScale(meter_type *meter, draw_area_type *area)
{
	FontID old_font;
	Int16 fh, fw, i;

	if(UtilIsVGA())
		old_font = FntSetFont(VgaBaseToVgaFont(meter->font));
	else
		old_font = FntSetFont(meter->font);

	fh = FntCharHeight() >> 1;
	for(i = 0; i < 4; i++)
	{
		Int32 r;
		Int16 angle;
		Coord x, y;

		angle = ANG_NSEW[i];
		if(meter->is_fixed == false)
			angle = 360 + angle - meter->aval;
		fw    = FntCharWidth(CHAR_NSEW[i]) >> 1;
		r     = meter->radius - MAX(fh, fw);

		x = (Coord)(r * UtilCos(angle) / 0x7FFF) + area->center_x;
		y = (Coord)(r * UtilSin(angle) / 0x7FFF) + area->center_y;
		WinDrawChars(&CHAR_NSEW[i], 1, x-(fw>>1), y-fh);
	}

	FntSetFont(old_font);
}

void DrawOdoBack(meter_type *meter, draw_area_type *area)
{
	WinEraseRectangle(&area->rect, 0);
	DrawCircle(		area->center_x, area->center_y,
					meter->radius-1);
	WinDrawLine(	area->center_x-1, area->center_y,
					area->center_x+1, area->center_y);
}

void DrawOdoNeedle(meter_type *meter, draw_area_type *area)
{
	double fac   = 180./PrefsGetPrefs()->maxspeed;
	meter->aval  = meter->aval*fac;
	meter->aval2 = meter->aval2*fac;

	DrawNeedle(meter, 270, TACH_ANG, 2, area);
}

void DrawOdoScale(meter_type *meter, draw_area_type *area)
{
	Int16 i, step;
	double maxspeed;

	maxspeed = (double)PrefsGetPrefs()->maxspeed;
	step     = (Int16)(180./(maxspeed/20.));
	for(i = 0; i <= 180; i+=step)
	{
		RectangleType rect;
		Int16 angle;
		Coord x, y;

		angle = i + 180;
		x = (Coord)(meter->radius * UtilCos(angle) / 0x7FFF) + area->center_x;
		y = (Coord)(meter->radius * UtilSin(angle) / 0x7FFF) + area->center_y;

		RctSetRectangle(&rect, x-1, y-1, 3, 3);
		WinDrawRectangle(&rect, 0);
	}
}

void DrawVarioBack(meter_type *meter, draw_area_type *area)
{
	WinEraseRectangle(&area->rect, 0);
	DrawCircle(		area->center_x, area->center_y,
					meter->radius-1);
	WinDrawLine(	area->center_x, area->center_y-1,
					area->center_x, area->center_y+1);
}

void DrawVarioNeedle(meter_type *meter, draw_area_type *area)
{
	DrawNeedle(meter, 90, TACH_ANG, 2, area);
}

void DrawVarioScale(meter_type *meter, draw_area_type *area)
{
	Int16 i;

	for(i = 0; i < 180; i+=VARIO_MUL)
	{
		Int16 x, y, angle;
		RectangleType rect;

		angle = i + 270;
		x =	(Int16)(meter->radius * UtilCos(angle) / 0x7FFF) + area->center_x;
		y =	(Int16)(meter->radius * UtilSin(angle) / 0x7FFF) + area->center_y;

		RctSetRectangle(&rect, x-1, y-1, 3, 3);
		WinDrawRectangle(&rect, 0);
	}
}

static void DrawNeedle(meter_type *meter, Int32 offset,
						UInt16 needle_ang, UInt16 part,
						draw_area_type *area)
{
	Coord cx[5], cy[5];
	Coord oa, backlength, angle;
	RectangleType re2;

	if(meter->has_aval)
	{
		angle = (meter->is_fixed ? meter->aval : 0) + offset + 270;
		oa = angle + 180;
		backlength = meter->iradius >> part;

		cx[0] = (Coord)(meter->iradius * UtilCos(angle) / 0x7FFF);
		cy[0] = (Coord)(meter->iradius * UtilSin(angle) / 0x7FFF);

		angle = oa;
		cx[1] = (Coord)((meter->iradius >> 1) * UtilCos(angle) / 0x7FFF);
		cy[1] = (Coord)((meter->iradius >> 1) * UtilSin(angle) / 0x7FFF);

		angle = oa - needle_ang;
		cx[2] = (Coord)(backlength * UtilCos(angle) / 0x7FFF);
		cy[2] = (Coord)(backlength * UtilSin(angle) / 0x7FFF);

		angle = oa + needle_ang;
		cx[3] = (Coord)(backlength * UtilCos(angle) / 0x7FFF);
		cy[3] = (Coord)(backlength * UtilSin(angle) / 0x7FFF);
	}

	if(meter->has_aval2)
	{
		Int32 angle2 = meter->aval2 + offset + 270;
		if(meter->is_fixed == false)
			angle2 = 360 + angle2 - meter->aval;
		cx[4] = (Coord)(meter->iradius * UtilCos(angle2) / 0x7FFF);
		cy[4] = (Coord)(meter->iradius * UtilSin(angle2) / 0x7FFF);
	}

	/* erase old lines only if not double buffering */
	if(meter->do_dbuf == false)
	{
		if(meter->eraselast == false)
			meter->eraselast = true;
		else
		{
			if(meter->has_aval)
			{
				// probably 32 would do, workaround for bug
				if(UtilIsOS(3, 3) == false && needle_ang == TACH_ANG)
				{
					WinEraseLine(	area->center_x,
									area->center_y,
									meter->cx[0],
									meter->cy[0]);
				}
				else
				{
					WinEraseLine(	meter->cx[0], meter->cy[0],
									meter->cx[2], meter->cy[2]);
					WinEraseLine(	meter->cx[0], meter->cy[0],
									meter->cx[3], meter->cy[3]);
					WinEraseLine(	meter->cx[2], meter->cy[2],
									meter->cx[1], meter->cy[1]);
					WinEraseLine(	meter->cx[3], meter->cy[3],
									meter->cx[1], meter->cy[1]);
				}
			}
			if(meter->has_aval2)
			{
				RctSetRectangle(&re2,
								meter->cx[4]-(meter->square >> 1),
								meter->cy[4]-(meter->square >> 1),
								meter->square, meter->square);
				WinEraseRectangle(&re2, 1);
			}
		}
	}

	if(meter->has_aval)
	{
		meter->cx[0] = area->center_x + cx[0];
		meter->cy[0] = area->center_y + cy[0];
		meter->cx[1] = area->center_x + cx[1];
		meter->cy[1] = area->center_y + cy[1];
		meter->cx[2] = area->center_x + cx[2];
		meter->cy[2] = area->center_y + cy[2];
		meter->cx[3] = area->center_x + cx[3];
		meter->cy[3] = area->center_y + cy[3];
	}
	if(meter->has_aval2)
	{
		meter->cx[4] = area->center_x + cx[4];
		meter->cy[4] = area->center_y + cy[4];
	}

	if(meter->has_aval)
	{
		if(UtilIsOS(3, 3) == false && needle_ang == TACH_ANG)
		{
			WinDrawLine(	area->center_x,
							area->center_y,
							meter->cx[0],
							meter->cy[0]);
		}
		else
		{
			WinDrawLine(meter->cx[0], meter->cy[0],
						meter->cx[2], meter->cy[2]);
			WinDrawLine(meter->cx[0], meter->cy[0],
						meter->cx[3], meter->cy[3]);
			WinDrawLine(meter->cx[2], meter->cy[2],
						meter->cx[1], meter->cy[1]);
			WinDrawLine(meter->cx[3], meter->cy[3],
						meter->cx[1], meter->cy[1]);
		}
	}
	if(meter->has_aval2)
	{
		RctSetRectangle(&re2,
						meter->cx[4]-(meter->square >> 1),
						meter->cy[4]-(meter->square >> 1),
						meter->square, meter->square);
		WinDrawRectangle(&re2, 1);
	}
}

void DrawValue(meter_type *meter, draw_area_type *area)
{
	if(meter->has_dval && meter->wants_dval)
	{
		Int16 l, sl;
		UInt32 km;
		Coord x, y;
		FontID old_font;
		char buf[32];

		if(UtilIsVGA())
			old_font = FntSetFont(VgaBaseToVgaFont(meter->ufont));
		else
			old_font = FntSetFont(meter->ufont);

		km = (UInt32)(meter->dval/1000.);
		StrPrintF(buf, "%05ld.%03ld", km, meter->dval-(km*1000));
		sl = StrLen(buf);
		l = FntLineWidth(buf, sl);
		x =	(area->rect.extent.x >> 1) +
			 area->rect.topLeft.x - (l >> 1);
		y =  area->rect.extent.y +
			 area->rect.topLeft.y -
			 FntCharHeight();
		WinDrawChars(buf, sl, x, y);

		FntSetFont(old_font);
	}
}

void DrawSky(Int16 x, Int16 y, Int16 r)
{
	Int16 i, r3, fh, fw;
	RectangleType rect;

	RctSetRectangle(&rect, x-r, y-r, r << 1, r << 1);
	WinEraseRectangle(&rect, 0);

	r3 = r / 3;
	DrawCircle(x, y, 3*r3);
	DrawCircle(x, y, 2*r3);
	DrawCircle(x, y, r3);

	for(i = 0; i < 180; i+=30)
	{
		Int16 x1, y1, x2, y2;

		x1 = (Int16)((Int32)r * UtilCos(i) / 0x7FFF);
		y1 = (Int16)((Int32)r * UtilSin(i) / 0x7FFF);
		x2 = -x1;
		y2 = -y1;
		WinDrawLine(x + x1, y + y1, x + x2, y + y2);
	}

	fh = FntCharHeight();
	fw = FntCharWidth(CHAR_NSEW[WEST]);
	WinDrawChars(&CHAR_NSEW[WEST], 1, x-r+3, y-(fh>>1));
	fw = FntCharWidth(CHAR_NSEW[EAST]);
	WinDrawChars(&CHAR_NSEW[EAST], 1, x+r-fw-2, y-(fh>>1));
	fw = FntCharWidth(CHAR_NSEW[NORTH]);
	WinDrawChars(&CHAR_NSEW[NORTH], 1, x-(fw>>1), y-r+2);
	fw = FntCharWidth(CHAR_NSEW[SOUTH]);
	WinDrawChars(&CHAR_NSEW[SOUTH], 1, x-(fw>>1), y+r-fh-1);
}

void DrawSkyObject(SkyObjectData *data)
{
	Int32 sx, sy, sr;
	IndexedColorType ofcolor = 0;
	IndexedColorType otcolor = 0;

	sr = (Int32)(-data->r * ((data->elevation-90.) / 90.));
	sx = (Int32)(sr * UtilCos(270+data->azimuth) / 0x7FFF + data->x);
	sy = (Int32)(sr * UtilSin(270+data->azimuth) / 0x7FFF + data->y);

	if(data->wants_color && UtilHasColor())
	{
		RGBColorType ctype;
		IndexedColorType color;

		ctype.r = data->red;
		ctype.g = data->green;
		ctype.b = data->blue;
		color = WinRGBToIndex(&ctype);
		ofcolor = WinSetForeColor(color);
		otcolor = WinSetTextColor(color);
	}

	if(data->bitmap > 0)
	{
		MemHandle imageH;

		imageH = DmGetResource(bitmapRsc, data->bitmap);
		if(imageH != NULL)
		{
			BitmapPtr bm;
			Coord w, h;

			bm = MemHandleLock(imageH);
			UtilBmpGetDimensions(bm, &w, &h);
			WinDrawBitmap(bm, sx-(w>>1), sy-(h>>1));
			MemHandleUnlock(imageH);
			DmReleaseResource(imageH);
		}
	}
	else
	{
		if(data->name != NULL)
		{
			Int16 len, fh, fw;

			len = StrLen(data->name);
			fh = FntCharHeight() >> 1;
			fw = FntCharsWidth(data->name, len) >> 1;

			if(data->inv == false)
				WinDrawChars(data->name, len, sx-fw, sy-fh);
			else
				WinDrawInvertedChars(data->name, len, sx-fw, sy-fh);
		}
		else
		{
			RectangleType rect;
			Coord psize = 4;

			if(UtilIsHires() == true)
				psize <<= 1;

			RctSetRectangle(&rect, sx-(psize >> 1), sy-(psize >> 1), psize, psize);
			WinDrawRectangle(&rect, 1);
		}
	}

	if(data->wants_color && UtilHasColor())
	{
		WinSetForeColor(ofcolor);
		WinSetTextColor(otcolor);
	}
}

void DrawScale(RectangleType *rect, double scale, Boolean is_upright)
{
	Coord x, x2, y, y2, w, h;
	UInt32 v, d;
	UInt16 len;
	Int16 fw, fh;
	Char buf[32];

	scale = PrefsTrackDistToPref(scale);

	x  = rect->topLeft.x;
	y  = rect->topLeft.y;
	w  = rect->extent.x;
	h  = rect->extent.y;
	y2 = y+h-4;
	x2 = (x+w)>>1;

	if(!is_upright)
	{
		WinDrawLine(x, y, x, y+h);
		WinDrawLine(x+w, y, x+w, y+h);
		WinDrawLine(x, y2, x+w, y2);

		scale *= (double)w;
	}
	else
	{
		WinDrawLine(x, y, x+w, y);
		WinDrawLine(x, y+h, x+w, y+h);
		WinDrawLine(x2, y, x2, y+h);
		scale *= (double)h;
	}

	v = (UInt32)scale;
	scale -= (double)v;
	scale *= 1000.;
	d = (UInt32)scale;
	StrPrintF(buf, "%ld.%ld %s", v, d, PrefsTrackDistUnit());


	fh = FntCharHeight();
	len = StrLen(buf);
	if(!is_upright)
	{
		fw = FntCharsWidth(buf, len);
		x = (w - fw) >> 1;
		y = y2 - fh - 2;
	}
	else
	{
		y = y + h + 1;
	}
	WinDrawChars(buf, len, x, y);
}

static void DrawPixel(Coord x, Coord y)
{
	if(UtilIsOS(3, 5))
		WinDrawPixel(x, y);
	else
		WinDrawLine(x, y, x, y);
}

/*
void DrawCircle(Int16 xCenter, Int16 yCenter, Int16 radius)
{
	RectangleType circle;

	circle.topLeft.x = xCenter - radius;
	circle.topLeft.y = yCenter - radius;
	circle.extent.x  = radius * 2;
	circle.extent.y  = radius * 2;
	WinDrawRectangle(&circle, radius);
	circle.topLeft.x += 1;
	circle.topLeft.y += 1;
	circle.extent.x  -= 2;
	circle.extent.y  -= 2;
	WinEraseRectangle(&circle, radius-1);
}
*/

void DrawCircle(Int16 xCenter, Int16 yCenter, Int16 radius)
{
	Int16 x = 0;
	Int16 y = radius;
	Int16 p = (5 - radius*4) >> 2;

	circlePoints(xCenter, yCenter, x, y);
	while(x < y)
	{
		x++;
		if(p < 0)
		{
			p += (x << 1)+1;
		}
		else
		{
			y--;
			p += ((x-y) << 1)+1;
		}
		circlePoints(xCenter, yCenter, x, y);
	}
}

static void circlePoints(Int16 cx, Int16 cy, Int16 x, Int16 y)
{
	if(x == 0)
	{
		DrawPixel(cx, cy + y);
		DrawPixel(cx, cy - y);
		DrawPixel(cx + y, cy);
		DrawPixel(cx - y, cy);
	}
	else 
		if(x == y)
		{
			DrawPixel(cx + x, cy + y);
			DrawPixel(cx - x, cy + y);
			DrawPixel(cx + x, cy - y);
			DrawPixel(cx - x, cy - y);
		}
		else 
			if(x < y)
			{
				DrawPixel(cx + x, cy + y);
				DrawPixel(cx - x, cy + y);
				DrawPixel(cx + x, cy - y);
				DrawPixel(cx - x, cy - y);
				DrawPixel(cx + y, cy + x);
				DrawPixel(cx - y, cy + x);
				DrawPixel(cx + y, cy - x);
				DrawPixel(cx - y, cy - x);
			}
}

/*
void DrawEllipse(Int16 xCenter, Int16 yCenter, Int16 rx, Int16 ry)
{
	Int32 t1 = rx*rx, t2 = t1<<1, t3 = t2<<1;
	Int32 t4 = ry*ry, t5 = t4<<1, t6 = t5<<1;
	Int32 t7 = rx*t5, t8 = t7<<1, t9 = 0L;
	Int32 d1 = t2 - t7 + (t4>>1);
	Int32 d2 = (t1>>1) - t8 + t5;
	Int32 x = rx, y = 0;

	while(d2 < 0)
	{
		DrawPixel(xCenter + x, yCenter + y);
		DrawPixel(xCenter + x, yCenter - y);
		DrawPixel(xCenter - x, yCenter + y);
		DrawPixel(xCenter - x, yCenter - y);

		y++;
		t9 += t3;	
		if(d1 < 0)
		{
			d1 += t9 + t2;
			d2 += t9;
		}
		else
		{
			x--;
			t8 -= t6;
			d1 += t9 + t2 - t8;
			d2 += t9 + t5 - t8;
		}
	}

	do
	{
		DrawPixel(xCenter + x, yCenter + y);
		DrawPixel(xCenter + x, yCenter - y);
		DrawPixel(xCenter - x, yCenter + y);
		DrawPixel(xCenter - x, yCenter - y);

		x--;
		t8 -= t6;	
		if(d2 < 0)
		{
			y++;
			t9 += t3;
			d2 += t9 + t5 - t8;
		}
		else
			d2 += t5 - t8;
	} while(x >= 0);
}
*/

