/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include "MathLib.h"
#include "main.h"
#include "util.h"
#include "prefs.h"
#include "forms.h"
#include "meter.h"
#include "resource.h"
#include "def.h"

static void GPSHandler(FormPtr form, nmea_type *nmea) CODE2_SECTION;
static void SizeForm(FormPtr form) CODE2_SECTION;
static void ResizeMeters(FormPtr form) CODE2_SECTION;
static void UpdatePBatt(FormPtr form) CODE2_SECTION;

static Boolean form_ready = false;
static Int32 old_alt;
static UInt32 last_check;
static Boolean compass_dwn;
static meter_type compass;
static meter_type vario;


Boolean GPS2FormHandler(EventPtr event)
{
	FormPtr	form = FrmGetActiveForm();
	Boolean handled = false;

	switch(event->eType)
	{
		case frmOpenEvent:
			form_ready = false;
			last_check = 0L;
			if(UtilIsVGA())
				VgaFormModify(form, vgaFormModify160To240);
			if(UtilIsResizeableNoDia())
				SizeForm(form);
			FrmDrawForm(form);
			MeterInit(&vario, TYPE_VARIO);
			MeterInit(&compass, TYPE_COMPASS);
			MeterSetDoublebuffered(&compass, true);
			MeterSetFixed(&compass, PrefsGetPrefs()->compass_fixed);
			ResizeMeters(form);
			form_ready = true;
			handled = true;
			break;
		case frmCloseEvent:
			form_ready = false;
			MeterFinish(&vario);
			MeterFinish(&compass);
			break;
		case frmUpdateEvent:
			FrmDrawForm(form);
			ResizeMeters(form);
			UtilUpdateChar(form, SpeedToTargetID, '-');
			UtilUpdateChar(form, MeritFactorID, '-');
			handled = true;
			break;
		case winDisplayChangedEvent:
		case displayExtentChangedEvent:
			if(form_ready == true)
			{
				SizeForm(form);
				FrmDrawForm(form);
				ResizeMeters(form);
			}
			handled = true;
			break;
		case penDownEvent:
			if(MeterPenInside(&compass, event))
				compass_dwn = true;
			else
				compass_dwn = false;
			break;
		case penUpEvent:
			if(compass_dwn && MeterPenInside(&compass, event))
			{
				CAppPrefs *prefs = PrefsGetPrefs();
				prefs->compass_fixed = !prefs->compass_fixed;
				MeterSetFixed(&compass, prefs->compass_fixed);
			}
			compass_dwn = false;
			break;
		case GpsNmeaEvent:
				GPSHandler(form, CASTNMEA(event));
			break;
		case nilEvent:
				{
					UInt32 time = TimGetSeconds();
					if(time != last_check)
					{
						UpdatePBatt(form);
						last_check = time;
					}
				}
			break;
		default:
			break;
	}
	return(handled);
}

static void GPSHandler(FormPtr form, nmea_type *nmea)
{
	MainSetGPSTitle(form);

	MeterUpdate(&compass, nmea);
	MeterUpdate(&vario, nmea);

	if(nmea->is_new_gga || nmea->is_new_nexrmc)
	{
		PrefsUpdateLabelAlt(form, AltID, nmea->alt, nmea->alt_unit, nmea->alt_WGS84, nmea->alt_WGS84_unit, nmea->lat, nmea->lon);
		PrefsUpdateLabelAlt(form, ClimbID, nmea->alt - old_alt, nmea->alt_unit, 0, '*', 0., 0.);

		old_alt = nmea->alt;
		nmea->is_new_gga = false;
	}

	if(nmea->is_new_rmc || nmea->is_new_vtg || nmea->is_new_nexrmc)
	{
		PrefsUpdateLabelSpeed(form, SpeedID, nmea->speed_f);
		UtilUpdateSingle(form, CourseID, nmea->course, "");

		if(MarkerHasDestination() && (nmea->speed != 0))
		{
			Char *unit;
			double stt, ctt; // speed-to-target, course-to-target

			ctt = MarkerGetCourse(nmea) - nmea->course;
			stt = nmea->speed_f * cos(ctt * M_PI/180.);

			UtilUpdateDouble(form, MeritFactorID, stt/nmea->speed_f, 2);

			PrefsSpeedToPref(&stt, &unit);
			UtilUpdateDouble(form, SpeedToTargetID, stt, 1);
		}
		else
		{
			UtilUpdateChar(form, SpeedToTargetID, '-');
			UtilUpdateChar(form, MeritFactorID, '-');
		}


		if(nmea->mag_var_dir == '-')
		{
			double de = UtilGetDeviation(nmea->lat, nmea->lon);
			UtilUpdateDouble(form, MagVarID, de, 1);
			if(de < 0.)
				nmea->mag_var_dir = 'E';
			else
				nmea->mag_var_dir = 'W';
		}
		else
		{
			// Care! This only works for mag_var_dec == 1!!!
			// otherwise leading zeros get ignored!
			UtilUpdateDec(form, MagVarID, nmea->mag_var, nmea->mag_var_fra);
		}
		UtilUpdateChar(form, MagVarDirID, nmea->mag_var_dir);

		if(nmea->has_rmc)
			nmea->is_new_rmc = false;
		if(nmea->has_vtg)
			nmea->is_new_vtg = false;
		if(nmea->has_nexrmc)
			nmea->is_new_nexrmc = false;
	}

	if(nmea->is_new_pmgnst)
	{
		UtilUpdateSingle(form, GBatteryID, nmea->mgn_bat_left, "'");
		nmea->is_new_pmgnst = false;
	}
}

static void SizeForm(FormPtr form)
{
	RectangleType curBounds, displayBounds;
	Coord diff;

	if(UtilIsVGA())
	{
		MeterErase(&compass);
		MeterErase(&vario);
	}
	UtilGetFrmBounds(form, &curBounds);
	UtilGetDpyBounds(&displayBounds);

	diff = displayBounds.extent.y - (curBounds.topLeft.y + curBounds.extent.y);

	if(diff)
	{
		curBounds.extent.y += diff;
		WinSetBounds(FrmGetWindowHandle(form), &curBounds);
	}
}

static void ResizeMeters(FormPtr form)
{
	RectangleType rfield_rect, lfield_rect, rect;
	Coord w, h, cw, r;

	FrmGetObjectBounds(form, FrmGetObjectIndex(form, MagVarID),
						&lfield_rect);
	FrmGetObjectBounds(form, FrmGetObjectIndex(form, GBatteryID),
						&rfield_rect);
	UtilGetFrmBounds(form, &rect);
	w = rect.extent.x;
	h = rect.extent.y;

	cw  = MIN(w, h - lfield_rect.topLeft.y - FntCharHeight() - 5);
	MeterSetSize(&vario, 1, h-1-cw, cw>>1, cw);

	cw = MIN(w>>1, h - rfield_rect.topLeft.y - FntCharHeight() - 5);
	r  = MIN(w - (cw>>1) - 2, cw);
	MeterSetSize(&compass, w-1-r, h-1-cw, r, r);

	MeterDraw(&compass);
	MeterDraw(&vario);

	if(PrefsGetPrefs()->format_distspeed != NAUTICAL)
		UtilUpdateLabel(form, STTLabelID, "STT:");
	else
		UtilUpdateLabel(form, STTLabelID, "VMG:");
}

static void UpdatePBatt(FormPtr form)
{
	UInt16 bleft;

	bleft = UtilGetBatteryState();

	if(bleft == 0xFFFF)
		UtilUpdateChar(form, PBatteryID, '~');
	else
		UtilUpdateSingle(form, PBatteryID, (Int32)bleft, "%");
}

