/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include "main.h"
#include "forms.h"
#include "meter.h"
#include "serial.h"
#include "util.h"
#include "nmea.h"
#include "draw.h"
#include "marker.h"
#include "time.h"
#include "log.h"
#include "prefs.h"
#include "help.h"
#include "resource.h"
#include "def.h"

static void GPSHandler(FormPtr form, nmea_type *nmea) CODE2_SECTION;
static void FillRectangles(FormPtr form) CODE2_SECTION;
static void SizeForm(FormPtr form) CODE2_SECTION;

static Boolean form_ready = false;
static Boolean compass_dwn = false;
static Boolean odo_dwn = false;;
static meter_type compass;
static meter_type odo;

Boolean GPSFormHandler(EventPtr event)
{
	FormPtr	form = FrmGetActiveForm();
	Boolean handled = false;

	switch(event->eType)
	{
		case frmOpenEvent:
			form_ready = false;
			if(UtilIsVGA())
				VgaFormModify(form, vgaFormModify160To240);
			MeterInit(&odo, TYPE_ODO);
			MeterInit(&compass, TYPE_COMPASS);
			MeterSetDoublebuffered(&compass, true);
			MeterSetFixed(&compass, PrefsGetPrefs()->compass_fixed);
			if(UtilIsResizeableNoDia())
				SizeForm(form);
			FrmDrawForm(form);
			PrefsUpdateLabelLatLon(form, LatLabelID, LonLabelID, false);
			FillRectangles(form);
			FrmCopyTitle(form, UtilGetAppName());
			MeterDraw(&compass);
			MeterDraw(&odo);
			form_ready = true;
			handled = true;
			break;
		case frmCloseEvent:
			form_ready = false;
			MeterFinish(&compass);
			MeterFinish(&odo);
			break;
		case frmUpdateEvent:
			FrmDrawForm(form);
			if(event->data.frmUpdate.updateCode == frmRedrawUpdateCode)
			{
				double dist = PrefsGetPrefs()->dist_f;
				PrefsNauticToPref(&dist, NULL);
				odo.dval  = (UInt32)dist;
				PrefsUpdateLabelLatLon(form, LatLabelID, LonLabelID, false);
			}
			MeterErase(&compass);
			MeterErase(&odo);
			FillRectangles(form);
			MeterDraw(&compass);
			MeterDraw(&odo);
			handled = true;
			break;
		case winDisplayChangedEvent:
		case displayExtentChangedEvent:
			if(form_ready == true)
			{
				//TODO erst updateEvent, dann displayChanged
				SizeForm(form);
				FrmDrawForm(form);
				FillRectangles(form);
				MeterDraw(&compass);
				MeterDraw(&odo);
			}
			handled = true;
			break;
		case penDownEvent:
			compass_dwn = false;
			odo_dwn = false;
			if(MeterPenInside(&compass, event))
				compass_dwn = true;
			if(MeterPenInside(&odo, event))
				odo_dwn = true;
			break;
		case penUpEvent:
			if(compass_dwn && MeterPenInside(&compass, event))
			{
				CAppPrefs *prefs = PrefsGetPrefs();
				prefs->compass_fixed = !prefs->compass_fixed;
				MeterSetFixed(&compass, prefs->compass_fixed);
			}
			if(odo_dwn && MeterPenInside(&odo, event))
			{
				FrmPopupForm(StartOdoForm);
			}
			compass_dwn = false;
			odo_dwn = false;
			break;
		case GpsNmeaEvent:
				GPSHandler(form, CASTNMEA(event));
				handled = true;
			break;
		default:
			break;
	}
	return(handled);
}

static void GPSHandler(FormPtr form, nmea_type *nmea)
{
	MainSetGPSTitle(form);
	MeterUpdate(&compass, nmea);
	MeterUpdate(&odo, nmea);

	if(nmea->is_new_gga || nmea->is_new_nexrmc)
	{
		UtilUpdateString(form, TimeID, TimeGetTimeString());
		PrefsUpdateLabelCoord(form,
								LatitudeID, LongitudeID,
								nmea->lat, nmea->lon);
		PrefsUpdateLabelAlt(form, AltID, nmea->alt, nmea->alt_unit,
								nmea->alt_WGS84, nmea->alt_WGS84_unit,
								nmea->lat, nmea->lon);

		if(nmea->has_gsa == false)
		{
			if(nmea->is_new_gga)
				UtilUpdateDec(form, HDOPID, nmea->gga_hdop, nmea->gga_hdop_fra);
			else
				UtilUpdateDec(form, HDOPID, nmea->hdop, nmea->hdop_fra);
		}
	}

	if(nmea->is_new_gsa)
	{
		UtilUpdateDec(form, PDOPID, nmea->pdop, nmea->pdop_fra);
		UtilUpdateDec(form, HDOPID, nmea->hdop, nmea->hdop_fra);
		UtilUpdateDec(form, VDOPID, nmea->vdop, nmea->vdop_fra);
	}

	if(nmea->is_new_rmc || nmea->is_new_vtg || nmea->is_new_nexrmc)
	{
		PrefsUpdateLabelSpeed(form, SpeedID, nmea->speed_f);
		UtilUpdateSingle(form, CourseID, nmea->course, "");
		UtilUpdateString(form, DateID, TimeGetDateString());
	}

	nmea->is_new_rmc = false;
	nmea->is_new_vtg = false;
	nmea->is_new_nexrmc = false;
	nmea->is_new_gga = false;
	nmea->is_new_gsa = false;
}

static void FillRectangles(FormPtr form)
{
	RectangleType rect, toprect;
	Coord w, h, cw, ch;

	FrmGetObjectBounds(form, FrmGetObjectIndex(form, LonLabelID),
						&toprect);
	UtilGetFrmBounds(form, &rect);
	w = rect.extent.x;
	h = rect.extent.y;

	ch = h - (toprect.topLeft.y + toprect.extent.y);
	cw = MIN(w/3, ch);
	MeterSetSize(&odo, 0, h-1-cw, w - cw, cw);
	MeterSetSize(&compass, w-1-cw, h-1-cw, cw, cw);
}

static void SizeForm(FormPtr form)
{
	RectangleType curBounds, displayBounds;
	Coord diff;

	UtilGetFrmBounds(form, &curBounds);
	UtilGetDpyBounds(&displayBounds);

	diff = displayBounds.extent.y - (curBounds.topLeft.y + curBounds.extent.y);

	if(diff)
	{
		if(UtilIsVGA())
		{
			MeterErase(&compass);
			MeterErase(&odo);
		}

		curBounds.extent.y += diff;
		WinSetBounds(FrmGetWindowHandle(form), &curBounds);
	}
}

