/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include "forms.h"
#include "help.h"
#include "main.h"
#include "prefs.h"
#include "util.h"
#include "resource.h"
#include "def.h"

static void SizeForm(FormPtr form) CODE4_SECTION;
static void ShowHelpMessageInit(FormPtr form, Boolean clean) CODE4_SECTION;
static void ShowHelpMessage(FormPtr form, Boolean clean) CODE4_SECTION;
static void ChangeFont(FormPtr form, UInt8 ft) CODE4_SECTION;
static void ButtonHandler(FormPtr form, WChar direction) CODE4_SECTION;
static void setTop(Boolean b) CODE4_SECTION;
static void setBottom(Boolean b) CODE4_SECTION;
static Boolean isTop() CODE4_SECTION;
static Boolean isBottom() CODE4_SECTION;

static const Char *msg;
static Boolean form_ready = false;
static UInt16 title_height;
static Int16 sl, curr_line;
static Boolean is_bottom, is_top;
static Coord x, y, w, h;

static FontID help_fonts[] = { stdFont, boldFont, largeBoldFont };

Boolean HelpFormHandler(EventPtr event)
{
	Boolean handled = false;
	UInt16 ctl_index;
	FormPtr	form = FrmGetActiveForm();

	switch(event->eType)
	{
		case frmOpenEvent:
			form_ready    = false;
			title_height  = 16;
			curr_line     = 0;
			x = y = w = h = 0;
			sl            = 0;
			msg           = NULL;
			if(UtilIsVGA())
			{
				VgaFormModify(form, vgaFormModify160To240);
				title_height = VgaGetFrmTitleHeight()+1;
			}
			if(UtilIsResizeableNoDia())
				SizeForm(form);
			FrmDrawForm(form);
			setBottom(false);
			setTop(false);
			ctl_index = FrmGetObjectIndex(form, HelpFont1ID+PrefsGetPrefs()->help_font);
			FrmSetControlValue(form, ctl_index, 1);
			msg = HelpGetString();
			ShowHelpMessageInit(form, false);
			form_ready = true;
			handled = true;
			break;
		case frmCloseEvent:
			form_ready = false;
			HelpReleaseString();
			msg = NULL;
			break;
		case winDisplayChangedEvent:
		case displayExtentChangedEvent:
			if(form_ready == true)
			{
				SizeForm(form);
				setTop(false);
				setBottom(false);
				WinEraseWindow();
				FrmDrawForm(form);
				ShowHelpMessageInit(form, false);
			}
			handled = true;
			break;
		case ctlSelectEvent:
			switch(event->data.ctlSelect.controlID)
			{
				case HelpDoneID:
					//HelpReleaseString();
					//msg = NULL;
					//FrmUpdateForm(UtilLastFormId(), frmRedrawUpdateCode);
					UtilReturnToLastForm();
					handled = true;
					break;
				case HelpFont1ID:
					ChangeFont(form, 0);
					break;
				case HelpFont2ID:
					ChangeFont(form, 1);
					break;
				case HelpFont3ID:
					ChangeFont(form, 2);
					break;
			}
		case ctlRepeatEvent:
			switch(event->data.ctlRepeat.controlID)
			{
				case HelpUpID:
					ButtonHandler(form, vchrPageUp);
					break;
				case HelpDownID:
					ButtonHandler(form, vchrPageDown);
					break;
				default:
					break;
			}
			break;
		case keyDownEvent:
			if(UtilIsHardBtn(event) && UtilIsArrowBtn(event))
			{
				ButtonHandler(form, UtilBtnDirection(event));
				handled = true;
			}
			break;
		default:
			break;
	}
	return(handled);
}

static void ChangeFont(FormPtr form, UInt8 ft)
{
	PrefsGetPrefs()->help_font = ft;
	setTop(false);
	setBottom(false);
	ShowHelpMessage(form, true);
}

static void ShowHelpMessageInit(FormPtr form, Boolean clean)
{
	RectangleType rect, field_rect;

	if(HelpIsError())
		FrmCopyTitle(form, HelpGetTitleErr());
	else
		FrmCopyTitle(form, HelpGetTitleHelp());

	FrmGetObjectBounds(form, FrmGetObjectIndex(form, HelpDoneID),
						&field_rect);

	UtilGetFrmBounds(form, &rect);

	x  = rect.topLeft.x;
	y  = rect.topLeft.y + title_height;
	w  = rect.extent.x  - x;
	h  = field_rect.topLeft.y - y - 2;
	sl = StrLen(msg);

	ShowHelpMessage(form, clean);
}

static void ShowHelpMessage(FormPtr form, Boolean clean)
{
	const Char *tmsg;
	RectangleType rect;
	Int16 len, fh, line;
	Coord ty;
	UnderlineModeType old_ul;
	FontID old_ft, ft;

	tmsg = msg;

	if(clean)
	{
		RctSetRectangle(&rect, x, y, w, h);
		WinEraseRectangle(&rect, 0);
	}

	ft = help_fonts[PrefsGetPrefs()->help_font];
	if(UtilIsVGA())
		old_ft = FntSetFont(VgaBaseToVgaFont(ft));
	else
		old_ft = FntSetFont(ft);
	old_ul = WinSetUnderlineMode(noUnderline);
	fh     = FntCharHeight();

	line = 0;
	len  = sl;
	ty   = y;
	while((ty < h) && (len >= 0))
	{
		UInt16 tlen;

		RctSetRectangle(&rect, x, ty, w, fh);
		WinEraseRectangle(&rect, 0);

		tlen = FntWordWrap(tmsg, w);
		if(tlen == 0)
		{
			setBottom(true);
			break;
		}

		if(line >= curr_line)
		{
			if(UtilCharIsSpace(tmsg[tlen-1]))
				WinDrawChars(tmsg, (tlen > 0 ? tlen-1 : tlen), x, ty);
			else
				WinDrawChars(tmsg, tlen, x, ty);
			ty += fh;
		}
		tmsg += tlen;
		len  -= tlen;
		line++;
	}

	FntSetFont(old_ft);
	WinSetUnderlineMode(old_ul);

	if(curr_line == 0)
		setTop(true);
}

static void ButtonHandler(FormPtr form, WChar direction)
{
	if(direction == vchrPageUp)
	{
		if(isTop() == false)
		{
			curr_line--;
			setBottom(false);
		}
	}
	else
	if(direction == vchrPageDown)
	{
		if(isBottom() == false)
		{
			curr_line++;
			setTop(false);
		}
	}

	ShowHelpMessage(form, false);
}

static void SizeForm(FormPtr form)
{
	RectangleType curBounds, displayBounds;
	Coord diff;

	UtilGetFrmBounds(form, &curBounds);
	UtilGetDpyBounds(&displayBounds);

	diff = displayBounds.extent.y - (curBounds.topLeft.y + curBounds.extent.y);

	if(diff)
	{
		curBounds.extent.y += diff;
		WinSetBounds(FrmGetWindowHandle(form), &curBounds);

		UtilMoveObject(form, HelpDownID, diff, true);
		UtilMoveObject(form, HelpUpID, diff, true);
		UtilMoveObject(form, HelpFont3ID, diff, true);
		UtilMoveObject(form, HelpFont2ID, diff, true);
		UtilMoveObject(form, HelpFont1ID, diff, true);
		UtilMoveObject(form, HelpDoneID, diff, true);
	}
}

static void setTop(Boolean b)
{
	is_top = b;
	UtilShowObject(FrmGetActiveForm(), HelpUpID, !is_top);
}

static void setBottom(Boolean b)
{
	is_bottom = b;
	UtilShowObject(FrmGetActiveForm(), HelpDownID, !is_bottom);
}

static Boolean isTop()
{
	return(is_top);
}

static Boolean isBottom()
{
	return(is_bottom);
}
