/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <Garmin/GPSLib68K.h>
#include "ique.h"
#include "time.h"
#include "util.h"
#include "main.h"
#include "MathLib.h"
#include "resource.h"

#define D_KMH2MS 0.278
#define D_NM2KM	 1.852
#define S_MS2KN  (1./D_KMH2MS/D_NM2KM)
#define secondsPerFrac (1. / 4294967296.) /* Seconds per fractional second */

static UInt16 gps_ref;
static UInt8 max_sat;
static GPSSatDataType *sat;

static void HandleGPSData(nmea_type *nmea) CODE2_SECTION;
static void HandleSatData(nmea_type *nmea) CODE2_SECTION;
static Err  GpsLibEventCallback(SysNotifyParamType *notifyParamsP) CODE2_SECTION;

Err IqueOpenGPS()
{
	Err err;

	max_sat = 0;
	sat = NULL;

	if((err = SysLibFind(gpsLibName, &gps_ref)))
	{
		if(err == sysErrLibNotFound)
			err = SysLibLoad(gpsLibType, gpsLibCreator, &gps_ref);

		if(err)
			UtilAlert(MainNOGPSLIB);
	}

	if(err == errNone)
	{
		err = GPSOpen(gps_ref);
		if(err != gpsErrNone)
		{
			UtilAlert(MainGPSLIBOPEN);
		}
		else
		{
			UInt16 cardNo;
			LocalID dbID;

			if(SysCurAppDatabase(&cardNo, &dbID) == errNone)
			{
				if(SysNotifyRegister(cardNo, dbID,
								sysNotifyGPSDataEvent, GpsLibEventCallback,
								sysNotifyNormalPriority, NULL) != errNone)
				{
					GPSClose(gps_ref);
					gps_ref = sysInvalidRefNum;
					UtilAlert(MainGPSLIBOPEN);
				}
			}

			max_sat = GPSGetMaxSatellites(gps_ref);
			sat = UtilMalloc(max_sat * sizeof(GPSSatDataType));
			if(max_sat > MAXSAT)
				max_sat = MAXSAT;
		}
	}
	return(err);
}

Err IqueCloseGPS()
{
	Err err;
	UInt16 cardNo;
	LocalID dbID;

	if(SysCurAppDatabase(&cardNo, &dbID) == errNone)
	{
		SysNotifyUnregister(cardNo, dbID, sysNotifyGPSDataEvent,
			sysNotifyNormalPriority);
	}

	if(sat != NULL)
		UtilFree(sat);

	err = GPSClose(gps_ref);
	if(err != gpsErrStillOpen)
		SysLibRemove(gps_ref);

	return(err);
}

static Err GpsLibEventCallback(SysNotifyParamType *nparam)
{
	DoIque((UInt32)nparam);
	return(errNone);
}


void IqueNmeaSim(UInt32 nparam, nmea_type *nmea)
{
	nmea->has_gps = true;
	HandleGPSData(nmea);
	HandleSatData(nmea);
/*
	switch(nparam)
	{
		case gpsModeChange:
		case gpsLostFix:
		case gpsStatusChange:
		case gpsLocationChange:
			HandleGPSData(nmea);
			break;
		case gpsSatDataChange:
			HandleSatData(nmea);
			break;
		default:
			break;
	}
*/
}

static void HandleGPSData(nmea_type *nmea)
{
	Err err;
	GPSPVTDataType pvt;
	DateTimeType pdatetime;
	double d;

	err = GPSGetPVT(gps_ref, &pvt);
	if(err != gpsErrNone)
		return;

	nmea->has_gga    = true;
	nmea->is_new_gga = true;
	nmea->has_gsa    = true;
	nmea->is_new_gsa = true;
	nmea->has_rmc    = true;
	nmea->is_new_rmc = true;

	switch(pvt.status.fix)
	{
		case gpsFix2D:
			nmea->fix_quality = 1;
			nmea->fix_dim = 2;
			break;
		case gpsFix3D:
			nmea->fix_quality = 1;
			nmea->fix_dim = 3;
			break;
		case gpsFix2DDiff:
			nmea->fix_quality = 2;
			nmea->fix_dim = 2;
			break;
		case gpsFix3DDiff:
			nmea->fix_quality = 2;
			nmea->fix_dim = 3;
			break;
		default:
		case gpsFixUnusable:
		case gpsFixInvalid:
			nmea->fix_quality = 0;
			nmea->fix_dim = 0;
			nmea->warning = 1;
	}

	nmea->mag_var_dir = '-';

	nmea->pdop     = pvt.status.epe;
	nmea->pdop_fra = (10 * pvt.status.epe) - nmea->pdop;
	nmea->pdop_dec = 1;
	if(nmea->pdop > 50)
	{
		nmea->pdop     = 50;
		nmea->pdop_fra = 0;
	}

	nmea->hdop     = pvt.status.eph;
	nmea->hdop_fra = (10 * pvt.status.eph) - nmea->hdop;
	nmea->hdop_dec = 1;
	if(nmea->hdop > 50)
	{
		nmea->hdop     = 50;
		nmea->hdop_fra = 0;
	}

	nmea->vdop     = pvt.status.epv;
	nmea->vdop_fra = (10 * pvt.status.epv) - nmea->vdop;
	nmea->vdop_dec = 1;
	if(nmea->vdop > 50)
	{
		nmea->vdop     = 50;
		nmea->vdop_fra = 0;
	}

	nmea->lat =  pvt.position.lat * (M_PI/0x80000000);
	nmea->lon = -pvt.position.lon * (M_PI/0x80000000);
	NMEAAveragePos(nmea);
	nmea->alt           = pvt.position.altMSL;
	nmea->alt_fra       = (10 * pvt.position.altMSL) - nmea->alt;
	nmea->alt_dec       = 1;
	nmea->alt_WGS84     = pvt.position.altWGS84;
	nmea->alt_WGS84_fra = (10 * pvt.position.altWGS84) - nmea->alt_WGS84;
	nmea->alt_WGS84_dec = 1;

	TimSecondsToDateTime(TimGetSeconds(), &pdatetime);
	TimAdjust(&pdatetime, 3600*TimeGetTimeOffset());

	nmea->date_of_fix_dy = pdatetime.day;
	nmea->date_of_fix_mo = pdatetime.month;
	nmea->date_of_fix_yr = pdatetime.year-2000;
	nmea->fix_taken_hour = (pvt.time.seconds/3600);
	nmea->fix_taken_min  = (pvt.time.seconds-nmea->fix_taken_hour*3600)/60;
	nmea->fix_taken_sec  = (pvt.time.seconds-nmea->fix_taken_hour*3600-nmea->fix_taken_min*60);
	nmea->fix_taken_hsec = (pvt.time.seconds+pvt.time.fracSeconds*secondsPerFrac);

	nmea->speed_f   = (double)pvt.velocity.speed*S_MS2KN;
	nmea->speed     = (Int32)nmea->speed_f;
	nmea->speed_fra = (Int32)(10.*(nmea->speed_f - (double)nmea->speed));
	nmea->speed_dec = 1;

	d = fmod(((double)(pvt.velocity.track+2*M_PI) * 180. / M_PI), 360);
	nmea->course     = (Int32)d;
	nmea->course_fra = (Int32)(10.*(d - (float)nmea->course));
	nmea->course_dec = 1;
}

static void HandleSatData(nmea_type *nmea)
{
	UInt8 i;

	GPSGetSatellites(gps_ref, sat);

	nmea->has_gga    = true;
	nmea->is_new_gga = true;
	nmea->has_gsv    = true;
	nmea->is_new_gsv = true;

	nmea->sat_tracked = 0;

	for(i = 0; i < MAXSAT; i++)
	{
		nmea->sat_data[i].prn       = 0;
		nmea->sat_data[i].elevation = 0;
		nmea->sat_data[i].azimuth   = 0;
		nmea->sat_data[i].signal    = 0;
	}

	for(i = 0; i < max_sat; i++)
	{
		nmea->sat_data[i].prn       = (Int32) sat[i].svid;
		nmea->sat_data[i].elevation = sat[i].elevation * 180. / M_PI;
		nmea->sat_data[i].azimuth   = sat[i].azimuth   * 180. / M_PI;
		nmea->sat_data[i].signal    = (Int32)(sat[i].snr/100.);

		if(nmea->sat_data[i].prn    == 255) // 255 mean unused slot
			nmea->sat_data[i].prn    = 0;

		if(sat[i].status & gpsSatUsedMask)
		{
			nmea->sat_prn[i] = sat[i].svid;
			nmea->sat_tracked++;
		}
		else
			nmea->sat_prn[i] = -1;
	}
}

