/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include <Garmin/GPSLib.h>
#include <unix_stdarg.h>
#include <HostControl.h>
#include "main.h"
#include "forms.h"
#include "serial.h"
#include "help.h"
#include "time.h"
#include "track.h"
#include "nmea.h"
#include "util.h"
#include "time.h"
#include "log.h"
#include "marker.h"
#include "prefs.h"
#include "beam.h"
#include "ique.h"
#include "dist.h"
#include "resource.h"
#include "MathLib.h"

static UInt16 StartApplication(void);
static void EventLoop(void);
static void StopApplication(void);
static Boolean HardbuttonHandleEvent(EventType *event);
static void SwitchForm(WChar direction);

static void DoGPS();
static void DoTick();
static void GPSHandleNewData();
static void GPSHandleNoData();
static Boolean DefaultFormHandler(EventPtr event);
static void GotoForm(UInt16 formId);
static void SyncTime();
static void ResetStats();
static void HandleDisplayChange(SysNotifyParamType* nparam);
static void SendGPSEvent(UInt16 type, MemPtr data);

static nmea_type *nmea;
static Int32 timeout;

/* order of forms */
static UInt16 formtable[] =
{
	GPSForm
	,GPS2Form
	,SatDataForm
	,TargetForm
	,TrackForm
#ifdef MOVINGMAP
	,MapDrawForm
#endif /* MOVINGMAP */
};

#define GPSTIMEOUT 15
#define DISPLAYTIMEOUT 30

static UInt32 time_now, time_last;
static UInt32 last_nmea_data;
static UInt32 next_track_save;
static Boolean is_track_save, save_track;
static Boolean sync_time;
static Boolean add_marker;
static Boolean menu_is_activated;
static Boolean new_ique_data;
static Boolean serial_on;
static Boolean want_serial;
static Boolean is_connecting;
static Boolean dispison;
static Boolean seen_data;
static UInt32 ique_data;
static dist_type dist;
static Char *title_str_demo = NULL;
static Char *title_str_log = NULL;
static Char *title_str_track = NULL;
static Char *title_str_sync = NULL;
static Char *title_str_marker = NULL;
static Char *title_str_nofix = NULL;
static Char *title_str_fix = NULL;

static void (*frmhelp)(UInt16 helpMsgId) = NULL;

UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
	int error;
	SysNotifyParamType *nparam;

	switch(cmd)
	{
		case sysAppLaunchCmdNormalLaunch:
			HostTraceInit();
			error = StartApplication();
			if(error != errNone)
				return(error);
			EventLoop();
			StopApplication();
			HostTraceClose();
			break;
		case sysAppLaunchCmdNotify:
			nparam = (SysNotifyParamType*) cmdPBP;
			switch(nparam->notifyType)
			{
				case sysNotifyDisplayChangeEvent:
				{
					Boolean app_active = (Boolean)(launchFlags & sysAppLaunchFlagSubCall);
					if(app_active)
					{
						HandleDisplayChange(nparam);
					}
					break;
				}
				case sysNotifyDisplayResizedEvent:
				{
					EventType resizedEvent;

					MemSet(&resizedEvent, sizeof(EventType), 0);
					resizedEvent.eType = winDisplayChangedEvent;
					EvtAddUniqueEventToQueue(&resizedEvent, 0, true);
					break;
				}
			}
			break;
		case sysAppLaunchCmdExgReceiveData:
			BeamReceive((ExgSocketPtr)cmdPBP, launchFlags & sysAppLaunchFlagSubCall);
			break;
		case sysAppLaunchCmdGoTo:
			{
				GoToParamsPtr gotoparam;
				Boolean	launched;

				gotoparam = (GoToParamsPtr)cmdPBP;
				launched = launchFlags & sysAppLaunchFlagNewGlobals;
				if(launched) 
				{
					error = StartApplication();
					if(error != errNone)
						return(error);

					PrefsGetPrefs()->marker_idx = gotoparam->recordNum;
					UtilPushFifo(PrefsGetPrefs()->start_form);
					FrmGotoForm(MarkerForm);

					EventLoop();
					StopApplication();	
				}
				else
				{
					FormPtr form;

					PrefsGetPrefs()->marker_idx = gotoparam->recordNum;
					form = FrmGetFormPtr(MarkerForm);
					if(form != NULL)
						FrmUpdateForm(MarkerForm, MARKERFORM_LISTADDED);
				}
			}
			break;
		default:
			break;
	}
	return(0);
}

static UInt16 StartApplication(void)
{
	Err err;

	nmea = NULL;

	UtilInit();

	if(UtilIsOS(3, 0) == false)
		UtilAlert(MainWRONGOS);

	if(UtilIsVGA() == true)
		VgaSetScreenMode(screenMode1To1, rotateModeNone);

	HelpInit();
	err = SysLibFind(MathLibName, &MathLibRef);
	if(err)
		err = SysLibLoad(LibType, MathLibCreator, &MathLibRef);
	if(err)
	{
		UtilAlert(MainNOMATHLIB);
		HelpFinish();
		UtilFinish();
		return(1);
	}
	err = MathLibOpen(MathLibRef, MathLibVersion);
	if(err)
	{
		UtilAlert(MainMATHLIBOPENFAILED);
		HelpFinish();
		UtilFinish();
		return(1);
	}

	timeout = SysTicksPerSecond()/20;
	time_now = TimGetSeconds();
	last_nmea_data = time_now;

	TrackSetCapturing(false);
	is_track_save = false;
	save_track = false;
	sync_time = false;
	add_marker = false;
	new_ique_data = false;
	serial_on = false;
	want_serial = true;
	is_connecting = false;
	dispison = true;
	seen_data = false;

	// OS 5 can't get patched anymore
	if(UtilIsOS(5,0) == false)
	{
		frmhelp = SysGetTrapAddress(sysTrapFrmHelp);
		err = SysSetTrapAddress(sysTrapFrmHelp, UtilFrmHelp);
		if(err)
		{
			UtilAlert(MainHELPPATCHFAILED); 
			frmhelp = NULL;
		}
	}

	TimeInit();
	nmea = UtilMalloc(sizeof(nmea_type));
	NMEAInit(nmea);

	if(SerialInit() == false)
	{
		NMEAFinish(nmea);
		SerialFinish();
		HelpFinish();
		UtilFinish();
		return(1);
	}
	PrefsInit(); // must be called after SerialInit!!!
	TrackSaveInit(PrefsGetPrefs()->track_name);
	DistInit(&dist, 0., 0., 0);

	title_str_demo   = UtilGetResourceString(TitleDEMO);
	title_str_log    = UtilGetResourceString(TitleLOG);
	title_str_track  = UtilGetResourceString(TitleTRACK);
	title_str_sync   = UtilGetResourceString(TitleSYNC);
	title_str_marker = UtilGetResourceString(TitleTAKEMARKER);
	title_str_nofix  = UtilGetResourceString(TitleNOFIX);
	title_str_fix    = UtilGetResourceString(TitleFIX);

	FrmGotoForm(PrefsGetPrefs()->start_form);

	if(PrefsHasPrefs() == true)
		SendGPSEvent(GpsStart, NULL);
	else
		FrmPopupForm(ConfigForm);

	return(0);
}

static void StopApplication()
{
	UInt16 usecount;
	Err err;

	FrmCloseAllForms();

	if(LogIsLogging() == true)
		LogEnd();

	DistFinish(&dist);
	TrackSaveFinish();
	SerialFinish();
	NMEAFinish(nmea);
	if(nmea != NULL)
	{
		UtilFree(nmea);
		nmea = NULL;
	}
	TimeFinish();
	HelpFinish();

	if(frmhelp != NULL)
	{
		err = SysSetTrapAddress(sysTrapFrmHelp, frmhelp);
		if(err)
			UtilAlert(MainHELPPATCHFAILED);
	}

	err = MathLibClose(MathLibRef, &usecount);
	if(err)
		UtilAlert(MainMATHLIBCLOSEFAILED);
	if(usecount == 0)
		SysLibRemove(MathLibRef);

	PrefsSave();
	PrefsFinish();

	if(title_str_demo != NULL)
		UtilFree(title_str_demo);
	if(title_str_log != NULL)
		UtilFree(title_str_log);
	if(title_str_track != NULL)
		UtilFree(title_str_track);
	if(title_str_sync != NULL)
		UtilFree(title_str_sync);
	if(title_str_marker != NULL)
		UtilFree(title_str_marker);
	if(title_str_nofix != NULL)
		UtilFree(title_str_nofix);
	if(title_str_fix != NULL)
		UtilFree(title_str_fix);

	UtilFinish();
}

static void EventLoop(void)
{
	Err err;
	EventType event;

	do
	{
		EvtGetEvent(&event, timeout);

		if(HardbuttonHandleEvent(&event))
			continue;

		if(SysHandleEvent(&event))
			continue;

		if(MenuHandleEvent((void *)0, &event, &err))
			continue;

		if(event.eType == winEnterEvent)
			if(event.data.winEnter.enterWindow == (WinHandle)FrmGetActiveForm())
				menu_is_activated = false;

		if(event.eType == winExitEvent)
			if(event.data.winExit.exitWindow == (WinHandle)FrmGetActiveForm())
				menu_is_activated = true;

		if(event.eType == frmLoadEvent)
		{
			int formID = event.data.frmLoad.formID;
			FormPtr form = FrmInitForm(formID);
			FrmSetActiveForm(form);

			switch(formID)
			{
				case GPSForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)GPSFormHandler);
						UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);
				break;
				case GPS2Form:
						FrmSetEventHandler(form, (FormEventHandlerPtr)GPS2FormHandler);
						UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);
				break;
				case ConfigForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)ConfigFormHandler);
				break;
				case MarkerForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)MarkerFormHandler);
						UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);
				break;
				case MarkerEditFormDD:
				case MarkerEditFormDMM:
				case MarkerEditFormDMS:
				case MarkerEditFormDMSS:
				case MarkerEditFormUTM:
				case MarkerEditFormOSGB:
				case MarkerEditFormQTH:
						FrmSetEventHandler(form, (FormEventHandlerPtr)MarkerEditFormHandler);
						UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);
				break;
				case SatDataForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)SatDataFormHandler);
						UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);
				break;
				case TargetForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)TargetFormHandler);
						UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);
				break;
				case WorldForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)WorldFormHandler);
						UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);
				break;
				case TrackForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)TrackFormHandler);
						UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);
				break;
				case TrackDrawForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)TrackDrawFormHandler);
						UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);
				break;
				case RawDataForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)RawDataFormHandler);
						UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);
				break;
				case TrackManagerForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)TrackManagerFormHandler);
						UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);
				break;
				case HelpForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)HelpFormHandler);
				break;
				case RenameForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)RenameFormHandler);
				break;
				case StartOdoForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)StartOdoFormHandler);
				break;
#ifdef MOVINGMAP
				case MapDrawForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)MapDrawFormHandler);
						UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);
				break;
#endif /* MOVINGMAP */
				default:
						FrmSetEventHandler(form, (FormEventHandlerPtr)DefaultFormHandler);
				break;
			}

			if(UtilHasDIA() == true)
			{
				WinHandle winhandle = FrmGetWindowHandle(form);

				WinSetConstraintsSize(winhandle, 160, 160, 225, 160, 160, 160);
				FrmSetDIAPolicyAttr(form, frmDIAPolicyCustom);
				PINSetInputTriggerState(pinInputTriggerEnabled);
				PINSetInputAreaState(pinInputAreaUser);
			}
		}

		if(event.eType == ctlSelectEvent)
		{
			if(event.data.ctlSelect.controlID == HelpBtnID)
			{
				// get help res-string and show
				UInt16 hres;

				hres = UtilGetHelpRes();
				if(hres != NULL)
					UtilFrmHelp(hres);
				continue;
			}

			if(event.data.ctlSelect.controlID == GPSOnOffID)
			{
				want_serial = !want_serial;
				UtilSetControlValue(FrmGetActiveForm(), GPSOnOffID, !want_serial);

				if(want_serial)
					MainStartGPS();
				else
					MainStopGPS();

				continue;
			}
		}

		if(event.eType == menuEvent)
		{
			Boolean menu_handled = true;
			menu_is_activated = true;

			switch(event.data.menu.itemID)
			{
				case MenuExitID:
						UtilExit();
					break;
				case MenuAboutID:
						FrmPopupForm(AboutForm);
					break;
				case DebugDataID:
						SerialSetDebug(!SerialIsDebug());
						MainSetGPSTitle(FrmGetActiveForm());
					break;
				case ShowRawDataID:
						UtilPopupForm(RawDataForm);
					break;
				case HelpID:
						UtilFrmHelp(IntroHelpStrID);
					break;
				case ConfigID:
						FrmPopupForm(ConfigForm);
					break;
				case ResetStatsID:
						ResetStats();
					break;
				case SyncTimeID:
						SyncTime();
					break;
				case MarkerFormID:
						UtilPopupForm(MarkerForm);
					break;
				case ShowWorldID:
						UtilPopupForm(WorldForm);
					break;
				case GPSFormID:
						GotoForm(GPSForm);
					break;
				case GPS2FormID:
						GotoForm(GPS2Form);
					break;
				case SatDataID:
						GotoForm(SatDataForm);
					break;
				case TrackFormID:
						GotoForm(TrackForm);
					break;
				case TargetFormID:
						GotoForm(TargetForm);
					break;
				case LogDBID:
						if(SerialIsIque())
							UtilBeep(BEEP_ERROR);
						else
						{
							if(LogIsLogging() == false)
								LogStart();
							else
								LogEnd();
						}
					break;
				case ResetLogDBID:
						if(LogIsLogging() == true)
						{
							LogEnd();
							LogRemove();
							LogStart();
						}
						else
							LogRemove();
					break;
				default:
						menu_handled = false;
					break;
			}
			menu_is_activated = false;
			if(menu_handled)
				continue;
		}

		if(event.eType == GpsStart)
			MainStartGPS();

		if(event.eType == GpsStop)
			MainStopGPS();

		if(event.eType == nilEvent)
		{
			time_last = time_now;
			time_now = TimGetSeconds();
			if(time_now != time_last)
				DoTick();
			DoGPS();
		}

		FrmDispatchEvent(&event);
	}
	while(event.eType != appStopEvent);
}

static Boolean HardbuttonHandleEvent(EventType *event)
{
	if(	event->eType == keyDownEvent &&
		TxtCharIsHardKey(	event->data.keyDown.modifiers,
							event->data.keyDown.chr))
	{
		if(event->data.keyDown.chr == vchrHard1)
		{
			TrackSavePosition(nmea, true);
			return(true);
		}
		if(event->data.keyDown.chr == vchrHard2)
		{
			TrackSetCapturing(!TrackIsCapturing());
			return(true);
		}
		if(event->data.keyDown.chr == vchrHard3)
		{
			dispison = !dispison;
			UtilSwitchDisplay(dispison);
			return(true);
		}
		if(event->data.keyDown.chr == vchrHard4)
		{
			add_marker = true;
			return(true);
		}
	}

	if((event->eType == keyDownEvent) && UtilIsHardBtn(event))
	{
		if(UtilIsArrowBtn(event))
		{
			if(FrmDispatchEvent(event) == false)
			{
				SwitchForm(UtilBtnDirection(event));
			}
			return(true);
		}
		if((event->data.keyDown.chr == vchrFind) ||
			(event->data.keyDown.chr == vchrCalc))
		{
			SndPlaySystemSound(sndError);
			return(true);
		}
	}
	return(false);
}

static void SwitchForm(WChar direction)
{
	UInt16 actfrm, nfrm, i, n, pos, updwn;

	n = sizeof(formtable) / sizeof(formtable[0]);
	updwn = (direction == vchrPageDown) ? 1 : n-1;
	actfrm = FrmGetActiveFormID();
	pos = 0;
	for(i = 0; i < n; i++)
	{
		if(formtable[i] == actfrm)
		{
			pos = i;
			break;
		}
	}
	nfrm = (pos+updwn)%n;
#ifdef MOVINGMAP
	if(UtilIsOS(3, 5) == false)
	{
		if(formtable[nfrm] == MapDrawForm)
			nfrm = (pos+updwn+updwn)%n;
	}
#endif /* MOVINGMAP */
	GotoForm(formtable[nfrm]);
}

static void GotoForm(UInt16 formId)
{
	PrefsGetPrefs()->start_form = formId;
	FrmGotoForm(formId);
}

void ResetStats()
{
	CAppPrefs *prefs = PrefsGetPrefs();

	DistInit(&dist, nmea->lat, nmea->lon, 0);

	prefs->dist_f    = 0.;
	prefs->max_speed = 0;
	prefs->min_alt   = 0;
	prefs->max_alt   = 0;

	if((nmea->has_gps == true) && (nmea->fix_quality > 0))
	{
		nmea->is_new_gga = true;
		nmea->is_new_gsa = true;
		nmea->is_new_rmc = true;
		nmea->is_new_gsv = true;
		nmea->is_new_pmgnst = true;
		SendGPSEvent(GpsNmeaEvent, nmea);
	}
}

void SyncTime()
{
	if(SerialIsDebug() == false)
		sync_time = true;
	else
		SndPlaySystemSound(sndError);
}

void MainSetGPSTitle(FormPtr form)
{
	Char title[TITLE_MAX];
	Int32 hdop, hdop_fra;

	if(!serial_on)
	{
		SetTitleRes(form, TitleGPSOFF);
		return;
	}

	if(nmea->has_gps == false)
	{
		SetTitleRes(form, TitleNOGPS);
		return;
	}

	if(nmea->has_gga)
	{
		hdop     = nmea->gga_hdop;
		hdop_fra = nmea->gga_hdop_fra;
	}
	else
	{
		hdop     = nmea->hdop;
		hdop_fra = nmea->hdop_fra;
	}

	switch(nmea->fix_quality)
	{
		case 0:
			StrPrintF(title, "%s (%ld sat)", title_str_nofix, nmea->sat_tracked);
		break;
		case 1:
			if(nmea->has_gsa)
				StrPrintF(title, "%ldD %ld sat (%ld.%ld)",
								nmea->fix_dim,
								nmea->sat_tracked,
								hdop,
								hdop_fra);
			else
				StrPrintF(title, "%s %ld sat (%ld.%ld)",
								title_str_fix,
								nmea->sat_tracked,
								hdop,
								hdop_fra);
		break;
		case 2:
			StrPrintF(title, "DGPS %ld sat (%ld.%ld)",
								nmea->sat_tracked,
								hdop,
								hdop_fra);
		break;
		/* This shouldn't happen */
		default:
			StrPrintF(title, "%ldDGPS %ld sat (%ld.%ld)",
								nmea->fix_quality,
								nmea->sat_tracked,
								hdop,
								hdop_fra);
		break;
	}

	SetTitle(form, title, false);
}

void SetTitleRes(FormPtr form, UInt16 res_id)
{
	Int8 i, len;
	Char *title;

	title = UtilGetResourceString(res_id);
	len = MIN(StrLen(title), TITLE_MAX);
	title[len] = chrNull;

	for(i = len-1; i >= 0; i--)
	{
		if(title[i] == chrSpace)
			title[i] = chrNull;
		else
			break;
	}

	SetTitle(form, title, true);
	UtilFree(title);
}

void SetTitle(FormPtr form, Char *title, Boolean lean)
{
	Char wintitle[TITLE_MAX];

	StrNCopy(wintitle, title, TITLE_MAX);

	if(lean == false)
	{
		if(SerialIsDebug())
			StrNCat(wintitle, title_str_demo, TITLE_MAX);

		if(LogIsLogging())
			StrNCat(wintitle, title_str_log, TITLE_MAX);

		if(TrackIsCapturing())
			StrNCat(wintitle, title_str_track, TITLE_MAX);

		if(sync_time == true)
			StrNCat(wintitle, title_str_sync, TITLE_MAX);

		if(add_marker == true)
			StrNCat(wintitle, title_str_marker, TITLE_MAX);
	}

	FrmCopyTitle(form, wintitle);
}

void DoIque(UInt32 nparam)
{
	new_ique_data = true;
	ique_data = nparam;
}

static void DoGPS()
{
	static char rawdata[NMEA_MAX_LENGTH];
	Boolean new_data = false;

	if(serial_on)
	{
		if(SerialIsIque())
		{
			if(new_ique_data)
			{
				nmea->is_new_gga = false;
				nmea->is_new_gsa = false;
				nmea->is_new_rmc = false;
				nmea->is_new_gsv = false;
				nmea->is_new_pmgnst = false;
				IqueNmeaSim(ique_data, nmea);
				new_data = true;
				new_ique_data = false;
			}
		}
		else
		{
			if(SerialNextLine(rawdata))
			{
				new_data = true;
				if(LogIsLogging() == true)
					LogAppend(rawdata);
				SendGPSEvent(GpsRawEvent, rawdata);
				nmea->is_new_gga = false;
				nmea->is_new_gsa = false;
				nmea->is_new_rmc = false;
				nmea->is_new_gsv = false;
				nmea->is_new_pmgnst = false;
				NMEAReadSentence(rawdata, nmea);
			}
		}
	}

	if(new_data || !seen_data)
		EvtResetAutoOffTimer();

	if(new_data)
	{
		seen_data = true;
		GPSHandleNewData();
		SendGPSEvent(GpsNmeaEvent, nmea);
	}
	else
	{
		GPSHandleNoData();
	}
}

static void DoTick()
{
	if(TrackIsCapturing() == true)
	{
		if(is_track_save == false)
		{
			next_track_save = time_now;
			is_track_save = true;
		}

		if(next_track_save <= time_now)
		{
			save_track = true;
			next_track_save = time_now + TrackGetSaveInterval();
		}
	}
	else
		is_track_save = false;
}

static void SendGPSEvent(UInt16 type, MemPtr data)
{
	EventType evt;
	gps_ptr_type *ptr;

	if(menu_is_activated == true)
		return;

	MemSet(&evt, sizeof(evt), 0);
	evt.eType = type;
	ptr = (gps_ptr_type*)evt.data.generic.datum;
	ptr->data = data;
	EvtAddEventToQueue(&evt);
}

static void GPSHandleNoData()
{
	if(SerialIsIque())
	{
		if(time_now - last_nmea_data > 2)
		{
			last_nmea_data = time_now;
			DoIque(gpsSatDataChange);
			return;
		}
	}

	if(	(is_connecting == false)	&&
		(want_serial == true)		&&
		(time_now - last_nmea_data > GPSTIMEOUT))
	{
		MainStopGPS();
		MainStartGPS();
		last_nmea_data = time_now;
	}
}

static void GPSHandleNewData()
{
	last_nmea_data = time_now;

	if((nmea->has_gps == true) && (nmea->warning == 0) && (nmea->fix_quality > 0))
	{
		if(nmea->is_new_rmc || nmea->is_new_gga || nmea->is_new_nexrmc)
		{
			TimeSetTime(nmea->fix_taken_hour, nmea->fix_taken_min,
						nmea->fix_taken_sec);

			if(nmea->has_rmc || nmea->has_nexrmc)
			{
				TimeSetDate(nmea->date_of_fix_yr, nmea->date_of_fix_mo,
							nmea->date_of_fix_dy);
			}
			else
			{
				TimeSetOwnDate();
			}

			if(sync_time == true)
			{
				TimeSetPalmTime();
				SndPlaySystemSound(sndInfo);
				sync_time = false;
			}
		}

		if(nmea->is_new_rmc || nmea->is_new_vtg || nmea->is_new_nexrmc)
		{
			CAppPrefs *prefs = PrefsGetPrefs();
			prefs->max_speed = MAX(nmea->speed, prefs->max_speed);

			if(nmea->speed_f < (double)PrefsGetPrefs()->drift/10.)
				nmea->speed_f = 0.;
		}

		if(nmea->is_new_gga || nmea->is_new_nexrmc)
		{
			CAppPrefs *prefs = PrefsGetPrefs();

			/* don't update dist when stopping, fixes ugly odo updates */
			/* while standing at red lights.                           */
			if(nmea->speed_f > 0.)
				prefs->dist_f += DistGetDistUpd(&dist, nmea->lat, nmea->lon);
			prefs->max_alt = MAX(nmea->alt, prefs->max_alt);
			prefs->min_alt = MIN(nmea->alt, prefs->min_alt);

			if(save_track == true)
			{
				TrackSavePosition(nmea, false);
				save_track = false;
			}

			if(add_marker == true)
			{
				Err err;
				MarkerAddMarker(NULL, nmea->lon, nmea->lat, &err);
				SndPlaySystemSound(sndInfo);
				add_marker = false;
			}
		}
	}
	else
	{
		if(sync_time == true)
		{
			SndPlaySystemSound(sndError);
			sync_time = false;
		}
		if(add_marker == true)
		{
			SndPlaySystemSound(sndError);
			add_marker = false;
		}
	}
}

static Boolean DefaultFormHandler(EventPtr event)
{
	Boolean handled = false;
	FormPtr	form = FrmGetActiveForm();

	switch(event->eType)
	{
		case frmOpenEvent:
			if(UtilIsVGA())
				VgaFormModify(form, vgaFormModify160To240);
			FrmDrawForm(form);
			handled = true;
			break;
		case displayExtentChangedEvent:
			FrmDrawForm(form);
			handled = true;
			break;
		case ctlSelectEvent:
			FrmReturnToForm(0);
			handled = true;
			break;
		default:
			break;
	}
	return(handled);
}

void MainStartGPS()
{
	if(SerialIsRunning())
		return;

	is_connecting = true;
	NMEAClear(nmea);

	if(SerialOpenPortByIndex(	PrefsGetPrefs()->port_index,
								PrefsGetPrefs()->port_speed_index) == errNone)
	{
		serial_on = true;
	}
	else
	{
		SerialClosePort();
		serial_on = false;
	}
	is_connecting = false;
}

void MainStopGPS()
{
	serial_on = false;
	if(SerialIsRunning())
		SerialClosePort();
}

static void HandleDisplayChange(SysNotifyParamType* nparam)
{
	EventType evt;
	SysNotifyDisplayChangeDetailsType* det;

	det = (SysNotifyDisplayChangeDetailsType*) nparam->notifyDetailsP;
	if(det->oldDepth == det->newDepth)
	{
		MemSet(&evt, sizeof(evt), 0);
		evt.eType = displayExtentChangedEvent;
		EvtAddUniqueEventToQueue(&evt, 0, true);
	}
}

