/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include "main.h"
#include "nmea.h"
#include "marker.h"
#include "prefs.h"
#include "resource.h"
#include "dist.h"
#include "util.h"
#include "beam.h"

#define NONOTE "*empty*\0"
#define DEFMARKERNAME "Marker"

static UInt16 dest_index;
static Int16 edit_index, view_index;
static dist_type dist, sortdist;
static Boolean is_editable;

static Boolean MarkerExists(Char *name);

typedef struct
{
	UInt16 index;
	double dist;
} dist_sorter;

void MarkerInit()
{
	edit_index = 0;
	view_index = -1;
	dest_index = PrefsGetPrefs()->marker_idx;
	MarkerSetDestination(dest_index);
	DistInit(&sortdist, 0., 0., 0);
}

void MarkerFinish()
{
}

void MarkerSetDestination(UInt16 di)
{
	marker_type dest;

	if(di >= MarkerGetMarkerCount())
		di = -1;

	dest_index = di;

	PrefsGetPrefs()->marker_idx = dest_index;

	dest.lon = 0.;
	dest.lat = 0.;

	if(dest_index != -1)
		MarkerGetMarker(dest_index, &dest);

	DistInit(&dist, dest.lat, dest.lon, 0);
}

void MarkerClearDestination()
{
	dest_index = -1;
	PrefsGetPrefs()->marker_idx = dest_index;

	DistInit(&dist, 0., 0., 0);
}

Boolean MarkerHasDestination()
{
	return(dest_index != -1);
}

Int32 MarkerGetCourse(nmea_type *nmea)
{
	Int32 course = 0;

	if(MarkerHasDestination())
		course = DistGetCourseDeg(&dist, nmea->lat, nmea->lon);

	return(course);
}

UInt32 MarkerGetDistance(nmea_type *nmea)
{
	double ddist = 0.;

	if(MarkerHasDestination())
	{
		ddist = DistGetDist(&dist, nmea->lat, nmea->lon);
		PrefsNauticToPref(&ddist, NULL);
	}

	return((UInt32)ddist);
}

void MarkerEdit(Int16 index, Boolean b)
{
	is_editable = b;
	edit_index  = index;

	switch(PrefsGetPrefs()->format_coord)
	{
		case 0:
			UtilPopupForm(MarkerEditFormDD);
		break;
		case 2:
			UtilPopupForm(MarkerEditFormDMS);
		break;
		case 3:
			UtilPopupForm(MarkerEditFormDMSS);
		break;
		case 4:
			UtilPopupForm(MarkerEditFormUTM);
		break;
		case 5:
			UtilPopupForm(MarkerEditFormQTH);
		break;
		case 6:
			UtilPopupForm(MarkerEditFormOSGB);
		break;
		default:
			UtilPopupForm(MarkerEditFormDMM);
		break;
	}
}

Int16 MarkerGetEditIndex()
{
	return(edit_index);
}

void MarkerSetViewIndex(Int16 index)
{
	view_index = index;
}

Int16 MarkerGetViewIndex()
{
	return(view_index);
}

Boolean MarkerIsEditable()
{
	return(is_editable);
}


DmOpenRef MarkerOpenDB(UInt16 mode)
{
	DmOpenRef db;

	db = UtilDBOpen(CGDBMarker, CGCreator, CGDBMarkerType, mode);
	if(db != NULL)
	{
		if(DmGetAppInfoID(db) == 0)
		{
			LocalID db_id;
			UInt16 card_no;

			/* Get database local ID and card number. We need these to
			 * initialize app info block.
			 */
			if(DmOpenDatabaseInfo(db, &db_id, NULL, NULL, &card_no, NULL) == errNone)
			{
				/* Allocate app info in storage heap. */
				MemHandle handle = DmNewHandle(db, sizeof(AppInfoType));
				if(handle != NULL)
				{
					AppInfoType *app_info;
					LocalID app_info_id;

					/* Associate app info with database. */
					app_info_id = MemHandleToLocalID(handle);
					DmSetDatabaseInfo(card_no, db_id, NULL, NULL, NULL, NULL, NULL,
									NULL, NULL, &app_info_id, NULL, NULL, NULL);

					/* Initialize app info block to 0. */
					app_info = MemHandleLock(handle);
					DmSet(app_info, 0, sizeof(AppInfoType), 0);

					/* Initialize the categories. */
					CategoryInitialize((AppInfoPtr)app_info, StringMARKERCATEGORY);

					/* Unlock the app info block. */
					MemHandleUnlock(handle);
				}
			}
		}
	}
	return(db);
}

void MarkerCloseDB(DmOpenRef db)
{
	DmCloseDatabase(db);
}

static Int16 sort_alpha(void *rec1, void *rec2, Int16 other,
				SortRecordInfoPtr rec1SortInfo,
				SortRecordInfoPtr rec2SortInfo,
				MemHandle app_info)
{
	marker_type *mk1, *mk2;

	mk1 = (marker_type*)rec1;
	mk2 = (marker_type*)rec2;

	return(StrCompare(mk1->name, mk2->name));
}

/*
static Int16 sort_dist(void *rec1, void *rec2, Int16 other,
				SortRecordInfoPtr rec1SortInfo,
				SortRecordInfoPtr rec2SortInfo,
				MemHandle app_info)
{
	double dist1, dist2;
	marker_type *mk1, *mk2;

	mk1 = (marker_type*)rec1;
	mk2 = (marker_type*)rec2;
	dist1 = DistGetDistValue(&sortdist, mk1->lat, mk1->lon);
	dist2 = DistGetDistValue(&sortdist, mk2->lat, mk2->lon);

	if(dist1 > dist2)
		return(1);
	if(dist1 < dist2)
		return(-1);
	return(0);
}
*/

static Int16 sort_dist(void *p1, void *p2, Int32 other)
{
	dist_sorter *ds1 = (dist_sorter*)p1;
	dist_sorter *ds2 = (dist_sorter*)p2;

	if(ds1->dist > ds2->dist)
		return(1);
	if(ds1->dist < ds2->dist)
		return(-1);
	return(0);
}

Boolean MarkerQSort()
{
	return(MarkerSort(sort_alpha));
}

/*
Boolean MarkerDSort(double lon, double lat)
{
	Boolean ret;

	DistInit(&sortdist, lat, lon, 0);
	ret = MarkerSort(sort_dist);
	DistInit(&sortdist, 0., 0., 0);
	return(ret);
}
*/

/*
 * Hier werden die Abstnde einmalig berechnet und in einem temporren Array
 * abgelegt. Dieses wird dann sortiert und die ebenfalls gespeicherten Indizes
 * werden benutzt, um die DB dementsprechend neu zu sortieren. Wir sparen uns
 * die hufige Entfernungsberechnung im Gegensatz zu frher.
 *
 * Das stndige ffnen/Schlieen der DB fra VIEL Zeit.
 */
Boolean MarkerDSort(double lon, double lat)
{
	DmOpenRef db;
	UInt16 i, mcnt;
	dist_sorter *dists;

	DistInit(&sortdist, lat, lon, 0);

	db = MarkerOpenDB(dmModeWrite);
	if(db == NULL)
		return(false);

	mcnt = DmNumRecords(db);

	dists = UtilMalloc(mcnt * sizeof(dist_sorter));
	if(dists == NULL)
	{
		MarkerCloseDB(db);
		return(false);
	}

	for(i = 0; i < mcnt; i++)
	{
		MemHandle mrec;

		dists[i].index = i;
		dists[i].dist  = 0;

		mrec = DmQueryRecord(db, i);
		if(mrec != NULL)
		{
			marker_type *marker = (marker_type *)MemHandleLock(mrec);
			dists[i].dist  = DistGetDistValue(&sortdist, marker->lat, marker->lon);
			MemHandleUnlock(mrec);
		}
	}

	SysQSort(dists, mcnt, sizeof(dist_sorter), sort_dist, 0);

	for(i = 0; i < mcnt; i++)
	{
		UInt16 from, j;

		from = dists[i].index;
		if(from != i)
			DmMoveRecord(db, from, i);

		for(j = i+1; j < mcnt; j++)
			if(dists[j].index < from)
				dists[j].index++;
	}

	UtilFree(dists);
	MarkerCloseDB(db);
	DistInit(&sortdist, 0., 0., 0);

	return(true);
}

Boolean MarkerSort(DmComparF sort)
{
	Err err;
	DmOpenRef db;

	db = MarkerOpenDB(dmModeWrite);
	if(db == NULL)
		return(false);

	err = DmQuickSort(db, sort, 0);
	MarkerCloseDB(db);

	if(err != errNone)
		return(false);

	return(true);
}

Boolean MarkerMoveMarker(UInt16 from, UInt16 to)
{
	Err err;
	DmOpenRef db;

	db = MarkerOpenDB(dmModeWrite);
	if(db == NULL)
		return(false);

	err = DmMoveRecord(db, from, to);
	MarkerCloseDB(db);

	if(err != errNone)
		return(false);

	return(true);
}

static Boolean MarkerExists(Char *name)
{
	marker_type marker;
	UInt32 i, cnt;

	cnt = MarkerGetMarkerCount();
	for(i = 0; i < cnt; i++)
		if(MarkerGetMarker(i, &marker) == true)
			if(StrCompare(name, marker.name) == 0)
				return(true);

	return(false);
}

UInt16 MarkerAddMarker(Char *name, double lon, double lat, Err *err)
{
	marker_type marker;

	if(name == NULL)
	{
		UInt32 i = 0;
		Char tname[MAX_MARKER_NAME_LENGTH];

		while(true)
		{
			StrPrintF(tname, "%s%ld", DEFMARKERNAME, i++);
			if(MarkerExists(tname) == false)
			{
				StrNCopy(marker.name, tname, MAX_MARKER_NAME_LENGTH);
				break;
			}
		}
	}
	else
		StrNCopy(marker.name, name, MAX_MARKER_NAME_LENGTH);

	marker.lon = lon;
	marker.lat = lat;

	return(UtilDBSaveRecord(CGDBMarker, CGCreator, CGDBMarkerType,
						&marker, sizeof(marker_type), err));
}

void MarkerUpdateMarker(UInt16 index, Char *name, double lon, double lat, Err *err)
{
	marker_type marker;

	marker.lon  = lon;
	marker.lat  = lat;
	StrNCopy(marker.name, name, MAX_MARKER_NAME_LENGTH);

	UtilDBUpdateRecord(CGDBMarker, CGCreator, CGDBMarkerType,
						index, &marker, sizeof(marker_type), err);
}

void MarkerDelMarker(UInt16 index, Err *err)
{
	UtilDBRemRecord(CGDBMarker, index, err);
}

UInt32 MarkerGetMarkerCount()
{
	return(UtilDBGetCountByName(CGDBMarker));
}

UInt32 MarkerGetMarkerCountInCategory(UInt16 category)
{
	return(UtilDBGetCountInCategory(CGDBMarker, category));
}

Boolean MarkerGetMarker(UInt16 index, marker_type *marker)
{
	DmOpenRef db;
	MemHandle mrec;

	if(index >= MarkerGetMarkerCount())
		return(false);

	db = MarkerOpenDB(dmModeReadOnly);
	if(db == NULL)
		return(false);

	mrec = DmQueryRecord(db, index);
	if(mrec != NULL)
	{
		marker_type *dbmarker;

		dbmarker = (marker_type *)MemHandleLock(mrec);

		marker->lon = dbmarker->lon;
		marker->lat = dbmarker->lat;
		StrNCopy(marker->name, dbmarker->name, MAX_MARKER_NAME_LENGTH);

		MemHandleUnlock(mrec);
	}
	MarkerCloseDB(db);
	return(true);
}

Boolean MarkerGetMarkerDescription(marker_note_type *note)
{
	if(note->index >= MarkerGetMarkerCount())
		return(false);

	note->offset = sizeof(marker_type);
	note->tindex = -1;
	if(note->mode != dmModeReadOnly)
	{
		MemPtr p;
		UInt16 nindex = dmMaxRecordIndex;

		/* get size of original record */
		note->db = MarkerOpenDB(note->mode);
		if(note->db == NULL)
			return(false);
		note->mhandle = DmGetRecord(note->db, note->index);
		note->size = MemHandleSize(note->mhandle) - note->offset;
		DmReleaseRecord(note->db, note->index, false);

		/* copy original record to end */
		note->mhandle = DmNewRecord(note->db, &nindex, note->size + note->offset);
		if(note->mhandle == 0)
		{
			MarkerCloseDB(note->db);
			note->mode = dmModeReadOnly;
		}
		else
		{
			MemHandle ohandle;
			MemPtr p2;

			note->tindex = nindex;

			/* copy original record over */
			ohandle = DmGetRecord(note->db, note->index);
			p  = MemHandleLock(note->mhandle);
			p2 = MemHandleLock(ohandle);
			if(DmWrite(p, 0, p2, note->size + note->offset) != errNone)
			{
				DmRemoveRecord(note->db, note->tindex);
				MarkerCloseDB(note->db);
				note->tindex = -1;
				note->mode = dmModeReadOnly;
			}
			MemHandleUnlock(ohandle);
			MemHandleUnlock(note->mhandle);
			DmReleaseRecord(note->db, note->index, false);
		}
	}

	if(note->mode == dmModeReadOnly)
	{
		note->db = MarkerOpenDB(note->mode);
		if(note->db == NULL)
			return(false);

		note->mhandle = DmQueryRecord(note->db, note->index);
		note->size = MemHandleSize(note->mhandle) - note->offset;
	}

	/* resize record to contain at least a null-length description */
	if((note->size < 1) && (note->mode != dmModeReadOnly))
	{
		MemPtr p;
		UInt16 index;

		if(note->tindex != -1)
			index = note->tindex;
		else
			index = note->index;

		DmReleaseRecord(note->db, index, false);
		DmResizeRecord(note->db, index, note->offset+1);
		note->mhandle = DmGetRecord(note->db, index);
		p = MemHandleLock(note->mhandle);
		DmWrite(p, note->offset, "\0", 1);
		note->size    = MemPtrSize(p) - note->offset;
		MemHandleUnlock(note->mhandle);
	}

	return(true);
}

void MarkerUpdateMarkerDescription(marker_note_type *note, Boolean update)
{
	if(note->mode != dmModeReadOnly)
	{
		if(update == false)
		{
			if(note->tindex != -1)
			{
				DmReleaseRecord(note->db, note->tindex, false);
				DmRemoveRecord(note->db, note->tindex);
			}
			else
				DmReleaseRecord(note->db, note->index, false);
		}
		else
		{
			DmReleaseRecord(note->db, note->tindex, false);
			DmRemoveRecord(note->db, note->index);
			DmMoveRecord(note->db, note->tindex-1, note->index);
		}
	}
	MarkerCloseDB(note->db);
	note->db = NULL;
}

void MarkerBeam(Int16 index)
{
	Char *description;
	marker_type marker;

	if(index < 0)
		return;

	if(MarkerGetMarker(index, &marker) == true)
		description = UtilStrDup(marker.name);
	else
		description = UtilGetResourceString(StringMARKER);

	BeamRecord(CGDBMarker, CGDBMarkerType, BEAM_MARKER, description, &index);
	MemPtrFree(description);
}

void MarkerBeamList()
{
	Char *description = UtilGetResourceString(StringMARKERLIST);
	BeamRecord(CGDBMarker, CGDBMarkerType, BEAM_MARKER, description, NULL);
	MemPtrFree(description);
}

void MarkerBeamDB()
{
	BeamDB(CGDBMarker, BEAM_MARKERDB);
}

void MarkerToClipboard(Int16 index)
{
	marker_type marker;

	if(index < 0)
		return;

	if(MarkerGetMarker(index, &marker) == true)
	{
		Char buf[64];
		Int32 adeg, amin, afrac;
		Int32 odeg, omin, ofrac;
		Char ac, oc;

		NMEAGetDeg_DMM(marker.lat, 3, &adeg, &amin, &afrac, &ac, LATLON_LAT);
		NMEAGetDeg_DMM(marker.lon, 3, &odeg, &omin, &ofrac, &oc, LATLON_LON);
		StrPrintF(buf, "%c %ld %ld.%03ld' %c %ld %ld.%03ld'",
					ac, adeg, amin, afrac, oc, odeg, omin, ofrac);
		ClipboardAddItem(clipboardText, buf, StrLen(buf));
	}
}

