/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include "main.h"
#include "forms.h"
#include "util.h"
#include "serial.h"
#include "prefs.h"
#include "dist.h"
#include "marker.h"
#include "resource.h"

static void UpdateMarkerList(FormPtr form, Int16 index) CODE3_SECTION;
static void UpdateFields(FormPtr form, Int16 index) CODE3_SECTION;
static void UpdateCategory(FormPtr form, Int16 index) CODE3_SECTION;
static void GPSHandler(FormPtr form, nmea_type *nmea) CODE3_SECTION;
static void UpdateData(FormPtr form) CODE3_SECTION;
static void ButtonHandler(FormPtr form, WChar direction) CODE3_SECTION;
static void SizeForm(FormPtr form) CODE3_SECTION;
static void ToggleRef(FormPtr form) CODE3_SECTION;
static void MoveMarker(FormPtr form, UInt16 id) CODE3_SECTION;
static void SelectCategory(FormPtr form) CODE3_SECTION;
static void ListDrawFunc(Int16 item, RectangleType *bounds, Char **data) CODE3_SECTION;
static void FreeListIndexes() CODE3_SECTION;
static Boolean LookupString(FormPtr form, EventPtr event) CODE3_SECTION;

static Boolean form_ready = false;
static dist_type dist;
static double cur_lat, cur_lon;
static Boolean has_coord, has_ref;
static Int16 refindex;
static UInt16 reflistindex;
static UInt16 *list_indexes;
static UInt16 list_indexes_cnt;
static UInt32 list_cnt;
static Char filter[MAX_MARKER_NAME_LENGTH];
static Int16 filter_len = 0;

Boolean MarkerFormHandler(EventPtr event)
{
	FormPtr	form;
	Boolean handled = false;

	form = FrmGetActiveForm();

	switch(event->eType)
	{
		case frmOpenEvent:
			form_ready = false;
			list_indexes = NULL;
			list_indexes_cnt = 0;
			list_cnt = 0;
			DistInit(&dist, 0., 0., 0);
			cur_lat = 0.;
			cur_lon = 0.;
			has_coord = false;
			has_ref   = false;
			refindex  = -1;
			reflistindex  = -1;
			if(UtilIsVGA())
				VgaFormModify(form, vgaFormModify160To240);
			if(UtilIsResizeableNoDia())
				SizeForm(form);
			FrmDrawForm(form);
			SetTitleRes(form, TitleMARKER);
			PrefsUpdateLabelLatLon(form, LatLabelID, LonLabelID, false);
			LstSetDrawFunction(UtilGetObjectPtr(form, DestListID), ListDrawFunc);
			UpdateMarkerList(form, MarkerGetViewIndex());
			if(filter_len > 0)
				UtilSetFieldText(form, DestLookUpID, filter);
			FrmSetFocus(form, FrmGetObjectIndex(form, DestLookUpID));
			form_ready = true;
			handled = true;
			break;
		case frmCloseEvent:
			form_ready = false;
			FreeListIndexes();
			break;
		case frmUpdateEvent:
			handled = true;
			switch(event->data.frmUpdate.updateCode)
			{
				case frmRedrawUpdateCode:
					FrmDrawForm(form);
					break;
				case MARKERFORM_LISTADDED:
					UpdateMarkerList(form, -1);
					break;
				default:
					handled = false;
					break;
			}
			break;
		case menuEvent:
			switch(event->data.menu.itemID)
			{
				case SendMarkerID:
					MarkerBeam(MarkerGetViewIndex());
					handled = true;
					break;
				case SendAllMarkerID:
					MarkerBeamList();
					handled = true;
					break;
				case SendMarkerDBID:
					MarkerBeamDB();
					handled = true;
					break;
				case SendMarkerClipID:
					MarkerToClipboard(MarkerGetViewIndex());
					handled = true;
					break;
				case DestSortAlphaID:
					if(MarkerQSort() == true)
						UpdateMarkerList(form, MarkerGetViewIndex());
					break;
				case DestSortDistID:
					if(has_coord || has_ref)
					{
						Boolean ret;
						//UInt32 t1, t2;

						SetTitleRes(form, TitleCALC);
						ToggleRef(form);
						//t1 = TimGetTicks();
						ret = MarkerDSort(cur_lon, cur_lat);
						//t2 = TimGetTicks();
						//TRACE("Ticks: %ld", (t2-t1));
						SetTitleRes(form, TitleMARKER);
						if(ret == true)
							UpdateMarkerList(form, 0);
					}
					break;
			}
			break;
		case ctlSelectEvent:
			switch(event->data.ctlSelect.controlID)
			{
				case DestDetailsID:
					MarkerEdit(MarkerGetViewIndex(), true);
					handled = true;
					break;
				case DestNewID:
					MarkerEdit(-1, true);
					handled = true;
					break;
				case DestUseID:
					MarkerSetDestination(MarkerGetViewIndex());
					MarkerSetViewIndex(-1);
					UtilReturnToLastForm();
					handled = true;
					break;
				case DestNonID:
					MarkerClearDestination();
					MarkerSetViewIndex(-1);
					UtilReturnToLastForm();
					handled = true;
					break;
				case DestKeepID:
					MarkerSetViewIndex(-1);
					UtilReturnToLastForm();
					handled = true;
					break;
				case DestCatPopupTriggerID:
					SelectCategory(form);
					break;
				case DestRefID:
					ToggleRef(form);
					handled = true;
					break;
			}
		case lstSelectEvent:
			if(event->data.lstSelect.listID == DestListID)
			{
				if(list_indexes != NULL)
					UpdateFields(form, list_indexes[event->data.lstSelect.selection]);
				handled = false;
			}
			break;
		case ctlRepeatEvent:
			switch(event->data.ctlRepeat.controlID)
			{
				case DestMoveUpID:
						MoveMarker(form, event->data.ctlRepeat.controlID);
					break;
				case DestMoveDownID:
						MoveMarker(form, event->data.ctlRepeat.controlID);
					break;
				default:
					break;
			}
			break;
		case winDisplayChangedEvent:
		case displayExtentChangedEvent:
			if(form_ready == true)
			{
				SizeForm(form);
				FrmDrawForm(form);
			}
			handled = true;
			break;
		case keyDownEvent:
			if(UtilIsHardBtn(event))
			{
				if(UtilIsArrowBtn(event))
				{
					ButtonHandler(form, UtilBtnDirection(event));
					handled = true;
				}
			}
			else
			{
				handled = LookupString(form, event);
			}
			break;
		//case fldChangedEvent:
		//	handled = LookupString(form, event);
		//	break;
		case GpsNmeaEvent:
			GPSHandler(form, CASTNMEA(event));
			break;
		default:
			break;
	}
	return(handled);
}

static void ToggleRef(FormPtr form)
{
	has_ref = !has_ref;
	if(has_ref)
	{
		refindex = MarkerGetViewIndex();
		reflistindex = LstGetSelection(UtilGetObjectPtr(form, DestListID));
		if(reflistindex != noListSelection)
		{
			RectangleType rlist, rbut;
			Coord lw;
			marker_type marker;
			Int16 slen;
			Boolean fits;
			Char name[MAX_MARKER_NAME_LENGTH];

// TODO berechnen, wieviel chars in den Platz passen und nur soviel rberkopieren!
			FrmGetObjectBounds(form,
				FrmGetObjectIndex(form, DestListID), &rlist);
			FrmGetObjectBounds(form,
				FrmGetObjectIndex(form, DestRefID), &rbut);
			lw = (rlist.topLeft.x+rlist.extent.x) -
					(rbut.topLeft.x+rbut.extent.x);
			MarkerGetMarker(refindex, &marker);
			slen = StrLen(marker.name);
			FntCharsInWidth(marker.name, &lw, &slen, &fits);
			MemSet(name, MAX_MARKER_NAME_LENGTH, chrNull);
			StrNCopy(name, marker.name, slen);
			UtilUpdateString(form, DestRefLabelID, name);
			cur_lat = marker.lat;
			cur_lon = marker.lon;
		}
		else
			has_ref = false;
	}
	else
	{
		UtilUpdateString(form, DestRefLabelID, "GPS");
	}
	UtilSetControlValue(form, DestRefID, has_ref);
	LstDrawList(UtilGetObjectPtr(form, DestListID));
	UpdateData(form);
}

static void MoveMarker(FormPtr form, UInt16 id)
{
	ListType *list;
	UInt16 from, to, sel;

	list = UtilGetObjectPtr(form, DestListID);
	sel = LstGetSelection(list);

	if(sel == noListSelection)
		return;

	to = from = list_indexes[sel];

	if(id == DestMoveUpID)
	{
		if(sel == 0)
			return;
		to = list_indexes[sel-1];
		sel = to;
	}
	else if(id == DestMoveDownID)
	{
		if((sel+1) == LstGetNumberOfItems(list))
			return;
		to = list_indexes[sel+1]+1;
		sel = to-1;
	}

	if(to != from)
		if(MarkerMoveMarker(from, to) == true)
			UpdateMarkerList(form, sel);
}

static void UpdateMarkerList(FormPtr form, Int16 index)
{
	ListType *list;
	UInt16 i, j, category;
	DmOpenRef db;

	category	= PrefsGetPrefs()->marker_category;
	list		= UtilGetObjectPtr(form, DestListID);
	list_cnt	= MarkerGetMarkerCountInCategory(category);
	j			= 0;

	if(list_cnt > 0)
	{
		if(list_cnt > list_indexes_cnt)
		{
			FreeListIndexes();
			/* make room for a little more */
			list_indexes_cnt = list_cnt + 5;
			list_indexes = (UInt16*)UtilMalloc(list_indexes_cnt * sizeof(UInt16));
		}
		db = MarkerOpenDB(dmModeReadOnly);
		if(db != NULL)
		{
			UInt16 catind = 0;
			for(i = 0, j = 0; i < list_cnt; i++)
			{
				MemHandle h = DmQueryNextInCategory(db, &catind, category);
				if(h != NULL)
				{
					if(filter_len > 0)
					{
						marker_type *dbmarker;
						Char s1[MAX_MARKER_NAME_LENGTH];

						dbmarker = (marker_type *)MemHandleLock(h);
						StrNCopy(s1, dbmarker->name, MAX_MARKER_NAME_LENGTH);
						StrToLower(s1, s1);
						if(StrStr(s1, filter) != NULL)
							list_indexes[j++] = catind;
						MemHandleUnlock(h);
					}
					else
						list_indexes[j++] = catind;
				}
				catind++;
			}
			MarkerCloseDB(db);
			list_cnt = j;
		}
	}
	else
		index = -1;
	LstSetListChoices(list, NULL, list_cnt);
	LstDrawList(list);

	if(list_cnt > 0)
	{
		Int16 lindex = -1;

		if(index == -1)
			index = PrefsGetPrefs()->marker_idx;

		if(index+1 > MarkerGetMarkerCount())
			index = list_indexes[0];
		else
			for(i = 0; i < list_cnt; i++)
				if(list_indexes[i] == index)
					lindex = i;
		LstSetSelection(list, lindex);
	}
	else
		LstSetSelection(list, noListSelection);

	UpdateFields(form, index);
}

static void UpdateFields(FormPtr form, Int16 index)
{
	marker_type tm;

	if(MarkerGetMarker(index, &tm) == true)
	{
		MarkerSetViewIndex(index);
		dist.lat = tm.lat;
		dist.lon = tm.lon;

		PrefsUpdateLabelCoord(form, DestLatID, DestLonID, dist.lat, dist.lon);
		UpdateData(form);
	}
	else
	{
		MarkerSetViewIndex(-1);
		UtilUpdateChar(form, DestLatID, '-');
		UtilUpdateChar(form, DestLonID, '-');
		UtilUpdateChar(form, DestCourseID, '-');
		UtilUpdateChar(form, DestDistID, '-');
	}
	UpdateCategory(form, index);
}

static void UpdateCategory(FormPtr form, Int16 index)
{
	DmOpenRef db = MarkerOpenDB(dmModeReadOnly);
	if(db != NULL)
	{
		ControlType *ctl;
		UInt16 category;
		CAppPrefs *prefs;
		static Char category_name[dmCategoryLength];

		prefs = PrefsGetPrefs();

		/* if current category is All, we need to look up category */
		if((index >= 0) && (prefs->marker_category == dmAllCategories))
		{
			UInt16 attr;

			DmRecordInfo(db, index, &attr, NULL, NULL);
			category = attr & dmRecAttrCategoryMask;
		}
		else
			category = prefs->marker_category;
		CategoryGetName(db, category, category_name);
		ctl = UtilGetObjectPtr(form, DestCatPopupTriggerID);
		CategorySetTriggerLabel(ctl, category_name);
		MarkerCloseDB(db);
	}
}

static void UpdateData(FormPtr form)
{
	if((MarkerGetViewIndex() != -1) && (has_coord || has_ref))
	{
		double ddist;
		Int32  course;

		ddist = DistGetDist(&dist, cur_lat, cur_lon);
		PrefsNauticToPref(&ddist, NULL);
		UtilUpdateDouble(form, DestDistID, ddist/1000., 3);

		course = DistGetCourseDeg(&dist, cur_lat, cur_lon);
		UtilUpdateSingle(form, DestCourseID, course, "");
	}
	else
	{
		UtilUpdateChar(form, DestCourseID, '-');
		UtilUpdateChar(form, DestDistID, '-');
	}
}

static void ButtonHandler(FormPtr form, WChar direction)
{
	ListType *list;
	Int16 vis;

	list = UtilGetObjectPtr(form, DestListID);
	vis = LstGetVisibleItems(list);
	LstScrollList(list, (direction == vchrPageUp ? winUp : winDown), vis-1);
}

static void GPSHandler(FormPtr form, nmea_type *nmea)
{
	if(has_ref)
		return;

	if(nmea->fix_quality > 0)
	{
		has_coord = true;
		cur_lat = nmea->lat;
		cur_lon = nmea->lon;
	}
	else
		has_coord = false;

	if(nmea->is_new_gga == true)
	{
		UpdateData(form);
		nmea->is_new_gga = false;
	}
}

static void SizeForm(FormPtr form)
{
	RectangleType curBounds, displayBounds;
	Coord diff;

	UtilGetFrmBounds(form, &curBounds);
	UtilGetDpyBounds(&displayBounds);

	diff = displayBounds.extent.y - (curBounds.topLeft.y + curBounds.extent.y);

	if(diff)
	{
		Int16 fh;
		Coord ldiff;
		FontID old_font;

		curBounds.extent.y += diff;
		WinSetBounds(FrmGetWindowHandle(form), &curBounds);

		if(UtilIsVGA())
			old_font = FntSetFont(VgaBaseToVgaFont(stdFont));
		else
			old_font = FntSetFont(stdFont);
		fh = FntCharHeight();
		FntSetFont(old_font);

		ldiff = fh*((Int16)(diff/fh));

		UtilMoveObject(form, DestNewID, diff, true);
		UtilMoveObject(form, DestDetailsID, diff, true);
		UtilMoveObject(form, DestUseID, diff, true);
		UtilMoveObject(form, DestNonID, diff, true);
		UtilMoveObject(form, DestKeepID, diff, true);
		UtilMoveObject(form, DestMoveDownID, diff, true);
		UtilMoveObject(form, DestMoveUpID, diff, true);
		UtilMoveObject(form, DestLookUpLabelID, diff, true);
		UtilMoveObject(form, DestLookUpID, diff, true);
		UtilResizeObject(form, DestListID, ldiff, true);
	}
}

static void SelectCategory(FormPtr form)
{
	DmOpenRef db;

	if(has_ref)
		ToggleRef(form);

	db = MarkerOpenDB(dmModeReadOnly);
	if(db != NULL)
	{
		static Char category_name[dmCategoryLength];
		Boolean catedited;
		UInt16 category;
		CAppPrefs *prefs;

		prefs = PrefsGetPrefs();

		category = prefs->marker_category;

		catedited = CategorySelect(db, form,
					DestCatPopupTriggerID, DestCatListID,
					true, &category, category_name, 1,
					categoryDefaultEditCategoryString);
		MarkerCloseDB(db);

		if(catedited || (category != prefs->marker_category))
		{
			prefs->marker_category = category;
			UpdateMarkerList(form, -1);
		}
	}
}

static void ListDrawFunc(Int16 index, RectangleType *bounds, Char **data)
{
	Coord w, x, y;
	Int16 slen;
	Boolean fits;
	marker_type marker;
	marker_note_type note;
	UInt16 list_index;

	if(list_indexes == NULL)
		return;

	list_index = list_indexes[index];
	if(MarkerGetMarker(list_index, &marker) == false)
	{
		marker.lat = 0.;
		marker.lon = 0.;
		*marker.name = '\0';
	}

	x    = bounds->topLeft.x;
	y    = bounds->topLeft.y;
	w    = bounds->extent.x;
	slen = StrLen(marker.name);

	FntCharsInWidth(marker.name, &w, &slen, &fits);
	WinDrawChars(marker.name, slen, x, y);

	note.index = list_index;
	note.mode  = dmModeReadOnly;
	if(MarkerGetMarkerDescription(&note))
	{
		if(note.size > 1)
		{
			FontID old_font;
			Int16 cw;
			Char notech = 0x09;

			if(UtilIsVGA())
				old_font = FntSetFont(VgaBaseToVgaFont(symbolFont));
			else
				old_font = FntSetFont(symbolFont);
			cw = FntCharsWidth(&notech, 1);
			WinDrawChars(&notech, 1, bounds->extent.x-cw, y);
			FntSetFont(old_font);
		}
		MarkerUpdateMarkerDescription(&note, false);
	}

	if(has_ref && (index == reflistindex))
	{
		FontID old_font;
		Int16 cw;
		Char isref = 0x8F;

		if(UtilIsVGA())
			old_font = FntSetFont(VgaBaseToVgaFont(boldFont));
		else
			old_font = FntSetFont(boldFont);
		cw = FntCharsWidth(&isref, 1);
		WinDrawChars(&isref, 1, bounds->extent.x-cw, y);
		FntSetFont(old_font);
	}
}

static void FreeListIndexes()
{
	if(list_indexes != NULL)
	{
		UtilFree(list_indexes);
		list_indexes = NULL;
	}
	list_indexes_cnt = 0;
}

static Boolean LookupString(FormPtr form, EventPtr event)
{
	Char *lus;

	FldHandleEvent(UtilGetObjectPtr(form, DestLookUpID), event);
	lus = UtilGetString(form, DestLookUpID);
	StrNCopy(filter, lus, MAX_MARKER_NAME_LENGTH);
	StrToLower(filter, filter);
	filter_len = StrLen(filter);

	UpdateMarkerList(form, -1);
	return(true);
}

