/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include "main.h"
#include "forms.h"
#include "util.h"
#include "serial.h"
#include "marker.h"
#include "prefs.h"
#include "resource.h"
#include "def.h"

static void GPSHandler(FormPtr form, nmea_type *nmea) CODE2_SECTION;
static void SizeForm(FormPtr form) CODE2_SECTION;
static void SelectCategory(FormPtr form) CODE2_SECTION;
static void CryptSelection(FormPtr form) CODE2_SECTION;
static void SetCategory() CODE2_SECTION;
static void ButtonHandler(FormPtr form, WChar direction) CODE2_SECTION;
static void UpdateFields(FormPtr form, Int16 marker_index) CODE2_SECTION;
static void UpdateScrollbar(FormPtr form, FieldType *field) CODE2_SECTION;
static void UpdateNote(FormPtr form, Boolean update) CODE2_SECTION;
static void UpdateLatLon(FormPtr form, double lon, double lat) CODE2_SECTION;
static void UpdateMarker(FormPtr form) CODE2_SECTION;
static Boolean ReadLatLon(FormPtr form, double *lat, double *lon) CODE2_SECTION;

static double curr_lat, curr_lon;
static Int16 marker_index;
static UInt16 category;
static Boolean is_new;
static marker_note_type note;

Boolean MarkerEditFormHandler(EventPtr event)
{
	FormPtr	form;
	Boolean handled = false;
	CAppPrefs *prefs;
	Err err;

	form = FrmGetActiveForm();
	prefs = PrefsGetPrefs();

	switch(event->eType)
	{
		case frmOpenEvent:
				curr_lat = 0.;
				curr_lon = 0.;
				is_new = false;
				if(UtilIsVGA())
					VgaFormModify(form, vgaFormModify160To240);
				if(UtilIsResizeableNoDia())
					SizeForm(form);
				FrmDrawForm(form);
				UtilShowObject(form, DestEditScrollBarID, true);
				UpdateFields(form, MarkerGetEditIndex());
				handled = true;
			break;
		case frmUpdateEvent:
				FrmDrawForm(form);
				handled = true;
			break;
		case frmCloseEvent:
				if(note.db != NULL)
					UpdateNote(form, false);
				if(is_new)
					MarkerDelMarker(marker_index, &err);
			break;
		case menuEvent:
			switch(event->data.menu.itemID)
			{
				case MarkerEditMenuDD:
						FrmGotoForm(MarkerEditFormDD);
						handled = true;
					break;
				case MarkerEditMenuDMM:
						FrmGotoForm(MarkerEditFormDMM);
						handled = true;
					break;
				case MarkerEditMenuDMS:
						FrmGotoForm(MarkerEditFormDMS);
						handled = true;
					break;
				case MarkerEditMenuDMSS:
						FrmGotoForm(MarkerEditFormDMSS);
						handled = true;
					break;
				case MarkerEditMenuUTM:
						FrmGotoForm(MarkerEditFormUTM);
						handled = true;
					break;
				case MarkerEditMenuOSGB:
						FrmGotoForm(MarkerEditFormOSGB);
						handled = true;
					break;
				case MarkerEditMenuQTH:
						FrmGotoForm(MarkerEditFormQTH);
						handled = true;
					break;
				case MarkerEditCrypt:
						CryptSelection(form);
						handled = true;
					break;
			}
			break;
		case winDisplayChangedEvent:
		case displayExtentChangedEvent:
				SizeForm(form);
				FrmDrawForm(form);
				handled = true;
			break;
		case sclRepeatEvent:
				UtilScrollField(UtilGetObjectPtr(form, DestEditFieldID),
							UtilGetObjectPtr(form, DestEditScrollBarID),
							(event->data.sclRepeat.newValue - 
								event->data.sclRepeat.value));
			break;
		case fldChangedEvent:
				UpdateScrollbar(form, UtilGetObjectPtr(form, DestEditFieldID));
			break;
		case keyDownEvent:
			if(UtilIsHardBtn(event) && UtilIsArrowBtn(event))
			{
				ButtonHandler(form, UtilBtnDirection(event));
				handled = true;
			}
			else switch(event->data.ctlSelect.controlID)
			{
				case pageUpChr:
						UtilPageScroll(UtilGetObjectPtr(form, DestEditFieldID), 
							UtilGetObjectPtr(form, DestEditScrollBarID),
							winUp);
						handled = true;
					break;
				case pageDownChr:
						UtilPageScroll(UtilGetObjectPtr(form, DestEditFieldID), 
							UtilGetObjectPtr(form, DestEditScrollBarID),
							winDown);
						handled = true;
					break;
				default:
						handled = false;
			}
			break;
		case ctlSelectEvent:
				switch(event->data.ctlSelect.controlID)
				{
					case DestEditOKID:
						UpdateNote(form, true);
						UpdateMarker(form);
						UtilReturnToLastForm();
						handled = true;
						break;
					case DestEditCancelID:
						UtilReturnToLastForm();
						handled = true;
						break;
					case DestEditHereID:
						UpdateLatLon(form, curr_lon, curr_lat);
						handled = true;
						break;
					case DestEditDeleteID:
						//TODO ask "really?"
						// this deletes marker at frmClose
						is_new = true;
						UtilReturnToLastForm();
						handled = true;
						break;
					case DestCatPopupTriggerID:
						SelectCategory(form);
						break;
				}
			break;
		case GpsNmeaEvent:
				GPSHandler(form, CASTNMEA(event));
			break;
		default:
			break;
	}
	return(handled);
}

static void ButtonHandler(FormPtr form, WChar direction)
{
	UtilPageScroll(UtilGetObjectPtr(form, DestEditFieldID), 
		UtilGetObjectPtr(form, DestEditScrollBarID),
		(direction == vchrPageUp ? winUp : winDown));
}

static void GPSHandler(FormPtr form, nmea_type *nmea)
{
	if((nmea->fix_quality > 0) && (nmea->is_new_gga || nmea->is_new_nexrmc))
	{
		curr_lon = nmea->lon;
		curr_lat = nmea->lat;
		nmea->is_new_gga = false;
		nmea->is_new_nexrmc = false;
	}
}

static void SizeForm(FormPtr form)
{
	RectangleType curBounds, displayBounds;
	Coord diff;

	UtilGetFrmBounds(form, &curBounds);
	UtilGetDpyBounds(&displayBounds);

	diff = displayBounds.extent.y - (curBounds.topLeft.y + curBounds.extent.y);

	if(diff)
	{
		FieldType *field;

		curBounds.extent.y += diff;
		WinSetBounds(FrmGetWindowHandle(form), &curBounds);

		UtilResizeObject(form, DestEditFieldID, diff, true);
		UtilResizeObject(form, DestEditScrollBarID, diff, true);
		UtilMoveObject(form, DestEditOKID, diff, true);
		UtilMoveObject(form, DestEditCancelID, diff, true);
		UtilMoveObject(form, DestEditDeleteID, diff, true);
		UtilMoveObject(form, DestEditHereID, diff, true);

		field = (FieldType*)UtilGetObjectPtr(form, DestEditFieldID);
		UpdateScrollbar(form, field);
		FldRecalculateField(field, true);
	}
}

static void CryptSelection(FormPtr form)
{
	FieldType *field;
	UInt16 start, end;

	field = (FieldType*)UtilGetObjectPtr(form, DestEditFieldID);
	FldGetSelection(field, &start, &end);
	if(start != end)
	{
		UInt16 length, i;
		Char *text;
		MemHandle mh;

		FldCopy(field);
		mh = ClipboardGetItem(clipboardText, &length);
		text = (Char*)MemHandleLock(mh);
		for(i = 0; i < length; i++)
		{
			if((text[i] >= 'A') && (text[i] <= 'M'))
			{
				text[i] += 13;
			}
			else if((text[i] >= 'M') && (text[i] <= 'Z'))
			{
				text[i] -= 13;
			}
			else if((text[i] >= 'a') && (text[i] <= 'm'))
			{
				text[i] += 13;
			}
			else if((text[i] >= 'm') && (text[i] <= 'z'))
			{
				text[i] -= 13;
			}
		}
		MemHandleUnlock(mh);
		FldPaste(field);
		FldSetSelection(field, start, end);
	}
}

static void SelectCategory(FormPtr form)
{
	if((note.db != NULL) && (note.mode == dmModeReadWrite))
	{
		Boolean catedited;
		UInt16 ncat;
		static Char category_name[dmCategoryLength];

		ncat = category;

		catedited = CategorySelect(note.db, form,
					DestCatPopupTriggerID, DestCatListID,
					false, &ncat, category_name, 1,
					categoryHideEditCategory);

		if(catedited || (category != ncat))
		{
			category = ncat;
			SetCategory();
		}
	}
}

static void SetCategory()
{
	UInt16 attr;

	// Set it's category to the category being viewed.
	// If the category is All then set the category to unfiled.
	DmRecordInfo(note.db, note.tindex, &attr, NULL, NULL);

	attr &= ~dmRecAttrCategoryMask;
	attr |= ((category == dmAllCategories) ? dmUnfiledCategory : category) | dmRecAttrDirty;
	DmSetRecordInfo(note.db, note.tindex, &attr, NULL);
}

static void UpdateFields(FormPtr form, Int16 index)
{
	marker_type marker;
	ControlType *ctl;
	UInt16 attr;
	static Char category_name[dmCategoryLength];

	marker_index = index;
	is_new = false;
	if(marker_index == -1)
	{
		UInt32 cnt;
		Char buf[64];
		Err err;

		cnt = MarkerGetMarkerCount();
		StrPrintF(buf, "Marker %lu", cnt);
		marker_index = MarkerAddMarker(buf, 0., 0., &err);
		if(err != errNone)
			return;
		is_new = true;
	}

	MarkerGetMarker(marker_index, &marker);
	UpdateLatLon(form, marker.lon, marker.lat);
	UtilSetFieldText(form, DestEditNameID, marker.name);

	note.index = marker_index;
	note.mode  = dmModeReadWrite;
	MarkerGetMarkerDescription(&note);
	if(note.size != 0)
	{
		FieldType *field;

		field = (FieldType*)UtilGetObjectPtr(form, DestEditFieldID);
		FldSetText(field, note.mhandle, note.offset, note.size);
		UpdateScrollbar(form, field);
		FldDrawField(field);
	}

	if((note.size == 0) || (note.mode == dmModeReadOnly))
	{
		UtilSetEnabled(form, DestEditNameID, false);
	}

	if(is_new)
		UtilSelectField(form, DestEditNameID);

	/* select category */
	DmRecordInfo(note.db, note.index, &attr, NULL, NULL);
	category = attr & dmRecAttrCategoryMask;
	CategoryGetName(note.db, category, category_name);
	ctl = UtilGetObjectPtr(form, DestCatPopupTriggerID);
	CategorySetTriggerLabel(ctl, category_name);
	SetCategory();

	if(MarkerIsEditable() == false)
	{
		UtilShowObject(form, DestEditOKID, false);
		UtilShowObject(form, DestEditDeleteID, false);
		UtilShowObject(form, DestEditHereID, false);
	}
}

static void UpdateScrollbar(FormPtr form, FieldType *field)
{
	UInt16 scroll_pos, text_h, field_h;
	Int16 max;
	ScrollBarType *sb;

	sb = (ScrollBarType*)UtilGetObjectPtr(form, DestEditScrollBarID);

	FldGetScrollValues(field, &scroll_pos, &text_h, &field_h);
	if(text_h > field_h)
		max = text_h - field_h + FldGetNumberOfBlankLines(field);
	//else if(scroll_pos)
	//	max = scroll_pos;
	else
		max = 0;
	SclSetScrollBar(sb, scroll_pos, 0, max, field_h-1);
}

static void UpdateNote(FormPtr form, Boolean update)
{
	FieldType *field = (FieldType*)UtilGetObjectPtr(form, DestEditFieldID);

	if(note.size != 0)
	{
		FldCompactText(field);
		FldSetText(field, NULL, 0, 0);
	}
	else
		update = false;

	MarkerUpdateMarkerDescription(&note, update);
}

static void UpdateMarker(FormPtr form)
{
	double lat, lon;

	if(ReadLatLon(form, &lat, &lon))
	{
		Err err;
		Char *name, *note;

		name = UtilGetString(form, DestEditNameID);
		note = UtilGetString(form, DestEditFieldID);
		MarkerUpdateMarker(marker_index, name, lon, lat, &err);
		is_new = false;
	}
	else
		SndPlaySystemSound(sndError);
}

static void UpdateLatLon(FormPtr form, double lon, double lat)
{
	Int32 deg1, min1, sec1, frac1;
	Int32 deg2, min2, sec2, frac2;
	UInt16 fid, dec;
	Char c1, c2;
	Char zone[4];
	Char buf[16];
	Char fbuf[8];

	MemSet(zone, 4, 0);
	MemSet(buf, 16, 0);
	MemSet(fbuf, 8, 0);

	fid = FrmGetFormId(form);
	dec = PrefsGetDec();

	switch(fid)
	{
		case MarkerEditFormDD:
			NMEAGetDeg_DD(lon, 6, &deg1, &min1, &c1, LATLON_LON);
			NMEAGetDeg_DD(lat, 6, &deg2, &min2, &c2, LATLON_LAT);

			StrPrintF(buf, "%ld", deg1);
			UtilSetFieldText(form, DestEditLonDegID, buf);
			StrPrintF(buf, "%ld", deg2);
			UtilSetFieldText(form, DestEditLatDegID, buf);
			StrPrintF(buf, "%06ld", min1);
			UtilSetFieldText(form, DestEditLonDecID, buf);
			StrPrintF(buf, "%06ld", min2);
			UtilSetFieldText(form, DestEditLatDecID, buf);

			UtilSetListSelection(form, DestEditLonListID, DestEditLonEWID,
							(c1 == 'W' ? 0 : 1));
			UtilSetListSelection(form, DestEditLatListID, DestEditLatNSID,
							(c2 == 'N' ? 0 : 1));
			break;
		case MarkerEditFormDMM:
			NMEAGetDeg_DMM(lon, dec, &deg1, &min1, &frac1, &c1, LATLON_LON);
			NMEAGetDeg_DMM(lat, dec, &deg2, &min2, &frac2, &c2, LATLON_LAT);

			StrPrintF(buf, "%ld", deg1);
			UtilSetFieldText(form, DestEditLonDegID, buf);
			StrPrintF(buf, "%ld", min1);
			UtilSetFieldText(form, DestEditLonMinID, buf);
			StrPrintF(fbuf, "%%0%uld", dec);
			StrPrintF(buf, fbuf, frac1);
			UtilSetFieldText(form, DestEditLonDecID, buf);

			StrPrintF(buf, "%ld", deg2);
			UtilSetFieldText(form, DestEditLatDegID,  buf);
			StrPrintF(buf, "%ld", min2);
			UtilSetFieldText(form, DestEditLatMinID,  buf);
			StrPrintF(fbuf, "%%0%uld", dec);
			StrPrintF(buf, fbuf, frac2);
			UtilSetFieldText(form, DestEditLatDecID,  buf);

			UtilSetListSelection(form, DestEditLonListID, DestEditLonEWID,
							(c1 == 'W' ? 0 : 1));
			UtilSetListSelection(form, DestEditLatListID, DestEditLatNSID,
							(c2 == 'N' ? 0 : 1));
			break;
		case MarkerEditFormDMS:
			NMEAGetDeg_DMS(lon, &deg1, &min1, &sec1, &c1, LATLON_LON);
			NMEAGetDeg_DMS(lat, &deg2, &min2, &sec2, &c2, LATLON_LAT);

			StrPrintF(buf, "%ld", deg1);
			UtilSetFieldText(form, DestEditLonDegID, buf);
			StrPrintF(buf, "%ld", min1);
			UtilSetFieldText(form, DestEditLonMinID, buf);
			StrPrintF(buf, "%ld", sec1);
			UtilSetFieldText(form, DestEditLonSecID, buf);
			StrPrintF(buf, "%ld", deg2);
			UtilSetFieldText(form, DestEditLatDegID, buf);
			StrPrintF(buf, "%ld", min2);
			UtilSetFieldText(form, DestEditLatMinID, buf);
			StrPrintF(buf, "%ld", sec2);
			UtilSetFieldText(form, DestEditLatSecID, buf);

			UtilSetListSelection(form, DestEditLonListID, DestEditLonEWID,
							(c1 == 'W' ? 0 : 1));
			UtilSetListSelection(form, DestEditLatListID, DestEditLatNSID,
							(c2 == 'N' ? 0 : 1));
			break;
		case MarkerEditFormDMSS:
			NMEAGetDeg_DMSS(lon, 2, &deg1, &min1, &sec1, &frac1, &c1, LATLON_LON);
			NMEAGetDeg_DMSS(lat, 2, &deg2, &min2, &sec2, &frac2, &c2, LATLON_LAT);

			StrPrintF(buf, "%ld", deg1);
			UtilSetFieldText(form, DestEditLonDegID, buf);
			StrPrintF(buf, "%ld", deg2);
			UtilSetFieldText(form, DestEditLatDegID, buf);
			StrPrintF(buf, "%ld", min1);
			UtilSetFieldText(form, DestEditLonMinID, buf);
			StrPrintF(buf, "%ld", min2);
			UtilSetFieldText(form, DestEditLatMinID, buf);
			StrPrintF(buf, "%ld", sec1);
			UtilSetFieldText(form, DestEditLonSecID, buf);
			StrPrintF(buf, "%ld", sec2);
			UtilSetFieldText(form, DestEditLatSecID, buf);
			StrPrintF(buf, "%02ld", frac1);
			UtilSetFieldText(form, DestEditLonDecID, buf);
			StrPrintF(buf, "%02ld", frac2);
			UtilSetFieldText(form, DestEditLatDecID, buf);

			UtilSetListSelection(form, DestEditLonListID, DestEditLonEWID,
							(c1 == 'W' ? 0 : 1));
			UtilSetListSelection(form, DestEditLatListID, DestEditLatNSID,
							(c2 == 'N' ? 0 : 1));
			break;
		case MarkerEditFormUTM:
			// deg1 -> northing, deg2 -> easting
			NMEAGetDeg_UTM(lat, lon, &deg1, &deg2, zone);
			StrPrintF(buf, "%ld", deg1);
			UtilSetFieldText(form, DestEditLatDegID, buf);
			StrPrintF(buf, "%ld", deg2);
			UtilSetFieldText(form, DestEditLonDegID, buf);
			UtilSetFieldText(form, DestEditLatDecID, zone);
			break;
		case MarkerEditFormOSGB:
			// deg1 -> northing, deg2 -> easting, zone->gridsquare
			NMEAGetDeg_OSGB(lat, lon, &deg1, &deg2, zone);
			StrPrintF(buf, "%ld", deg1);
			UtilSetFieldText(form, DestEditLatDegID, buf);
			StrPrintF(buf, "%ld", deg2);
			UtilSetFieldText(form, DestEditLonDegID, buf);
			UtilSetFieldText(form, DestEditLatDecID, zone);
			break;
		case MarkerEditFormQTH:
			NMEAGetDeg_QTH(lat, lon, buf);
			UtilSetFieldText(form, DestEditLatDegID, buf);
			break;
	}
}

static Boolean ReadLatLon(FormPtr form, double *lat, double *lon)
{
	ListType *list;
	Int32 deg1, min1, sec1, frac1;
	Int32 deg2, min2, sec2, frac2;
	Int16 dec1, dec2;
	Char c1, c2;
	Char sbuf[7];
	UInt16 fid;

	fid = FrmGetFormId(form);

	*lon = 0.;
	*lat = 0.;

	switch(fid)
	{
		case MarkerEditFormDD:
			UtilGetInt(form, DestEditLonDegID, &deg1, NULL);
			UtilGetInt(form, DestEditLatDegID, &deg2, NULL);
			UtilGetInt(form, DestEditLonDecID, &min1, &dec1);
			UtilGetInt(form, DestEditLatDecID, &min2, &dec2);

			list = UtilGetObjectPtr(form, DestEditLonListID);
			c1 = (LstGetSelectionText(list, LstGetSelection(list)))[0];
			list = UtilGetObjectPtr(form, DestEditLatListID);
			c2 = (LstGetSelectionText(list, LstGetSelection(list)))[0];

			*lon = NMEAGetRadianDD(deg1, min1, dec1, c1);
			*lat = NMEAGetRadianDD(deg2, min2, dec2, c2);
			break;
		case MarkerEditFormDMM:
			UtilGetInt(form, DestEditLonDegID, &deg1, NULL);
			UtilGetInt(form, DestEditLatDegID, &deg2, NULL);
			UtilGetInt(form, DestEditLonMinID, &min1, NULL);
			UtilGetInt(form, DestEditLatMinID, &min2, NULL);
			UtilGetInt(form, DestEditLonDecID, &frac1, &dec1);
			UtilGetInt(form, DestEditLatDecID, &frac2, &dec2);

			list = UtilGetObjectPtr(form, DestEditLonListID);
			c1 = (LstGetSelectionText(list, LstGetSelection(list)))[0];
			list = UtilGetObjectPtr(form, DestEditLatListID);
			c2 = (LstGetSelectionText(list, LstGetSelection(list)))[0];

			*lon = NMEAGetRadianDMM(deg1, min1, frac1, dec1, c1);
			*lat = NMEAGetRadianDMM(deg2, min2, frac2, dec2, c2);
			break;
		case MarkerEditFormDMS:
			UtilGetInt(form, DestEditLonDegID, &deg1, NULL);
			UtilGetInt(form, DestEditLatDegID, &deg2, NULL);
			UtilGetInt(form, DestEditLonMinID, &min1, NULL);
			UtilGetInt(form, DestEditLatMinID, &min2, NULL);
			UtilGetInt(form, DestEditLonSecID, &sec1, NULL);
			UtilGetInt(form, DestEditLatSecID, &sec2, NULL);

			list = UtilGetObjectPtr(form, DestEditLonListID);
			c1 = (LstGetSelectionText(list, LstGetSelection(list)))[0];
			list = UtilGetObjectPtr(form, DestEditLatListID);
			c2 = (LstGetSelectionText(list, LstGetSelection(list)))[0];

			*lon = NMEAGetRadianDMS(deg1, min1, sec1, c1);
			*lat = NMEAGetRadianDMS(deg2, min2, sec2, c2);
			break;
		case MarkerEditFormDMSS:
			UtilGetInt(form, DestEditLonDegID, &deg1, NULL);
			UtilGetInt(form, DestEditLatDegID, &deg2, NULL);
			UtilGetInt(form, DestEditLonMinID, &min1, NULL);
			UtilGetInt(form, DestEditLatMinID, &min2, NULL);
			UtilGetInt(form, DestEditLonSecID, &sec1, NULL);
			UtilGetInt(form, DestEditLatSecID, &sec2, NULL);
			UtilGetInt(form, DestEditLonDecID, &frac1, &dec1);
			UtilGetInt(form, DestEditLatDecID, &frac2, &dec2);

			list = UtilGetObjectPtr(form, DestEditLonListID);
			c1 = (LstGetSelectionText(list, LstGetSelection(list)))[0];
			list = UtilGetObjectPtr(form, DestEditLatListID);
			c2 = (LstGetSelectionText(list, LstGetSelection(list)))[0];

			*lon = NMEAGetRadianDMSS(deg1, min1, sec1, frac1, dec1, c1);
			*lat = NMEAGetRadianDMSS(deg2, min2, sec2, frac2, dec2, c2);
			break;
		case MarkerEditFormUTM:
			// deg1 -> northing, deg2 -> easting
			UtilGetInt(form, DestEditLatDegID, &deg1, NULL);
			UtilGetInt(form, DestEditLonDegID, &deg2, NULL);
			StrCopy(sbuf, UtilGetString(form, DestEditLatDecID));

			NMEAGetRadianUTM(deg1, deg2, sbuf, lat, lon);
			break;
		case MarkerEditFormOSGB:
			// deg1 -> northing, deg2 -> easting
			UtilGetInt(form, DestEditLatDegID, &deg1, NULL);
			UtilGetInt(form, DestEditLonDegID, &deg2, NULL);
			StrCopy(sbuf, UtilGetString(form, DestEditLatDecID));

			NMEAGetRadianOSGB(deg1, deg2, sbuf, lat, lon);
			break;
		case MarkerEditFormQTH:
			StrCopy(sbuf, UtilGetString(form, DestEditLatDegID));
			NMEAGetRadianQTH(sbuf, lat, lon);
	}
	return(true);
}

