/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include "util.h"
#include "draw.h"
#include "marker.h"
#include "prefs.h"
#include "meter.h"

static void MeterDoDraw(meter_type *meter, Boolean draw_real) CODE3_SECTION;
static void UpdateOdo(meter_type *meter, nmea_type *nmea) CODE3_SECTION;
static void UpdateCompass(meter_type *meter, nmea_type *nmea) CODE3_SECTION;
static void UpdateVario(meter_type *meter, nmea_type *nmea) CODE3_SECTION;

void MeterInit(meter_type *meter, MeterType type)
{
	meter->type         = type;
	meter->x_dbuf       = 0;
	meter->y_dbuf       = 0;
	meter->firsttime    = true;
	meter->eraselast    = false;
	meter->radius       = 0;
	meter->iradius      = 0;
	meter->aval         = -1;
	meter->aval2        = 0;
	meter->dval         = 0L;
	meter->has_aval     = true;
	meter->has_aval2    = true;
	meter->has_dval     = true;
	meter->fixtime		= 0L;
	meter->wants_dval   = true;
	meter->is_fixed     = true;
	meter->font         = stdFont;
	meter->ufont        = boldFont;
	meter->meter_update = NULL;
	meter->square       = UtilIsHires() ? 6 : 4;
	meter->do_dbuf		= false;
	meter->dbuf			= NULL;
	meter->draw_back    = NULL;
	meter->draw_scale   = NULL;
	meter->draw_needle  = NULL;
	meter->draw_value   = DrawValue;
	RctSetRectangle(&(meter->area_screen.rect), 0, 0, 0, 0);
	RctSetRectangle(&(meter->area_virt.rect), 0, 0, 0, 0);
	meter->area_screen.center_x = 0;
	meter->area_screen.center_y = 0;
	meter->area_virt.center_x   = 0;
	meter->area_virt.center_y   = 0;
	meter->mv_back		= false;
	meter->mv_scale		= false;
	meter->mv_needle	= true;

	switch(type)
	{
		case TYPE_COMPASS:
				meter->has_aval2 	= false;
				meter->has_dval  	= false;
				meter->draw_back    = DrawCompassBack;
				meter->draw_scale   = DrawCompassScale;
				meter->draw_needle  = DrawCompassNeedle;
				meter->meter_update = UpdateCompass;
			break;
		case TYPE_ODO:
				meter->aval2 		= PrefsGetPrefs()->max_speed;
				meter->draw_back    = DrawOdoBack;
				meter->draw_scale   = DrawOdoScale;
				meter->draw_needle  = DrawOdoNeedle;
				meter->meter_update = UpdateOdo;
			break;
		case TYPE_VARIO:
				meter->has_aval2 	= false;
				meter->has_dval  	= false;
				meter->draw_back    = DrawVarioBack;
				meter->draw_scale   = DrawVarioScale;
				meter->draw_needle  = DrawVarioNeedle;
				meter->meter_update = UpdateVario;
			break;
		default:
				SndPlaySystemSound(sndError);
			break;
	}
}

void MeterFinish(meter_type *meter)
{
	if(meter->dbuf != NULL)
	{
		WinDeleteWindow(meter->dbuf, false);
		meter->dbuf = NULL;
	}
}

void MeterSetFixed(meter_type *meter, Boolean b)
{
	meter->is_fixed  = b;
	meter->mv_needle = b;
	meter->mv_scale  = !b;

	/* take care to not draw, if window is not initialized! */
	if(meter->firsttime == false)
	{
		meter->firsttime = true;
		MeterDraw(meter);
	}
}

/* must be set before the first MeterSetSize()! */
void MeterSetDoublebuffered(meter_type *meter, Boolean b)
{
	meter->do_dbuf = b;
}

void MeterSetSize(meter_type *meter, Coord x, Coord y, Coord w, Coord h)
{
	Int8 mul;
	RectangleType *dr = &meter->area_screen.rect;

	if(meter->dbuf != NULL)
	{
		WinDeleteWindow(meter->dbuf, false);
		meter->dbuf = NULL;
	}

	mul = 1;
	if(UtilIsHires() == true)
	{
		mul = 2;
		dr->topLeft.x = x<<1;
		dr->topLeft.y = y<<1;
		dr->extent.x  = w<<1;
		dr->extent.y  = h<<1;
	}
	else
	{
		dr->topLeft.x = x;
		dr->topLeft.y = y;
		dr->extent.x  = w;
		dr->extent.y  = h;
	}

	if(meter->do_dbuf)
	{
		meter->x_dbuf = dr->topLeft.x;
		meter->y_dbuf = dr->topLeft.y;
	}

	switch(meter->type)
	{
		case TYPE_ODO:
			meter->radius   = MIN(	 dr->extent.x >> 1,
									 dr->extent.y);
			meter->iradius  =		 meter->radius - (meter->square << 1);
			meter->area_screen.center_x =
									 (dr->extent.x >> 1) + dr->topLeft.x;
			meter->area_screen.center_y =
									 dr->extent.y + dr->topLeft.y;
		break;
		case TYPE_VARIO:
			meter->radius   = MIN(	 dr->extent.x,
									 dr->extent.y >> 1);
			meter->iradius  =		 meter->radius - 5;
			meter->area_screen.center_x =
									 dr->topLeft.x;
			meter->area_screen.center_y =
									(dr->extent.y >> 1) +
									 dr->topLeft.y;
		break;
		case TYPE_COMPASS:
		default:
			{
				FontID old_font;
				if(UtilIsVGA())
					old_font = FntSetFont(VgaBaseToVgaFont(meter->font));
				else
					old_font = FntSetFont(meter->font);

				meter->radius   =		 dr->extent.x >> 1;
				meter->iradius  =		 meter->radius-mul*FntCharHeight()-2;
				meter->area_screen.center_x =
										 dr->topLeft.x + meter->radius;
				meter->area_screen.center_y =
										 dr->topLeft.y + meter->radius;
				FntSetFont(old_font);
			}
		break;
	}

	if(meter->do_dbuf)
	{
		Err err;

		meter->area_virt.rect.topLeft.x = 0;
		meter->area_virt.rect.topLeft.y = 0;
		meter->area_virt.rect.extent.x = meter->area_screen.rect.extent.x;
		meter->area_virt.rect.extent.y = meter->area_screen.rect.extent.y;
		meter->area_virt.center_x = meter->area_screen.center_x - meter->area_screen.rect.topLeft.x;
		meter->area_virt.center_y = meter->area_screen.center_y - meter->area_screen.rect.topLeft.y;

		meter->dbuf = WinCreateOffscreenWindow(
							dr->extent.x, dr->extent.y,
							nativeFormat, &err);
		if(err != errNone)
		{
			meter->do_dbuf = false;
			meter->dbuf    = NULL;
			MeterSetSize(meter, x, y, w, h);
		}
	}
	else
	{
		RctCopyRectangle(&meter->area_screen.rect, &meter->area_virt.rect);
		meter->area_virt.center_x = meter->area_screen.center_x;
		meter->area_virt.center_y = meter->area_screen.center_y;
	}
	meter->firsttime = true;
}

void MeterDraw(meter_type *meter)
{
	WinHandle cw = NULL;
	CSYS_INIT

	CSYS_DOUBLE

	if(meter->do_dbuf)
		cw = WinSetDrawWindow(meter->dbuf);

	MeterDoDraw(meter, false);

	if(meter->do_dbuf)
	{
		WinSetDrawWindow(cw);
		WinCopyRectangle(	meter->dbuf, cw, &meter->area_virt.rect,
							meter->x_dbuf, meter->y_dbuf, winPaint);
	}

	MeterDoDraw(meter, true);

	CSYS_RESET
}

static void MeterDoDraw(meter_type *meter, Boolean draw_real)
{
	if(draw_real == false)
	{
		/* redraw all when not dbuf and not fixed */
		if((meter->do_dbuf == false) && (meter->is_fixed == false))
			meter->firsttime = true;

		if(meter->firsttime)
		{
			WinEraseRectangle(&meter->area_screen.rect, 0);
			meter->eraselast = false; // XXX required?
			meter->firsttime = false;

			if(meter->mv_back == false)
				meter->draw_back(meter, &meter->area_virt);

			if(meter->mv_scale == false)
				meter->draw_scale(meter, &meter->area_virt);

			//XXX ugly hack (draws needle even when not moving,
			//    just because of the target point
			if((meter->mv_needle == false) && (meter->is_fixed == true))
				meter->draw_needle(meter, &meter->area_virt);
		}
	}
	else
	{
		if(meter->is_fixed == false)
			meter->eraselast = false; // XXX required?

		if(meter->mv_back == true)
			meter->draw_back(meter, &meter->area_screen);

		if(meter->mv_scale == true)
			meter->draw_scale(meter, &meter->area_screen);

		if((meter->mv_needle == true) || (meter->is_fixed == false))
			meter->draw_needle(meter, &meter->area_screen);

		meter->draw_value(meter, &meter->area_screen);
	}
}

void MeterErase(meter_type *meter)
{
	CSYS_INIT

	CSYS_DOUBLE

	WinEraseRectangle(&meter->area_screen.rect, 0);

	CSYS_RESET
}

void MeterUpdate(meter_type *meter, nmea_type *nmea)
{
	meter->meter_update(meter, nmea);
}

Boolean MeterPenInside(meter_type *meter, EventPtr event)
{
	Coord x, y;

	x = event->screenX;
	y = event->screenY;

	if(UtilIsHires() == true)
	{
		x <<= 1;
		y <<= 1;
	}

	return(RctPtInRectangle(x, y, &meter->area_screen.rect));
}

static void UpdateOdo(meter_type *meter, nmea_type *nmea)
{
	if(nmea->is_new_rmc || nmea->is_new_vtg || nmea->is_new_nexrmc)
	{
		double speed, dist;
		Int32 max_speed, ispeed;
		Char *unit;
		CAppPrefs *prefs;

		prefs = PrefsGetPrefs();

		dist = prefs->dist_f;
		PrefsNauticToPref(&dist, NULL);
		speed = nmea->speed;
		PrefsSpeedToPref(&speed, &unit);

		ispeed       = (Int32)speed;
		meter->aval  = ispeed > 180 ? 180 : ispeed;
		max_speed    = MAX(meter->aval, prefs->max_speed);
		meter->aval2 = max_speed;
		meter->dval  = (UInt32)dist;
		prefs->max_speed = max_speed;

		MeterDraw(meter);
	}
}

static void UpdateCompass(meter_type *meter, nmea_type *nmea)
{
	if(nmea->has_gps == false)
		return;

	if(nmea->is_new_gga || nmea->is_new_nexrmc)
	{
		meter->fixtime = nmea->fix_taken_hsec;

		if(MarkerHasDestination() == true)
		{
			meter->aval2 = MarkerGetCourse(nmea);
			meter->dval  = MarkerGetDistance(nmea);

			meter->has_aval2 = true;
			meter->has_dval  = true;
		}
		else
		{
			if(meter->has_aval2)
				meter->firsttime = true; /* redraw back */

			meter->has_aval2 = false;
			meter->has_dval  = false;
		}
		MeterDraw(meter);
	}

	if(nmea->is_new_rmc || nmea->is_new_vtg || nmea->is_new_nexrmc)
	{
		meter->aval = nmea->course;

		/* delay to next GGA or NORMC sentence, if available */
		if(nmea->has_gga == false && nmea->has_nexrmc == false)
			MeterDraw(meter);
	}
}

static void UpdateVario(meter_type *meter, nmea_type *nmea)
{
	if(nmea->has_gps == false)
		return;

	if(nmea->is_new_gga || nmea->is_new_nexrmc)
	{
		meter->fixtime = nmea->fix_taken_hsec;
		meter->aval = meter->aval2 - nmea->alt;
		if(meter->aval > VARIO_MAX_METER)
			meter->aval = VARIO_MAX_METER;
		if(meter->aval < -VARIO_MAX_METER)
			meter->aval = -VARIO_MAX_METER;
		meter->aval *= VARIO_MUL;
		MeterDraw(meter);
		meter->aval2 = nmea->alt;
	}
}

