/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include "nmea.h"
#include "main.h"
#include "util.h"
#include "resource.h"
#include "MathLib.h"
#include "def.h"

static Char *NMEAStrTok(Char *str) CODE4_SECTION;
static void NMEAParseFrac(Int32 *val, Char *s) CODE4_SECTION;
static void NMEAParseMinutes(UInt32 *minutes, Char *s) CODE4_SECTION;
static Boolean NMEACheckChecksum(char *nmeabuf) CODE4_SECTION;
static UInt32 NMEAHex2Int(UInt8 *buf) CODE4_SECTION;

static void NMEAParseGGA(UInt8 *nmeabuf, nmea_type *nmea) CODE4_SECTION;
static void NMEAParseGSA(UInt8 *nmeabuf, nmea_type *nmea) CODE4_SECTION;
static void NMEAParseRMC(UInt8 *nmeabuf, nmea_type *nmea) CODE4_SECTION;
static void NMEAParseNexianRMC(UInt8 *nmeabuf, nmea_type *nmea) CODE4_SECTION;
static void NMEAParseGSV(UInt8 *nmeabuf, nmea_type *nmea) CODE4_SECTION;
static void NMEAParseVTG(UInt8 *nmeabuf, nmea_type *nmea) CODE4_SECTION;
static void NMEAParsePMGNST(UInt8 *nmeabuf, nmea_type *nmea) CODE4_SECTION;
static Char UTMLetterDesignator(double lat) CODE4_SECTION;
static void CoordsToOSGBSquare(double easting, double northing,  
				  char *OSGBGridSquare, Int32 *OSGBEasting, Int32 *OSGBNorthing) CODE4_SECTION;
static void OSGBSquareToRefCoords(const char* OSGBGridSquare,
						   Int32 *RefEasting, Int32 *RefNorthing) CODE4_SECTION;

static double lon_last=0.;
static double lat_last=0.;

#define DELIM ','

static double k0, majoraxis, minoraxis;
static double eccSquared, eccSquared2, eccSquared3, eccPrimeSquared;
static double LongOrigin, LongOriginRad, LatOrigin, LatOriginRad;
static double M0, e1;
static char GridSquare[] = "VWXYZQRSTULMNOPFGHJKABCDE";

void NMEAInit(nmea_type *nmea)
{
	NMEAClear(nmea);

	// these values are for OSGB only, UTM values are a little bit different!
	k0				= 0.9996012717;
	majoraxis		= 6377563.396;//Airy
	minoraxis		= 6356256.91;//Airy
	eccSquared		= (majoraxis * majoraxis - minoraxis * minoraxis) / (majoraxis * majoraxis);
	eccSquared2		= eccSquared*eccSquared;
	eccSquared3		= eccSquared2*eccSquared;
	eccPrimeSquared	= (eccSquared)/(1-eccSquared);
	e1 = (1-sqrt(1-eccSquared))/(1+sqrt(1-eccSquared));

	LongOrigin		= -2;
	LongOriginRad	= LongOrigin * M_PI/180.;
	LatOrigin		= 49;
	LatOriginRad	= LatOrigin * M_PI/180.;
	//only calculate M0 once since it is based on the origin 
	//of the OSGB projection, which is fixed
	M0 = majoraxis*((1 - eccSquared/4 - 3*eccSquared2/64 - 5*eccSquared3/256)*LatOriginRad - (3*eccSquared/8 + 3*eccSquared2/32 + 45*eccSquared3/1024)*sin(2*LatOriginRad) + (15*eccSquared2/256 + 45*eccSquared3/1024)*sin(4*LatOriginRad) - (35*eccSquared3/3072)*sin(6*LatOriginRad));
}

void NMEAFinish(nmea_type *nmea)
{
}

void NMEAClear(nmea_type *nmea)
{
	lon_last = 0.;
	lat_last = 0.;

	if(nmea == NULL)
		return;

	MemSet(nmea, sizeof(nmea_type), 0);

	nmea->has_gps = false;
	nmea->has_gga = nmea->has_gsa = false;
	nmea->has_rmc = nmea->has_gsv = false;
	nmea->has_vtg = nmea->has_pmgnst = false;
	nmea->has_nexrmc = false;

	nmea->is_new_gga = nmea->is_new_gsa = false;
	nmea->is_new_rmc = nmea->is_new_gsv = false;
	nmea->is_new_vtg = nmea->is_new_pmgnst = false;
	nmea->is_new_nexrmc = false;

	nmea->avg_pos_cnt = 0;
	nmea->avg_pos_next = 0;
}

static Char *NMEAStrTok(Char *str)
{
	static Char *sp, *ep;

	if((str == NULL) && (sp == NULL))
		return(NULL);

	if(str != NULL)
		ep = StrChr(str, DELIM) + 1;

	sp = ep;
	ep = StrChr(sp, DELIM);
	if(ep == NULL)
		ep = StrChr(sp, '\0');
	else
		*ep++ = '\0';

	return(sp);
}

static void NMEAParseFrac(Int32 *val, Char *s)
{
	char *sp;
	Int32 *frac;
	UInt16 *dec;

	frac = val + 1;
	dec = (UInt16*) (frac + 1);

	*dec  = 0;
	*frac = 0;

	sp = StrChr(s, '.');
	if(sp != NULL)
	{
		*sp++ = '\0';
		*dec  = StrLen(sp);
		*frac = StrAToI(sp);
	}
	*val = StrAToI(s);
}

static void NMEAParseMinutes(UInt32 *minutes, Char *s)
{
	// keep order! must not be on stack!
	static Int32 val;
	static Int32 val_frac;
	static UInt16 val_dec;

	Int16 i;

	val = val_frac = val_dec = 0;
	NMEAParseFrac(&val, s);

	*minutes = val * 60;

	for(i = 0; i < val_dec; i++)
		val_frac *= 10;

	*minutes += (UInt32)(val_frac * .6);
}

void NMEAReadSentence(UInt8 *nmeabuf, nmea_type *nmea)
{
	UInt8 nmeadata[NMEA_MAX_LENGTH];
	UInt8 *cp;

	cp = StrChr(nmeabuf, '$');
	if(cp == NULL) /* not a NMEA sentence */
		return;

	StrCopy(nmeadata, cp);

	if(NMEACheckChecksum(nmeadata) == false)
		return;

	nmea->has_gps = true;

	if(StrNCompare("$GPGGA,", nmeadata, 7) == 0)
	{
		NMEAParseGGA(nmeadata, nmea);
	}
	else
	if(StrNCompare("$GPGSA,", nmeadata, 7) == 0)
	{
		NMEAParseGSA(nmeadata, nmea);
	}
	else
	if(StrNCompare("$GPRMC,", nmeadata, 7) == 0)
	{
		NMEAParseRMC(nmeadata, nmea);
	}
	else
	if(StrNCompare("$GPGSV,", nmeadata, 7) == 0)
	{
		NMEAParseGSV(nmeadata, nmea);
	}
	else
	if(StrNCompare("$GPVTG,", nmeadata, 7) == 0)
	{
		NMEAParseVTG(nmeadata, nmea);
	}
	else
	if(StrNCompare("$PMGNST,", nmeadata, 8) == 0)
	{
		NMEAParsePMGNST(nmeadata, nmea);
	}
	else
	if(StrNCompare("$NORMC,", nmeadata, 7) == 0)
	{
		NMEAParseNexianRMC(nmeadata, nmea);
	}
	else
	if(StrNCompare("$NOGSV,", nmeadata, 7) == 0)
	{
		NMEAParseGSV(nmeadata, nmea);
		nmea->sat_tracked = nmea->sat_in_view;
	}
}

static Boolean NMEACheckChecksum(char *nmeabuf)
{
	UInt8 *cp;
	UInt32 cs_is, cs_should;

	cp = StrChr(nmeabuf, '*');
	if(cp)
	{
		*cp++ = '\0';	// cut off checksum
		cs_should = NMEAHex2Int(cp);

		for(cs_is = 0, cp = nmeabuf+1; *cp; cp++)
			cs_is ^= *cp;

		return(cs_is == cs_should);
	}

	return(true);
}

/* converts a two byte char array containing hex values to int */
static UInt32 NMEAHex2Int(UInt8 *buf)
{
	UInt32 n1, n2;
	char h = buf[0];
	char l = buf[1];

	n1 = n2 = 0;

	if(UtilCharIsDigit(h))
		n1 = h - '0';
	else
		if(UtilCharIsAlpha(h))
			n1 = h - (UtilCharIsUpper(h) ? 'A' - 10 : 'a' - 10);

	if(UtilCharIsDigit(l))
		n2 = l - '0';
	else
		if(UtilCharIsAlpha(l))
			n2 = l - (UtilCharIsUpper(l) ? 'A' - 10 : 'a' - 10);

	return((n1 << 4)+n2);
}

static void NmeaParseFixTime(Char *sp, nmea_type *nmea)
{
	UInt8 has_comma;
	has_comma = (StrChr(sp, '.') != NULL);
	sp[6] = '\0';
	nmea->fix_taken_sec = StrAToI(&sp[4]);
	sp[4] = '\0';
	nmea->fix_taken_min = StrAToI(&sp[2]);
	sp[2] = '\0';
	nmea->fix_taken_hour = StrAToI(&sp[0]);

	/*
	 * Garmin e-trex summit sends time without comma
	 */
	nmea->fix_taken_hsec  = nmea->fix_taken_hour * 360000; //60*60*100;
	nmea->fix_taken_hsec += nmea->fix_taken_min  *   6000; //60*100;
	nmea->fix_taken_hsec += nmea->fix_taken_sec  *    100;
	if(has_comma)
	{
		nmea->fix_taken_hsec += StrAToI(&sp[8]);
		sp[8] = '\0';
		nmea->fix_taken_hsec += 10 * StrAToI(&sp[7]);
	}
}

static void NMEAParseGGA(UInt8 *nmeabuf, nmea_type *nmea)
{
	Char *sp;
	Int32 deg, min, frac;
	UInt16 dec;
	char c;

	/* fix taken */
	sp = NMEAStrTok(nmeabuf);
	NmeaParseFixTime(sp, nmea);

	/* latitude */
	sp = NMEAStrTok(NULL);
	dec = StrLen(&sp[5]);
	frac = StrAToI(&sp[5]);
	sp[5] = '\0';
	min = StrAToI(&sp[2]);
	sp[2] = '\0';
	deg = StrAToI(&sp[0]);

	/* North/South */
	sp = NMEAStrTok(NULL);
	c = *sp;

	nmea->lat = NMEAGetRadianDMM(deg, min, frac, dec, c);

	/* longitude */
	sp = NMEAStrTok(NULL);
	dec = StrLen(&sp[6]);
	frac = StrAToI(&sp[6]);
	sp[6] = '\0';
	min = StrAToI(&sp[3]);
	sp[3] = '\0';
	deg = StrAToI(&sp[0]);

	/* East/West */
	sp = NMEAStrTok(NULL);
	c = *sp;

	nmea->lon = NMEAGetRadianDMM(deg, min, frac, dec, c);

	NMEAAveragePos(nmea);

	/* fix quality */
	sp = NMEAStrTok(NULL);
	nmea->fix_quality = StrAToI(sp);
	if(nmea->has_rmc == false)
		nmea->warning = !nmea->fix_quality;

	/* number of satellites being tracked */
	sp = NMEAStrTok(NULL);
	nmea->sat_tracked = StrAToI(sp);

	/* horizontal dilution of position */
	sp = NMEAStrTok(NULL);
	NMEAParseFrac(&nmea->gga_hdop, sp);
	if((nmea->gga_hdop == 0) && (nmea->gga_hdop_fra == 0))
	{
		nmea->gga_hdop     = 50;
		nmea->gga_hdop_fra = 0;
	}

	/* altitude */
	sp = NMEAStrTok(NULL);
	NMEAParseFrac(&nmea->alt, sp);

	/* altitude unit */
	sp = NMEAStrTok(NULL);
	nmea->alt_unit = *sp;

	/* altitude above WGS84 */
	sp = NMEAStrTok(NULL);
	NMEAParseFrac(&nmea->alt_WGS84, sp);

	/* altitude WGS84 unit */
	sp = NMEAStrTok(NULL);
	nmea->alt_WGS84_unit = *sp;

	nmea->is_new_gga = true;
	nmea->has_gga = true;
}

static void NMEAParseGSA(UInt8 *nmeabuf, nmea_type *nmea)
{
	UInt16 i;
	Char *sp;

	/* Auto selection of 2D or 3D fix (we don't care) */
	sp = NMEAStrTok(nmeabuf);

	/* dimension of fix */
	sp = NMEAStrTok(NULL);
	nmea->fix_dim = StrAToI(sp);

	for(i = 0; i < MAXSAT; i++)
	{
		nmea->sat_prn[i] = -1;
		sp = NMEAStrTok(NULL);

		if(*sp != '\0')
			nmea->sat_prn[i] = StrAToI(sp);
	}

	/* dilution of precision */
	sp = NMEAStrTok(NULL);
	NMEAParseFrac(&nmea->pdop, sp);
	if((nmea->pdop == 0) && (nmea->pdop_fra == 0))
	{
		nmea->pdop     = 50;
		nmea->pdop_fra = 0;
	}

	/* horizontal dilution of precision */
	sp = NMEAStrTok(NULL);
	NMEAParseFrac(&nmea->hdop, sp);
	if((nmea->hdop == 0) && (nmea->hdop_fra == 0))
	{
		nmea->hdop     = 50;
		nmea->hdop_fra = 0;
	}

	/* vertical dilution of precision */
	sp = NMEAStrTok(NULL);
	NMEAParseFrac(&nmea->vdop, sp);
	if((nmea->vdop == 0) && (nmea->vdop_fra == 0))
	{
		nmea->vdop     = 50;
		nmea->vdop_fra = 0;
	}

	nmea->is_new_gsa = true;
	nmea->has_gsa = true;
}

static void NMEAParseRMC(UInt8 *nmeabuf, nmea_type *nmea)
{
	Char *sp;

	/* time of fix */
	sp = NMEAStrTok(nmeabuf);
	NmeaParseFixTime(sp, nmea);

	/* warning */
	sp = NMEAStrTok(NULL);
	nmea->warning = (*sp == 'V') ? 1 : 0;

	/* latitude */
	sp = NMEAStrTok(NULL);
	sp = NMEAStrTok(NULL);

	/* longitude */
	sp = NMEAStrTok(NULL);
	sp = NMEAStrTok(NULL);

	/* speed */
	sp = NMEAStrTok(NULL);
	NMEAParseFrac(&nmea->speed, sp);
	nmea->speed_f = UtilGetDouble(nmea->speed, nmea->speed_fra, nmea->speed_dec);

	/* course made good */
	sp = NMEAStrTok(NULL);
	if(nmea->speed_f != 0.)
		NMEAParseFrac(&nmea->course, sp);

	/* date of fix */
	sp = NMEAStrTok(NULL);
	nmea->date_of_fix_yr = StrAToI(&sp[4]);
	sp[4] = '\0';
	nmea->date_of_fix_mo = StrAToI(&sp[2]);
	sp[2] = '\0';
	nmea->date_of_fix_dy = StrAToI(&sp[0]);

	/* magnetic variation */
	sp = NMEAStrTok(NULL);
	NMEAParseFrac(&nmea->mag_var, sp);

	/* magnetic variation dir */
	sp = NMEAStrTok(NULL);
	if((*sp >= 'A') && (*sp <= 'Z'))
		nmea->mag_var_dir = *sp;
	else
		nmea->mag_var_dir = '-';

	nmea->is_new_rmc = true;
	nmea->has_rmc = true;
}

// These are guessed values, no doc exists!
static void NMEAParseNexianRMC(UInt8 *nmeabuf, nmea_type *nmea)
{
	Char *sp;
	Int32 deg, min, frac;
	UInt16 dec;
	char c;

	/* time of fix */
	sp = NMEAStrTok(nmeabuf);
	NmeaParseFixTime(sp, nmea);

	/* GUESS! warning */
	sp = NMEAStrTok(NULL);
	nmea->warning = (*sp == 'V') ? 1 : 0;

	/* GUESS! fix quality */
	sp = NMEAStrTok(NULL);
	nmea->fix_quality = StrAToI(sp);

	/* GUESS!  horizontal dilution of precision */
	sp = NMEAStrTok(NULL);
	NMEAParseFrac(&nmea->hdop, sp);
	if((nmea->hdop == 0) && (nmea->hdop_fra == 0))
	{
		nmea->hdop     = 50;
		nmea->hdop_fra = 0;
	}

	/* latitude */
	sp = NMEAStrTok(NULL);
	dec = StrLen(&sp[5]);
	frac = StrAToI(&sp[5]);
	sp[5] = '\0';
	min = StrAToI(&sp[2]);
	sp[2] = '\0';
	deg = StrAToI(&sp[0]);

	/* North/South */
	sp = NMEAStrTok(NULL);
	c = *sp;

	nmea->lat = NMEAGetRadianDMM(deg, min, frac, dec, c);

	/* longitude */
	sp = NMEAStrTok(NULL);
	dec = StrLen(&sp[6]);
	frac = StrAToI(&sp[6]);
	sp[6] = '\0';
	min = StrAToI(&sp[3]);
	sp[3] = '\0';
	deg = StrAToI(&sp[0]);

	/* East/West */
	sp = NMEAStrTok(NULL);
	c = *sp;

	nmea->lon = NMEAGetRadianDMM(deg, min, frac, dec, c);

	NMEAAveragePos(nmea);

	/* GUESS! altitude */
	sp = NMEAStrTok(NULL);
	NMEAParseFrac(&nmea->alt, sp);

	/* altitude unit (seems to be fix, guess meters) */
	nmea->alt_unit = 'M';

	/* GUESS! speed */
	sp = NMEAStrTok(NULL);
	NMEAParseFrac(&nmea->speed, sp);
	nmea->speed_f = UtilGetDouble(nmea->speed, nmea->speed_fra, nmea->speed_dec);

	/* GUESS! course made good */
	sp = NMEAStrTok(NULL);
	if(nmea->speed_f != 0.)
		NMEAParseFrac(&nmea->course, sp);

	/* date of fix */
	sp = NMEAStrTok(NULL);
	nmea->date_of_fix_yr = StrAToI(&sp[4]);
	sp[4] = '\0';
	nmea->date_of_fix_mo = StrAToI(&sp[2]);
	sp[2] = '\0';
	nmea->date_of_fix_dy = StrAToI(&sp[0]);

	nmea->has_nexrmc = true;
	nmea->is_new_nexrmc = true;
}

static void NMEAParseGSV(UInt8 *nmeabuf, nmea_type *nmea)
{
	Int32 sen, act, i;
	Char *sp;

	/* number of sentences */
	sp = NMEAStrTok(nmeabuf);
	sen = StrAToI(sp);

	/* this sentence is number ... */
	sp = NMEAStrTok(NULL);
	act = StrAToI(sp);

	/* number of satellites in view */
	sp = NMEAStrTok(NULL);
	nmea->sat_in_view = StrAToI(sp);

	/* reset counter and clear unseen sats */
	if(act == 1)
	{
		nmea->last_gsv = 0;
		for(i = nmea->sat_in_view; i < MAXSAT; i++)
		{
			nmea->sat_data[i].prn = 0;
			nmea->sat_data[i].elevation = 0;
			nmea->sat_data[i].azimuth = 0;
			nmea->sat_data[i].signal = 0;
		}
	}

	while(true)
	{
		if(nmea->last_gsv >= MAXSAT)
			break;

		nmea->sat_data[nmea->last_gsv].prn = 0;
		nmea->sat_data[nmea->last_gsv].elevation = 0;
		nmea->sat_data[nmea->last_gsv].azimuth = 0;
		nmea->sat_data[nmea->last_gsv].signal = 0;

		sp = NMEAStrTok(NULL);
		if(*sp == '\0')
			break;
		nmea->sat_data[nmea->last_gsv].prn = StrAToI(sp);

		sp = NMEAStrTok(NULL);
		nmea->sat_data[nmea->last_gsv].elevation = StrAToI(sp);

		sp = NMEAStrTok(NULL);
		nmea->sat_data[nmea->last_gsv].azimuth = StrAToI(sp);

		sp = NMEAStrTok(NULL);
		nmea->sat_data[nmea->last_gsv].signal = StrAToI(sp);

		nmea->last_gsv++;
	}

	if(act == sen)
		nmea->is_new_gsv = true;
	else
		nmea->is_new_gsv = false;
	nmea->has_gsv = true;
}

static void NMEAParseVTG(UInt8 *nmeabuf, nmea_type *nmea)
{
	Char *sp;

	/* track made good */
	sp = NMEAStrTok(nmeabuf);
	if(nmea->speed_f != 0.)
		NMEAParseFrac(&nmea->course, sp);

	/* Fixed text 'T' indicates that track made good is relative to true north */
	sp = NMEAStrTok(NULL);

	/* Magnetic track made good */
	sp = NMEAStrTok(NULL);

	/* Fixed text 'M' indicates that track made good is relative to magnetic north */
	sp = NMEAStrTok(NULL);

	/* Speed over ground in knots */
	sp = NMEAStrTok(NULL);
	NMEAParseFrac(&nmea->speed, sp);
	nmea->speed_f = UtilGetDouble(nmea->speed, nmea->speed_fra, nmea->speed_dec);

	/* Fixed text 'N' indicates that speed over ground in in knots */
	sp = NMEAStrTok(NULL);

	/* Speed over ground in kilometers/hour */
	sp = NMEAStrTok(NULL);

	/* Fixed text 'K' indicates that speed over ground is in kilometers/hour */
	sp = NMEAStrTok(NULL);

	nmea->is_new_vtg = true;
	nmea->has_vtg = true;
}

// $PMGNST,02.12,3,T,534,05.0,+03327,00*40
/*
$PMGNST,mm.nn,t,f,bbb,hh.h,oooooo,pp*cc
                  mm       =  Software major version number
                  nn       =  Software minor version number
                  t        =  Type of last fix (0 - none, 1 - init, 2 
2D, 3 - 3D)
                  f        =  T if first fix, F if no fix yet
                  bbb      =  Battery voltage level (0-999) in arbitrary
units
                  hh       =  Hours of operation left on battery
(approximate)
                  oooooo   =  Oscillator offset in Hz at L-band.
                  Pp       =  proprietary information
                  *cc      = Checksum
*/

static void NMEAParsePMGNST(UInt8 *nmeabuf, nmea_type *nmea)
{
	Char *sp;

	/* Version number */
	sp = NMEAStrTok(nmeabuf);

	/* type of last fix */
	sp = NMEAStrTok(NULL);

	/* 'T' if first fix; 'F' if no fix yet */
	sp = NMEAStrTok(NULL);

	/* battery voltage level */
	sp = NMEAStrTok(NULL);

	/* time left on the gps battery in hours */
	sp = NMEAStrTok(NULL);
	NMEAParseMinutes(&nmea->mgn_bat_left, sp);

	/* oscillator offset */
	sp = NMEAStrTok(NULL);

	/* PRN number receiving current focus */
	sp = NMEAStrTok(NULL);

	nmea->is_new_pmgnst = true;
	nmea->has_pmgnst = true;
}

void NMEAAveragePos(nmea_type *nmea)
{
	if(nmea->speed == 0)
	{
		UInt16 i;
		double slon, slat;

		if(nmea->avg_pos_cnt == 0)
		{
			for(i = 0; i < MAX_POS_COLLECT; i++)
			{
				nmea->_lat[i] = 0.;
				nmea->_lon[i] = 0.;
			}
			nmea->avg_pos_next = 0;
		}
		nmea->_lat[nmea->avg_pos_next] = nmea->lat;
		nmea->_lon[nmea->avg_pos_next] = nmea->lon;

		nmea->avg_pos_next = (nmea->avg_pos_next + 1) % MAX_POS_COLLECT;
		nmea->avg_pos_cnt = MIN(nmea->avg_pos_cnt+1, MAX_POS_COLLECT);

		slon = slat = 0.;
		for(i = 0; i < nmea->avg_pos_cnt; i++)
		{
			slon += nmea->_lon[i];
			slat += nmea->_lat[i];
		}
		nmea->lon = slon / (double)(nmea->avg_pos_cnt);
		nmea->lat = slat / (double)(nmea->avg_pos_cnt);
	}
	else
	{
		nmea->avg_pos_cnt = 0;
	}
}

double NMEAGetRadianDD(Int32 deg, Int32 min, UInt16 dec, Char c)
{
	Int16 i;
	double rad;

	rad = (double)min;
	for(i = 0; i < dec; i++)
		rad /= 10.;
	rad += deg;
	rad  = rad * M_PI/180.;

	if((c == 'S') || (c == 'E'))
		rad = -rad;

	return(rad);
}

double NMEAGetRadianDMM(Int32 deg, Int32 min, Int32 fra,
						UInt16 dec, Char c)
{
	Int16 i;
	double rad;

	rad = (double)fra;
	for(i = 0; i < dec; i++)
		rad /= 10.;
	rad += min;
	rad /= 60.;
	rad += deg;
	rad  = rad * M_PI/180.;

	if((c == 'S') || (c == 'E'))
		rad = -rad;

	return(rad);
}

double NMEAGetRadianDMS(Int32 deg, Int32 min, Int32 sec, Char c)
{
	double rad;

	rad  = (double)sec / 3600.;
	rad += (double)min / 60.;
	rad += (double)deg;
	rad  = rad * M_PI/180.;

	if((c == 'S') || (c == 'E'))
		rad = -rad;

	return(rad);
}

double NMEAGetRadianDMSS(Int32 deg, Int32 min, Int32 sec, Int32 fra,
						UInt16 dec, Char c)
{
	Int16 i;
	double rad;

	rad = (double)fra;
	for(i = 0; i < dec; i++)
		rad /= 10.;
	rad += sec;
	rad /= 3600.;
	rad += min / 60.;
	rad += deg;
	rad  = rad * M_PI/180.;

	if((c == 'S') || (c == 'E'))
		rad = -rad;

	return(rad);
}

/* converts UTM coords to lat/long.  Equations from USGS Bulletin 1532 
 * East Longitudes are positive, West longitudes are negative. 
 * North latitudes are positive, South latitudes are negative
 * Written by Chuck Gantz- chuck.gantz@globalstar.com
 */
void NMEAGetRadianUTM(Int32 northing, Int32 easting, Char *zone,
						double *lat, double *lon)
{

	static const double a			= 6378137;
	static const double k0			= 0.9996;
	static const double eccSquared	= 0.00669438;
	static const double eccPrimeSquared = (eccSquared)/(1-eccSquared);
	static const double eccSquared2 = eccSquared*eccSquared;
	static const double eccSquared3 = eccSquared2*eccSquared;

	double LongOrigin;
	double N, T, C1, C2, R, M;
	double D1, D2, D3, D4, D5, D6;
	double mu, phi;
	double x, y;
	Int16 ZoneNumber;
	char ZoneLetter;
	Char z[3];

	/* remove 500000 meter offset for longitude */
	x = easting - 500000.;
	y = northing;

	z[0] = zone[0];
	z[1] = zone[1];
	z[2] = '\0';
	ZoneLetter = zone[2];
	ZoneNumber = (Int16)StrAToI(z);
	if((ZoneLetter - 'N') < 0)
	{
		/* point is in southern hemisphere */
		/* remove 10,000,000 meter offset used for southern hemisphere */
		y -= 10000000.;
	}

	/* +3 puts origin in middle of zone */
	LongOrigin = (ZoneNumber-1)*6 - 180 + 3;
	LongOrigin *= M_PI / 180.;

	M = y / k0;
	mu = M/(a*(1-eccSquared/4-3*eccSquared2/64-5*eccSquared3/256));

	phi = mu	+ (3*e1/2-27*e1*e1*e1/32)*sin(2*mu) 
				+ (21*e1*e1/16-55*e1*e1*e1*e1/32)*sin(4*mu)
				+(151*e1*e1*e1/96)*sin(6*mu);

	N  = a/sqrt(1-eccSquared*sin(phi)*sin(phi));
	T  = tan(phi)*tan(phi);
	C1 = eccPrimeSquared*cos(phi)*cos(phi);
	C2 = C1*C1;
	R  = a*(1-eccSquared)/pow(1-eccSquared*sin(phi)*sin(phi), 1.5);
	D1 = x/(N*k0);
	D2 = D1*D1;
	D3 = D2*D1;
	D4 = D3*D1;
	D5 = D4*D1;
	D6 = D5*D1;

	*lat = phi - (N*tan(phi)/R)*(D2/2-(5+3*T+10*C1-4*C2-9*eccPrimeSquared)*D4/24+(61+90*T+298*C1+45*T*T-252*eccPrimeSquared-3*C2)*D6/720);

	*lon = LongOrigin + (D1-(1+2*T+C1)*D3/6+(5-2*C1+28*T-3*C2+8*eccPrimeSquared+24*T*T)*D5/120)/cos(phi);
	*lon = -(*lon);
}

/* converts OSGB coords to lat/long.  Equations from USGS Bulletin 1532 
 * East Longitudes are positive, West longitudes are negative. 
 * North latitudes are positive, South latitudes are negative
 * Lat and Long are in decimal degrees. 
 * Written by Chuck Gantz- chuck.gantz@globalstar.com
 */
void NMEAGetRadianOSGB(double OSGBNorthing, double OSGBEasting, char* OSGBZone,
						double *lat, double *lon)
{

	double N1, T1, C1, R1, D, M;
	double mu, phi1, phi1Rad;
	double x, y;
	Int32 RefEasting, RefNorthing;

	OSGBSquareToRefCoords(OSGBZone, &RefEasting, &RefNorthing);
	x = OSGBEasting - 400000. + RefEasting; //remove 400,000 meter false easing for longitude
	y = OSGBNorthing + 100000. + RefNorthing; //remove 100,000 meter false easing for longitude

	M = M0 + y / k0;
	mu = M/(majoraxis*(1-eccSquared/4-3*eccSquared*eccSquared/64-5*eccSquared*eccSquared*eccSquared/256));

	phi1Rad = mu	+ (3*e1/2-27*e1*e1*e1/32)*sin(2*mu) 
					+ (21*e1*e1/16-55*e1*e1*e1*e1/32)*sin(4*mu)
					+ (151*e1*e1*e1/96)*sin(6*mu);
	phi1 = phi1Rad*180./M_PI;

	N1 = majoraxis/sqrt(1-eccSquared*sin(phi1Rad)*sin(phi1Rad));
	T1 = tan(phi1Rad)*tan(phi1Rad);
	C1 = eccPrimeSquared*cos(phi1Rad)*cos(phi1Rad);
	R1 = majoraxis*(1-eccSquared)/pow(1-eccSquared*sin(phi1Rad)*sin(phi1Rad), 1.5);
	D = x/(N1*k0);

	*lat = phi1Rad -
			(N1*tan(phi1Rad)/R1)*(D*D/2-(5+3*T1+10*C1-4*C1*C1-9*eccPrimeSquared)*D*D*D*D/24
				+(61+90*T1+298*C1+45*T1*T1-252*eccPrimeSquared-3*C1*C1)*D*D*D*D*D*D/720);
	*lon = LongOriginRad +
			(D-(1+2*T1+C1)*D*D*D/6+(5-2*C1+28*T1-3*C1*C1+8*eccPrimeSquared+24*T1*T1)
				*D*D*D*D*D/120)/cos(phi1Rad);
	*lon = -(*lon);
}

void OSGBSquareToRefCoords(const char* OSGBGridSquare, Int32 *RefEasting, Int32 *RefNorthing)
{
	int pos, x_multiplier=0, y_multiplier=0;

	/* find 500km offset */
	char ch = OSGBGridSquare[0];
	switch(ch)
	{
		case 'S': x_multiplier = 0; y_multiplier = 0; break;
		case 'T': x_multiplier = 1; y_multiplier = 0; break;
		case 'N': x_multiplier = 0; y_multiplier = 1; break;
		case 'O': x_multiplier = 1; y_multiplier = 1; break;
		case 'H': x_multiplier = 0; y_multiplier = 2; break;
		case 'J': x_multiplier = 1; y_multiplier = 2; break;
	}
	*RefEasting = x_multiplier * 500000L;
	*RefNorthing = y_multiplier * 500000L;

	/* find 100km offset and add to 500km offset to get coordinate of square point is in */
	pos = StrChr(GridSquare, OSGBGridSquare[1]) - GridSquare;
	*RefEasting += ((pos % 5) * 100000L);
	*RefNorthing += ((pos / 5) * 100000L);
}

void NMEAGetRadianQTH(Char *qth, double *lat, double *lon)
{
	if(UtilCharIsLower(qth[0]))
		qth[0] -= 'a' - 'A';
	if(UtilCharIsLower(qth[1]))
		qth[1] -= 'a' - 'A';
	if(UtilCharIsLower(qth[4]))
		qth[4] -= 'a' - 'A';
	if(UtilCharIsLower(qth[5]))
		qth[5] -= 'a' - 'A';

	*lon  = (74 - qth[0]) * 20;
	*lon -= (qth[2] - 48) * 2;
	*lon -= (qth[4] - 65) / 12 - 1 / 24;
	*lat  = (qth[1] - 74) * 10;
	*lat += (qth[3] - 48);
	*lat += (qth[6] - 65) / 24 + 1 / 48;
}


void NMEAGetDeg_DMM(double val, UInt16 dec,
				Int32 *deg, Int32 *min, Int32 *frac,
				Char *c, Latlon latlon)
{
	double t;
	UInt16 i;

	if(val < 0)
	{
		*c = (latlon == LATLON_LAT) ? 'S' : 'E';
		val = -val;
	}
	else
		*c = (latlon == LATLON_LAT) ? 'N' : 'W';

	t = val * 180 / M_PI;
	*deg = (Int32) t;
	t -= *deg;
	t *= 60.;
	*min = (Int32) t;
	t -= *min;
	for(i = 0; i < dec; i++)
		t *= 10.;
	*frac = (Int32) t;
}

void NMEAGetDeg_DD(double val, UInt16 dec,
				Int32 *deg, Int32 *min,
				Char *c, Latlon latlon)
{
	double t;
	UInt16 i;

	if(val < 0)
	{
		*c = (latlon == LATLON_LAT) ? 'S' : 'E';
		val = -val;
	}
	else
		*c = (latlon == LATLON_LAT) ? 'N' : 'W';

	t = val * 180 / M_PI;
	*deg = (Int32) t;
	t -= *deg;
	for(i = 0; i < dec; i++)
		t *= 10.;
	*min = (Int32) t;
}

void NMEAGetDeg_DMS(double val,
				Int32 *deg, Int32 *min, Int32 *sec,
				Char *c, Latlon latlon)
{
	double t;

	if(val < 0)
	{
		*c = (latlon == LATLON_LAT) ? 'S' : 'E';
		val = -val;
	}
	else
		*c = (latlon == LATLON_LAT) ? 'N' : 'W';

	t = val * 180 / M_PI;
	*deg = (Int32) t;
	t -= *deg;
	t *= 60.;
	*min = (Int32) t;
	t -= *min;
	t *= 60.;
	*sec = (Int32) t;
}

void NMEAGetDeg_DMSS(double val, UInt16 dec,
				Int32 *deg, Int32 *min, Int32 *sec, Int32 *frac,
				Char *c, Latlon latlon)
{
	double t;
	UInt16 i;

	if(val < 0)
	{
		*c = (latlon == LATLON_LAT) ? 'S' : 'E';
		val = -val;
	}
	else
		*c = (latlon == LATLON_LAT) ? 'N' : 'W';

	t = val * 180 / M_PI;
	*deg = (Int32) t;
	t -= *deg;
	t *= 60.;
	*min = (Int32) t;
	t -= *min;
	t *= 60.;
	*sec = (Int32) t;
	t -= *sec;
	for(i = 0; i < dec; i++)
		t *= 10.;
	*frac = (Int32) t;
}

/* converts lat/long to UTM coords.  Equations from USGS Bulletin 1532 
 * East Longitudes are positive, West longitudes are negative. 
 * North latitudes are positive, South latitudes are negative
 * Written by Chuck Gantz- chuck.gantz@globalstar.com
 */
void NMEAGetDeg_UTM(double lat, double lon,
					Int32 *northing, Int32 *easting, Char *zone)
{
	static const double a			= 6378137;
	static const double k0			= 0.9996;
	static const double eccSquared	= 0.00669438;
	static const double eccPrimeSquared = (eccSquared)/(1-eccSquared);
	static const double eccSquared2 = eccSquared*eccSquared;
	static const double eccSquared3 = eccSquared2*eccSquared;

	double LongOrigin;
	double LongOriginRad;
	double LatRad;
	double LongRad;
	double N, T, C, A, M;
	double cosLat, tanLat;
	Int16  ZoneNumber;

	/* lat/lon must be in decimal degrees */
	lon		= -lon;
	LatRad	= lat;
	lat		= lat*180./M_PI;
	LongRad	= lon;
	lon		= lon*180./M_PI;

	/* Make sure the longitude is between -180.00 .. 179.9 */
	lon = (lon+180)-(Int32)((lon+180)/360)*360-180;

	ZoneNumber = (Int16)((lon + 180.)/6.) + 1;
  
	if((lat >= 56.) && (lat < 64.) && (lon >= 3.) && (lon < 12.))
		ZoneNumber = 32;

	/* Special zones for Svalbard */
	if((lat >= 72.) && (lat < 84.))
	{
		if(      lon >= 0.  && lon <  9.) ZoneNumber = 31;
		else if( lon >= 9.  && lon < 21.) ZoneNumber = 33;
		else if( lon >= 21. && lon < 33.) ZoneNumber = 35;
		else if( lon >= 33. && lon < 42.) ZoneNumber = 37;
	}

	/* +3 puts origin in middle of zone */
	LongOrigin = (ZoneNumber-1)*6 - 180 + 3;
	LongOriginRad = LongOrigin * M_PI/180.;

	/* compute the UTM Zone from the latitude and longitude */
	StrPrintF(zone, "%d%c", ZoneNumber, UTMLetterDesignator(lat));

	cosLat = cos(LatRad);
	tanLat = tan(LatRad);
	N = a/sqrt(1-eccSquared*sin(LatRad)*sin(LatRad));
	T = tanLat*tanLat;
	C = eccPrimeSquared*cosLat*cosLat;
	A = cosLat*(LongRad-LongOriginRad);

	M = a*((1-eccSquared/4 - 3*eccSquared2/64 - 5*eccSquared3/256)*LatRad - (3*eccSquared/8 + 3*eccSquared2/32 + 45*eccSquared3/1024)*sin(2*LatRad) + (15*eccSquared2/256 + 45*eccSquared3/1024)*sin(4*LatRad) - (35*eccSquared3/3072)*sin(6*LatRad));

	*easting  = (Int32)(k0*N*(A+(1-T+C)*A*A*A/6
					+ (5-18*T+T*T+72*C-58*eccPrimeSquared)*A*A*A*A*A/120)
					+ 500000.+.5);

	*northing = (Int32)(k0*(M+N*tanLat*(A*A/2+(5-T+9*C+4*C*C)*A*A*A*A/24
					+ (61-58*T+T*T+600*C-330*eccPrimeSquared)*A*A*A*A*A*A/720))+.5);
	if(lat < 0)
		*northing += 10000000; //10000000 meter offset for southern hemisphere
}

/* This routine determines the correct UTM letter designator
 * for the given latitude. returns 'Z' if latitude is outside
 * the UTM limits of 84N to 80S
 * Written by Chuck Gantz- chuck.gantz@globalstar.com
 */
static Char UTMLetterDesignator(double lat)
{
	Char LetterDesignator;

	if((84 >= lat) && (lat >= 72)) LetterDesignator = 'X';
	else if((72 > lat) && (lat >= 64)) LetterDesignator = 'W';
	else if((64 > lat) && (lat >= 56)) LetterDesignator = 'V';
	else if((56 > lat) && (lat >= 48)) LetterDesignator = 'U';
	else if((48 > lat) && (lat >= 40)) LetterDesignator = 'T';
	else if((40 > lat) && (lat >= 32)) LetterDesignator = 'S';
	else if((32 > lat) && (lat >= 24)) LetterDesignator = 'R';
	else if((24 > lat) && (lat >= 16)) LetterDesignator = 'Q';
	else if((16 > lat) && (lat >= 8)) LetterDesignator = 'P';
	else if(( 8 > lat) && (lat >= 0)) LetterDesignator = 'N';
	else if(( 0 > lat) && (lat >= -8)) LetterDesignator = 'M';
	else if((-8> lat) && (lat >= -16)) LetterDesignator = 'L';
	else if((-16 > lat) && (lat >= -24)) LetterDesignator = 'K';
	else if((-24 > lat) && (lat >= -32)) LetterDesignator = 'J';
	else if((-32 > lat) && (lat >= -40)) LetterDesignator = 'H';
	else if((-40 > lat) && (lat >= -48)) LetterDesignator = 'G';
	else if((-48 > lat) && (lat >= -56)) LetterDesignator = 'F';
	else if((-56 > lat) && (lat >= -64)) LetterDesignator = 'E';
	else if((-64 > lat) && (lat >= -72)) LetterDesignator = 'D';
	else if((-72 > lat) && (lat >= -80)) LetterDesignator = 'C';
	else LetterDesignator = 'Z'; //This is here as an error flag to show that the Latitude is outside the UTM limits

	return(LetterDesignator);
}

/* converts lat/long to OSGB coords.  Equations from USGS Bulletin 1532 
 * East Longitudes are positive, West longitudes are negative. 
 * North latitudes are positive, South latitudes are negative
 * Lat and Long are in decimal degrees
 * Written by Chuck Gantz- chuck.gantz@globalstar.com
 */
void NMEAGetDeg_OSGB(double lat, double lon,
					Int32 *OSGBNorthing, Int32 *OSGBEasting, char *OSGBGridSquare)
{
	double N, T, C, A, M;
	double easting, northing;
	double sinLat, cosLat, tanLat;

	lon = -lon;

	sinLat = sin(lat);
	cosLat = cos(lat);
	tanLat = tan(lat);

	N = majoraxis/sqrt(1-eccSquared*sinLat*sinLat);
	T = tanLat*tanLat;
	C = eccPrimeSquared*cosLat*cosLat;
	A = cosLat*(lon-LongOriginRad);

	M = majoraxis*((1 - eccSquared/4 - 3*eccSquared2/64 - 5*eccSquared3/256)*lat - (3*eccSquared/8 + 3*eccSquared2/32 + 45*eccSquared3/1024)*sin(2*lat) + (15*eccSquared2/256 + 45*eccSquared3/1024)*sin(4*lat) - (35*eccSquared3/3072)*sin(6*lat));
	
	easting = (double)(k0*N*(A+(1-T+C)*A*A*A/6
					+ (5-18*T+T*T+72*C-58*eccPrimeSquared)*A*A*A*A*A/120));
	easting += 400000.0; //false easting

	northing = (double)(k0*(M-M0+N*tanLat*(A*A/2+(5-T+9*C+4*C*C)*A*A*A*A/24
				 + (61-58*T+T*T+600*C-330*eccPrimeSquared)*A*A*A*A*A*A/720)));
	northing -= 100000.0;//false northing

	CoordsToOSGBSquare(easting, northing, OSGBGridSquare, OSGBEasting, OSGBNorthing);
}

void CoordsToOSGBSquare(double easting, double northing,  
				  char *OSGBGridSquare, Int32 *OSGBEasting, Int32 *OSGBNorthing)
{
	Int32 posx, posy; //positions in grid

	(*OSGBEasting) = (Int32)(easting + 0.5); //round to nearest int
	(*OSGBNorthing) = (Int32)(northing + 0.5); //round to nearest int

	//find correct 500km square
	posx = (*OSGBEasting) / 500000;
	posy = (*OSGBNorthing) / 500000;
	OSGBGridSquare[0] = GridSquare[posx + posy * 5 + 7];

	//find correct 100km square
	posx = (*OSGBEasting) % 500000L;//remove 500 km square
	posy = (*OSGBNorthing) % 500000L;//remove 500 km square
	posx = posx / 100000L;//find 100 km square
	posy = posy / 100000L;//find 100 km square
	OSGBGridSquare[1] = GridSquare[posx + posy * 5];

	OSGBGridSquare[2] = '\0';//terminate grid ref string with null

	//remainder is northing and easting
	(*OSGBNorthing) = (*OSGBNorthing) % 500000; 
	(*OSGBNorthing) = (*OSGBNorthing) % 100000;

	(*OSGBEasting) = (*OSGBEasting) % 500000;
	(*OSGBEasting) = (*OSGBEasting) % 100000;
}

/* QTH-Locator */
void NMEAGetDeg_QTH(double lat, double lon, Char *qth)
{
	Char c1, c2, c3, c4;
	UInt8 n1, n2;

	lat	= lat*180./M_PI;
	lon	= lon*180./M_PI;

	lat +=  90.;
	lon -= 180.;
	if(lon < 0)
		lon = -lon;

	lon = lon / 20.;
	c1 = (Int32)lon + 65;
	lon = 20. * (lon - (Int32)lon);
	lat = lat / 10.;
	c2 = (Int32)lat + 65;
	lat = 10. * (lat - (Int32)lat);
	lon = lon / 2.;
	n1 = (UInt8)lon;
	lon = 2. * (lon - n1);
	n2 = (UInt8)lat;
	lat = lat - n2;
	c3 = (Int32)(lon * 12) + 65;
	c4 = (Int32)(lat * 24) + 65;
	while(n1 > 9)
		n1 = (UInt8)(n1/10.);
	while(n2 > 9)
		n2 = (UInt8)(n2/10.);
	StrPrintF(qth, "%c%c%d%d%c%c", c1, c2, n1, n2, c3, c4);
}

