/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef _nmea_h_
#define _nmea_h_

#define NMEA_MAX_LENGTH 82
#define MAXSAT 12
#define MAX_POS_COLLECT 10

#include "def.h"

typedef enum
{
	LATLON_LAT=0,
	LATLON_LON
} Latlon;

typedef struct 
{
	Int32 prn;			/* sat number */
	Int32 elevation;	/* degrees */
	Int32 azimuth;		/* degrees */
	Int32 signal;		/* signal strength, higher is better */
} satdata_type;

typedef struct
{
	double lat;
	double lon;
	Boolean has_gps;

	/* GGA - Global Positioning System Fix Data */
	Boolean has_gga;
	Boolean is_new_gga;
	Int32 fix_taken_hour;		/* time in UTC */
	Int32 fix_taken_min;
	Int32 fix_taken_sec;
	UInt32 fix_taken_hsec;	/* 1/100th of a sec */
	Int32 fix_quality;		/* 0-invalid, 1-GPS, 2-DGPS */
	Int32 sat_tracked;		/* numbers of satellites being tracked */
	Int32 gga_hdop;			/* horizontal dilution of position */
	Int32 gga_hdop_fra;
	UInt16 gga_hdop_dec;
	Int32 alt;				/* altitude above mean sea level (in meters) */
	Int32 alt_fra;
	UInt16 alt_dec;
	char alt_unit;
	Int32 alt_WGS84;		/* height of geoid above WGS84 ellipsoid */
	Int32 alt_WGS84_fra;
	UInt16 alt_WGS84_dec;
	char alt_WGS84_unit;

	/* GSA - GPS DOP and active satellites */
	Boolean has_gsa;
	Boolean is_new_gsa;
	Int32 fix_dim;			/* dimension of fix */
	Int32 sat_prn[MAXSAT];	/* PRNs of satellites used for fix */
	Int32 pdop;				/* dilution of precision */
	Int32 pdop_fra;
	UInt16 pdop_dec;
	Int32 hdop;				/* horizontal dilution of precision */
	Int32 hdop_fra;
	UInt16 hdop_dec;
	Int32 vdop;				/* vertical dilution of precision */
	Int32 vdop_fra;
	UInt16 vdop_dec;

	/* RMC - Recommended minimum specific GPS/Transit data */
	Boolean has_rmc;
	Boolean is_new_rmc;
	Int32 date_of_fix_dy;	/* date of fix */
	Int32 date_of_fix_mo;
	Int32 date_of_fix_yr;
	Int32 warning;			/* 0-OK, 1-warning */
	Int32 speed;			/* speed over ground */
	Int32 speed_fra;
	UInt16 speed_dec;
	double speed_f;
	Int32 course;			/* course made good in degrees */
	Int32 course_fra;
	UInt16 course_dec;
	Int32 mag_var;			/* magnetic variation */
	Int32 mag_var_fra;
	UInt16 mag_var_dec;
	char mag_var_dir;

	/* NORMC (Nexian HandyGPS) */
	Boolean has_nexrmc;
	Boolean is_new_nexrmc;

	/* GSV - Satellites in view */
	Boolean has_gsv;
	Boolean is_new_gsv;
	Int32 sat_in_view;		/* Number of satellites in view */
	UInt16 last_gsv;		/* PRIVAT USE ONLY! */
	satdata_type sat_data[MAXSAT];

	/* VTG - Vector track an Speed over the Ground /Velocity made good */
	Boolean has_vtg;
	Boolean is_new_vtg;

	/* Magellan status */
	Boolean has_pmgnst;
	Boolean is_new_pmgnst;
	UInt32 mgn_bat_left;	/* time left on the gps battery in minutes */

	/*
	 * collect last MAX_POS_COLLECT positions if speed == 0
	 * and average that
	 */
	UInt16 avg_pos_cnt;				/* PRIVAT USE ONLY! */
	UInt16 avg_pos_next;			/* PRIVAT USE ONLY! */
	double _lat[MAX_POS_COLLECT];	/* PRIVAT USE ONLY! */
	double _lon[MAX_POS_COLLECT];	/* PRIVAT USE ONLY! */

} nmea_type;

void NMEAInit(nmea_type *nmea) CODE4_SECTION;
void NMEAClear(nmea_type *nmea) CODE4_SECTION;
void NMEAFinish(nmea_type *nmea) CODE4_SECTION;
void NMEAReadSentence(UInt8 *nmeabuf, nmea_type *nmea) CODE4_SECTION;
void NMEAAveragePos(nmea_type *nmea) CODE4_SECTION;

double NMEAGetRadianDD(Int32 deg, Int32 min, UInt16 dec, Char c) CODE4_SECTION;
double NMEAGetRadianDMM(Int32 deg, Int32 min, Int32 fra,
						UInt16 dec, Char c) CODE4_SECTION;
double NMEAGetRadianDMS(Int32 deg, Int32 min, Int32 sec, Char c) CODE4_SECTION;
double NMEAGetRadianDMSS(Int32 deg, Int32 min, Int32 sec, Int32 fra,
						UInt16 dec, Char c) CODE4_SECTION;
void NMEAGetRadianUTM(Int32 northing, Int32 easting, Char *zone,
						double *lat, double *lon) CODE4_SECTION;
void NMEAGetRadianOSGB(double northing, double easting, char* zone,
						double *lat, double *lon) CODE4_SECTION;
void NMEAGetRadianQTH(Char *qth, double *lat, double *lon) CODE4_SECTION;

void NMEAGetDeg_DD(double val, UInt16 dec,
				Int32 *deg, Int32 *min, Char *c,
				Latlon latlon) CODE4_SECTION;
void NMEAGetDeg_DMM(double val, UInt16 dec,
				Int32 *deg, Int32 *min, Int32 *frac, Char *c,
				Latlon latlon) CODE4_SECTION;
void NMEAGetDeg_DMS(double val,
				Int32 *deg, Int32 *min, Int32 *sec, Char *c,
				Latlon latlon) CODE4_SECTION;
void NMEAGetDeg_DMSS(double val, UInt16 dec,
				Int32 *deg, Int32 *min, Int32 *sec, Int32 *frac,
				Char *c, Latlon latlon) CODE4_SECTION;
void NMEAGetDeg_UTM(double lat, double lon,
					Int32 *northing, Int32 *easting, Char *zone) CODE4_SECTION;
void NMEAGetDeg_OSGB(double lat, double lon,
					Int32 *northing, Int32 *easting,
					char *gridsquare) CODE4_SECTION;
void NMEAGetDeg_QTH(double lat, double lon, Char *qth) CODE4_SECTION;

#endif /* _nmea_h_ */
