/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "prefs.h"
#include "main.h"
#include "forms.h"
#include "time.h"
#include "serial.h"
#include "geoidalheight.h"

#define D_M2FT 3.2808

#define D_KMH2MS 0.278
#define D_NM2KM	 1.852
#define D_NM2M	 (D_NM2KM * 1000.)
#define D_NM2ML	 1.1507768
#define D_NM2FT	 6076.115
#define D_NM2MS  D_NM2KM*D_KMH2MS
#define D_NM2YD  (D_NM2FT/3.)

#define A_NM2KM	 (D_NM2KM * D_NM2KM)
#define A_NM2HA	 (D_NM2KM * D_NM2KM * 100.)
#define A_NM2M	 (D_NM2M  * D_NM2M)
#define A_NM2ML	 (D_NM2ML * D_NM2ML)
#define A_NM2FT	 (D_NM2FT * D_NM2FT)
#define A_NM2AC	 (A_NM2KM * 247.1054)

static const UInt16 coord_dec = 4;

// see .rcp
static const char *TRACK_DISTUNIT[] =
	{
		"m", "km", "nm", "ft", "ml"
	};
static const double TRACK_DIST[] =
	{
		D_NM2M, D_NM2KM, 1., D_NM2FT, D_NM2ML
	};

// see .rcp
static const char *TRACK_AREAUNIT[] =
	{
		"m" "km" "ha" "acre" "ml" "ft"
	};
static const double TRACK_AREA[] =
	{
		A_NM2M, A_NM2KM, A_NM2HA, A_NM2AC, A_NM2ML, A_NM2FT
	};


// "meter" "feet"
static Char *UNIT_ALT[]  = { "m", "ft" };

// "meter" "feet"
static Char *UNIT_OFFSET[]  = { "m", "ft" };

// "metric" "nautical" "statute" "golf"
static Char *UNIT_DIST[] = { "km", "nm", "mi", "yd" };

// "metric" "nautical" "statute" "golf"
static Char *UNIT_SPEED[] = { "km/h", "kn", "mph", "mph" };

static Boolean has_prefs;
static CAppPrefs prefs;

static Char *string_lat;
static Char *string_lon;
static Char *string_lat_s;
static Char *string_lon_s;
static Char *string_north;
static Char *string_east;
static Char *string_north_s;
static Char *string_east_s;
static Char *string_qth;

void PrefsInit()
{
	UInt16 prefs_size;
	Int16 code;

	string_lat		= UtilGetResourceString(StringLAT);
	string_lon		= UtilGetResourceString(StringLON);
	string_lat_s	= UtilGetResourceString(StringLATSHORT);
	string_lon_s	= UtilGetResourceString(StringLONSHORT);
	string_north	= UtilGetResourceString(StringNORTHING);
	string_east		= UtilGetResourceString(StringEASTING);
	string_north_s	= UtilGetResourceString(StringNORTHINGSHORT);
	string_east_s	= UtilGetResourceString(StringEASTINGSHORT);
	string_qth		= UtilGetResourceString(StringMAIDENHEAD);

	prefs_size = sizeof(CAppPrefs);
	MemSet(&prefs, prefs_size, 0);

	has_prefs = true;
	code = PrefGetAppPreferences(CGCreator, 0, 
                              (void*)&prefs, &prefs_size, true);

	if((prefs_size != sizeof(CAppPrefs)) || (code == noPreferenceFound))
	{
		prefs.start_form = GPSForm;
		prefs.time_offset = TimeGetDefOffset();
		prefs.port_index = SerialGetDefPort();
		prefs.port_speed_index = SerialGetDefSpeed();
		prefs.format_distspeed = 0;
		prefs.format_alt = 0;
		prefs.format_coord = 1;
		prefs.useNMEAAltWGS84 = true;
		prefs.correctWGS84 = false;
		prefs.drift = 7;
		prefs.maxspeed = 180;
		prefs.track_side_inside = true;
		prefs.track_offset_m = 0;
		prefs.track_offset_c = 0;
		prefs.track_offset_c_l = 1;
		prefs.track_interval_index= 0;
		prefs.track_format_dist= 0;
		prefs.track_format_area= 0;
		prefs.track_beep = 1;
		prefs.track_last_dist = 0;
		prefs.track_last_area = 0;
		prefs.track_has_last_dist = false;
		prefs.track_has_last_area = false;
		prefs.sat_data_doastro = true;
		prefs.dist_f = 0;
		prefs.max_speed = 0;
		prefs.min_alt = 0;
		prefs.max_alt = 0;
		prefs.show_rawdata = false; // not used anymore
		prefs.compass_fixed = true;
		prefs.help_font  = 0;

		prefs.marker_idx = -1;
		prefs.marker_category = dmAllCategories;
		prefs.track_category  = dmAllCategories;
		prefs.marker_category = dmAllCategories;
		StrCopy(prefs.track_name, CGDBTrack);
		StrCopy(prefs.map_name,   CGDBMap);
		MemSet(&prefs.btaddress, btLibDeviceAddressSize, 0);

		has_prefs = false;
    }

	MarkerInit();

	if(UtilIsOS(4, 0))
		prefs.time_offset = TimeGetDefOffset();

	TimeSetCurrentOffset(prefs.time_offset);
}

void PrefsFinish()
{
	MarkerFinish();
	UtilFree(string_lat);
	UtilFree(string_lon);
	UtilFree(string_lat_s);
	UtilFree(string_lon_s);
	UtilFree(string_north);
	UtilFree(string_east);
	UtilFree(string_north_s);
	UtilFree(string_east_s);
	UtilFree(string_qth);
}

void PrefsSave()
{
	Int16 pref_version = 1;
	PrefSetAppPreferences(CGCreator, 0, pref_version, (void*)&prefs,
							sizeof(CAppPrefs), true);
}

CAppPrefs *PrefsGetPrefs()
{
	return(&prefs);
}

Boolean PrefsHasPrefs()
{
	return(has_prefs);
}

void PrefsSetPrefs(Boolean b)
{
	has_prefs = b;
}

UInt16 PrefsGetDec()
{
	return(coord_dec);
}

void PrefsUpdateLabelAlt(FormPtr form, UInt16 id,
							Int32 alt, char c,
							Int32 alt84, char c84,
							double lat, double lon)
{
	Char *unit, *s;
	Unit from;

	from = (c == 'M' ? METRIC : STATUTE);
	PrefsAltToPref(from, &alt, &unit);
	if(c84 != '*')
	{
		if(prefs.useNMEAAltWGS84 == true)
		{
			if(c84 != NULL)
			{
				from = (c84 == 'M' ? METRIC : STATUTE);
				PrefsAltToPref(from, &alt84, &s);
				alt -= alt84;
			}
		}
		if(prefs.correctWGS84 == true)
		{
			alt84 = (Int32)GetGeoidalHeightRad(lat, lon);
			from = 'M';
			PrefsAltToPref(from, &alt84, &s);
			alt -= alt84;
		}
	}
	UtilUpdateSingle(form, id, alt, unit);
}

void PrefsAltToPref(Unit from, Int32 *val, Char **unit)
{
	if(unit != NULL)
		*unit = UNIT_ALT[prefs.format_alt];

	if(from != prefs.format_alt)
	{
		if(from == 0) // metric
			*val *= D_M2FT;
		else
			*val /= D_M2FT;
	}
}

void PrefsUpdateLabelNautic(FormPtr form, UInt16 id, double value)
{
	Char *unit;

	PrefsNauticToPref(&value, &unit);
	UtilUpdateSingle(form, id, value, unit);
}

void PrefsNauticToPref(double *val, Char **unit)
{
	if(unit != NULL)
		*unit = UNIT_DIST[prefs.format_distspeed];

	if(prefs.format_distspeed == METRIC)
		*val *= D_NM2KM;
	if(prefs.format_distspeed == STATUTE)
		*val *= D_NM2ML;
	if(prefs.format_distspeed == GOLF)
		*val *= D_NM2YD;
}

void PrefsUpdateLabelSpeed(FormPtr form, UInt16 id, double value)
{
	Char *unit;

	PrefsSpeedToPref(&value, &unit);
	UtilUpdateDoubleUnit(form, id, value, 1, unit);
}

// from knots
void PrefsSpeedToPref(double *val, Char **unit)
{
	if(unit != NULL)
		*unit = UNIT_SPEED[prefs.format_distspeed];

	if(prefs.format_distspeed == METRIC)
		*val *= D_NM2KM;
	if(prefs.format_distspeed == STATUTE)
		*val *= D_NM2ML;
	if(prefs.format_distspeed == GOLF)
		*val *= D_NM2ML;
}

void PrefsUpdateLabelOffset(FormPtr form, UInt16 id)
{
	Char *unit = (prefs.format_distspeed == METRIC) ?
					UNIT_OFFSET[0] : UNIT_OFFSET[1];
	UtilUpdateString(form, id, unit);
}

double PrefsOffsetToNM(double offset)
{
	offset /= D_NM2KM;
	if(prefs.format_distspeed != METRIC)
		offset /= D_M2FT;
	return(offset);
}

void PrefsUpdateLabelLatLon(FormPtr form,
							UInt16 lat_id, UInt16 lon_id,
							Boolean is_short)
{
	switch(prefs.format_coord)
	{
		case 4: /* UTM */
		case 6: /* OSGB */
			if(is_short == false)
			{
				UtilUpdateString(form, lat_id, string_north);
				UtilUpdateString(form, lon_id, string_east);
			}
			else
			{
				UtilUpdateString(form, lat_id, string_north_s);
				UtilUpdateString(form, lon_id, string_east_s);
			}
			break;
		case 5: /* Maidenhead */
			UtilUpdateString(form, lat_id, string_qth);
			UtilUpdateString(form, lon_id, "");
			break;
		default:
			if(is_short == false)
			{
				UtilUpdateString(form, lat_id, string_lat);
				UtilUpdateString(form, lon_id, string_lon);
			}
			else
			{
				UtilUpdateString(form, lat_id, string_lat_s);
				UtilUpdateString(form, lon_id, string_lon_s);
			}
			break;
	}
}

void PrefsUpdateLabelCoord(FormPtr form,
							UInt16 lat_id, UInt16 lon_id,
							double lat_val, double lon_val)
{
	static char fbuf[MAX_LABEL_LENGTH];
	Int32 deg, min, sec, frac;
	Char c;

	switch(prefs.format_coord)
	{
		case 0:
				/* D.D */
				StrPrintF(fbuf, "%%c %%ld.%%0%uld", coord_dec+2);

				NMEAGetDeg_DD(lat_val, coord_dec+2, &deg, &min, &c, LATLON_LAT);
				UtilUpdateLabel(form, lat_id, fbuf, c, deg, min);

				NMEAGetDeg_DD(lon_val, coord_dec+2, &deg, &min, &c, LATLON_LON);
				UtilUpdateLabel(form, lon_id, fbuf, c, deg, min);
			break;
		case 2:
				/* D M' S'' */
				StrPrintF(fbuf, "%%c %%ld %%ld' %%ld''");

				NMEAGetDeg_DMS(lat_val, &deg, &min, &sec, &c, LATLON_LAT);
				UtilUpdateLabel(form, lat_id, fbuf, c, deg, min, sec);

				NMEAGetDeg_DMS(lon_val, &deg, &min, &sec, &c, LATLON_LON);
				UtilUpdateLabel(form, lon_id, fbuf, c, deg, min, sec);
			break;
		case 3:
				/* D M' S.S'' */
				StrPrintF(fbuf, "%%c %%ld %%ld' %%ld.%%0%uld''", coord_dec-2);

				NMEAGetDeg_DMSS(lat_val, coord_dec-2, &deg, &min, &sec, &frac, &c, LATLON_LAT);
				UtilUpdateLabel(form, lat_id, fbuf, c, deg, min, sec, frac);

				NMEAGetDeg_DMSS(lon_val, coord_dec-2, &deg, &min, &sec, &frac, &c, LATLON_LON);
				UtilUpdateLabel(form, lon_id, fbuf, c, deg, min, sec, frac);
			break;
		case 4:
			{
				Char zone[4];
				Int32 northing, easting;

				/* UTM */
				NMEAGetDeg_UTM(lat_val, lon_val, &northing, &easting, zone);
				UtilUpdateLabel(form, lat_id, "%ld %s", northing, zone);
				UtilUpdateLabel(form, lon_id, "%ld", easting);
			}
			break;
		case 5:
			{
				Char qth[7];

				/* Maidenhead */
				NMEAGetDeg_QTH(lat_val, lon_val, qth);
				UtilUpdateLabel(form, lat_id, "%s", qth);
				UtilUpdateLabel(form, lon_id, " ");
			}
			break;
		case 6:
			{
				Char gridsquare[3];
				Int32 northing, easting;

				/* OSGB */
				NMEAGetDeg_OSGB(lat_val, lon_val, &northing, &easting, gridsquare);
				UtilUpdateLabel(form, lat_id, "%ld %s", northing, gridsquare);
				UtilUpdateLabel(form, lon_id, "%ld", easting);
			}
			break;
		default:
		case 1:
				/* D M.M' */
				StrPrintF(fbuf, "%%c %%ld %%ld.%%0%uld'", coord_dec);

				NMEAGetDeg_DMM(lat_val, coord_dec, &deg, &min, &frac, &c, LATLON_LAT);
				UtilUpdateLabel(form, lat_id, fbuf, c, deg, min, frac);

				NMEAGetDeg_DMM(lon_val, coord_dec, &deg, &min, &frac, &c, LATLON_LON);
				UtilUpdateLabel(form, lon_id, fbuf, c, deg, min, frac);
			break;
	}
}

const Char *PrefsTrackDistUnit()
{
	return(TRACK_DISTUNIT[prefs.track_format_dist]);
}

const Char *PrefsTrackAreaUnit()
{
	return(TRACK_AREAUNIT[prefs.track_format_area]);
}

double PrefsTrackDistToPref(double val)
{
	return(val * TRACK_DIST[prefs.track_format_dist] / 1000.);
}

double PrefsTrackAreaToPref(double val)
{
	return(val * TRACK_AREA[prefs.track_format_area] / 1000000.);
}

Err PrefsGetBTDevice()
{
	Err err;
	UInt16 btlibref;

	if(SysLibFind(btLibName, &btlibref))
	{
		err = SysLibLoad(sysFileTLibrary, sysFileCBtLib, &btlibref);
		if(err != errNone)
			return(err);
	}

	err = BtLibOpen(btlibref, false);
	if(err == errNone)
	{
		err = BtLibDiscoverSingleDevice(btlibref, NULL, NULL, 0, &(prefs.btaddress), false, false);
		if(err == errNone)
		{
			err = BtLibClose(btlibref);
		}
		BtLibClose(btlibref);
	}

	return(err);
}

