/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include "main.h"
#include "forms.h"
#include "util.h"
#include "nmea.h"
#include "draw.h"
#include "astro.h"
#include "serial.h"
#include "prefs.h"
#include "resource.h"
#include "def.h"

static void GPSHandler(FormPtr form, nmea_type *nmea) CODE3_SECTION;
static Boolean IsFixSat(nmea_type *nmea, Int32 prn) CODE3_SECTION;
static void InitImage(FormPtr form) CODE3_SECTION;
static void DrawImage() CODE3_SECTION;
static void SizeForm(FormPtr form) CODE3_SECTION;

static WinHandle image = NULL;
static Boolean form_ready = false;
static UInt16 title_height;
static SkyObjectData data;
static UInt16 font_off;
static RectangleType form_rect;

Boolean SatDataFormHandler(EventPtr event)
{
	CAppPrefs *prefs;
	Boolean handled = false;
	FormPtr	form = FrmGetActiveForm();

	switch(event->eType)
	{
		case frmOpenEvent:
			form_ready = false;
			title_height = 16;
			font_off     = 3;
			if(UtilIsVGA())
			{
				VgaFormModify(form, vgaFormModify160To240);
				title_height = VgaGetFrmTitleHeight()+1;
			}
			if(UtilIsResizeableNoDia())
			{
				font_off = 4;
				SizeForm(form);
			}
			FrmDrawForm(form);
			if(image == NULL)
				InitImage(form);
			prefs = PrefsGetPrefs();
			UtilSetControlValue(form, SatDataDoAstroID, prefs->sat_data_doastro);
			form_ready = true;
			handled = true;
			break;
		case frmUpdateEvent:
			FrmDrawForm(form);
			DrawImage();
			handled = true;
			break;
		case frmCloseEvent:
			form_ready = false;
			if(image != NULL)
			{
				WinDeleteWindow(image, false);
				image = NULL;
			}
			break;
		case winDisplayChangedEvent:
		case displayExtentChangedEvent:
			if(form_ready == true)
			{
				SizeForm(form);
				FrmDrawForm(form);
				DrawImage();
			}
			handled = true;
			break;
		case ctlSelectEvent:
			switch(event->data.ctlSelect.controlID)
			{
				case SatDataDoAstroID:
						prefs = PrefsGetPrefs();
						prefs->sat_data_doastro =
							UtilGetControlValue(form, SatDataDoAstroID);
					break;
			}
			break;
		case GpsNmeaEvent:
				GPSHandler(form, CASTNMEA(event));
			break;
		default:
			break;
	}
	return(handled);
}

static void SizeForm(FormPtr form)
{
	RectangleType curBounds, displayBounds;
	Coord diff;

	UtilGetFrmBounds(form, &curBounds);
	UtilGetDpyBounds(&displayBounds);

	diff = displayBounds.extent.y - (curBounds.topLeft.y + curBounds.extent.y);

	if(diff)
	{
		curBounds.extent.y += diff;

		WinSetBounds(FrmGetWindowHandle(form), &curBounds);
		WinEraseRectangle(&curBounds, 0);

		if(image != NULL)
		{
			WinDeleteWindow(image, false);
			image = NULL;
			InitImage(form);
		}
		UtilMoveObject(form, SatDataDoAstroID, diff, true);
	}
}

static void GPSHandler(FormPtr form, nmea_type *nmea)
{
	Int16 i, y, step;
	FontID old_font;
	CAppPrefs *prefs;
	Char buf[16];
	CSYS_INIT

	MainSetGPSTitle(form);

	if((nmea->has_gsv && nmea->is_new_gsv) ||
		(!nmea->has_gsv && nmea->has_gga && nmea->is_new_gga))
	{
		if(UtilIsVGA())
			old_font = FntSetFont(VgaBaseToVgaFont(stdFont));
		else
			old_font = FntSetFont(stdFont);

		y = title_height;
		step = (Int16) ((form_rect.extent.y - title_height) / MAXSAT);
		for(i = 0; i < MAXSAT; i++)
		{
			satdata_type *sat = &nmea->sat_data[i];

			if(sat->prn == 0)
			{
				StrPrintF(buf, "--: --  ");
			}
			else
			{
				if(sat->signal)
					StrPrintF(buf, "%02ld: %02ld", sat->prn, sat->signal);
				else
					StrPrintF(buf, "%02ld: --  ", sat->prn);
			}
			WinDrawChars(buf, StrLen(buf), 0, y);

			y += step;
		}

		DrawImage();

		if(UtilIsVGA())
			FntSetFont(VgaBaseToVgaFont(boldFont));
		else
			FntSetFont(boldFont);

		CSYS_DOUBLE

		prefs = PrefsGetPrefs();
		if((prefs->sat_data_doastro) && (nmea->fix_quality > 0))
			AstroDraw(data.x, data.y, data.r, nmea);

		data.wants_color = 0;
		data.name = buf;
		data.bitmap = 0;
		for(i = 0; i < MAXSAT; i++)
		{
			satdata_type *sat = &nmea->sat_data[i];

			if(sat->prn)
			{
				StrPrintF(buf, "%02ld", sat->prn);
				data.inv = IsFixSat(nmea, sat->prn);
				data.elevation = sat->elevation;
				data.azimuth   = sat->azimuth;
				DrawSkyObject(&data);
			}
		}

		// draw direction indicator
		data.name = NULL;
		data.elevation = 5;
		data.azimuth = nmea->course;
		data.wants_color = 1;
		data.red   = 255;
		data.green = 0;
		data.blue  = 0;
		DrawSkyObject(&data);

		// draw target bearing
		if(MarkerHasDestination() == true)
		{
			data.name = "X";
			data.elevation = 5;
			data.azimuth = (Int16)MarkerGetCourse(nmea);
			data.wants_color = 1;
			data.red   = 0;
			data.green = 196;
			data.blue  = 0;
			DrawSkyObject(&data);
		}

		FntSetFont(old_font);

		CSYS_RESET

		nmea->is_new_gsv = false;
	}
}

static void InitImage(FormPtr form)
{
	Err err;
	Int16 w, h, d, fw;
	CSYS_INIT

	UtilGetFrmBounds(form, &form_rect);
	fw = FntCharsWidth("66", 2);
	w = form_rect.extent.x - font_off*fw;
	h = form_rect.extent.y - title_height - FntCharHeight() - 4;
	d = MIN(w, h);
	data.r = d >> 1;
	data.x = ((w-d) >> 1) + form_rect.topLeft.x + font_off*fw + data.r;
	data.y = ((h-d) >> 1) + title_height + data.r;

	CSYS_DOUBLE

	if(UtilIsHires() == true)
	{
		data.x <<= 1;
		data.y <<= 1;
		data.r <<= 1;
	}

	image = WinCreateOffscreenWindow(data.r << 1, data.r << 1,
										nativeFormat, &err);
	if(err == errNone)
	{
		RectangleType rect;
		WinHandle cw;

		cw = WinSetDrawWindow(image);
		DrawSky(data.r, data.r, data.r);
		WinSetDrawWindow(cw);
		RctSetRectangle(&rect, 0, 0, data.r << 1, data.r << 1);
		WinCopyRectangle(image, cw,
						&rect, data.x-data.r, data.y-data.r, winPaint);
	}
	else
	{
		image = NULL;
		DrawSky(data.x, data.y, data.r);
	}

	CSYS_RESET
}

static void DrawImage()
{
	CSYS_INIT
	CSYS_DOUBLE

	if(image != NULL)
	{
		RectangleType rect;

		RctSetRectangle(&rect, 0, 0, data.r << 1, data.r << 1);
		WinCopyRectangle(image, WinGetDisplayWindow(),
						&rect, data.x-data.r, data.y-data.r, winPaint);
	}
	else
	{
		DrawSky(data.x, data.y, data.r);
	}

	CSYS_RESET
}

static Boolean IsFixSat(nmea_type *nmea, Int32 prn)
{
	UInt16 i;

	for(i = 0; i < MAXSAT; i++)
	{
		if(nmea->sat_prn[i] == prn)
			return(true);
	}
	return(false);
}

