/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <SerialMgrOld.h>
#include <BtLib.h>
#include <BtCommVdrv.h>
#include "main.h"
#include "log.h"
#include "util.h"
#include "prefs.h"
#include "serial.h"
#include "ique.h"
#include "resource.h"

#define BUFMAX 1024
#define RCVBUFMAX 1024
#define ERRTIME 3
#define ERRMAX  10

static void SerialInitDevices() CODE2_SECTION;
static void SerialDrain() CODE2_SECTION;
static void HandleLineError() CODE2_SECTION;

static Char DEF_PORT[]  = "built-in";
static Char IQUE_PORT[] = "iQue";
#define IQUE_ID			'iQue'

static UInt32 has_new_sm;
static Boolean is_running, has_rcvbuffer, first_open;
static UInt16 ser_ref;
static Char **def_item;
static UInt16 def_cnt;
static UInt32 *serports;
static UInt16 defdev;
static UInt16 bufmax, rcvbufmax;

static char *buffer;
static char *rcv_buffer;
static UInt32 cport;
static UInt32 cspeed;
static UInt32 dlen;

static Char *SPEED[] =	{
							"1200", "2400", "4800", "9600",
							"19200", "38400", "57600", "115200",
							NULL
						};

Boolean SerialInit()
{
	Err err;

	is_running	= false;
	has_rcvbuffer  = false;
	first_open	= true;
	def_item	= NULL;
	serports	= NULL;
	def_cnt		= 0;
	defdev		= 0;
	rcv_buffer	= NULL;
	buffer		= NULL;
	ser_ref		= 0;
	cport		= 0;
	cspeed		= 0;

	err = FtrGet(sysFileCSerialMgr, sysFtrNewSerialVersion, &has_new_sm);
	if(err != errNone)
		has_new_sm = 0;

	if((has_new_sm > 1) && (UtilIsOS(4,0)) == false)
		has_new_sm = 1;

	rcvbufmax = RCVBUFMAX;
	bufmax    = BUFMAX;

	rcv_buffer = UtilMalloc(rcvbufmax);
	if(!rcv_buffer)
		return(false);
	MemSet(rcv_buffer, rcvbufmax, 0L);

	buffer = UtilMalloc(bufmax);
	if(!buffer)
		return(false);
	MemSet(buffer, bufmax, 0L);

	dlen = 0L;

	if(!has_new_sm)
	{
		if(SysLibFind("Serial Library", &ser_ref) != 0)
		{
			ser_ref = 0;
			return(false);
		}
		return(true);
	}

	SerialInitDevices();

	return(true);
}

void SerialFinish()
{
	UInt16 i;

	if(is_running)
		SerialClosePort();

	if(def_item != NULL)
	{
		for(i = 0; i < def_cnt; i++)
			if(def_item[i] != NULL)
				UtilFree(def_item[i]);
		UtilFree(def_item);
		def_item = NULL;
	}
	if(serports)
		UtilFree(serports);
	if(buffer)
		UtilFree(buffer);
	if(rcv_buffer)
		UtilFree(rcv_buffer);
	is_running = false;

	LogClose();
}

void SerialSetDebug(Boolean db)
{
	if(db == true)
		LogOpen();
	else
		LogClose();
}

Boolean SerialIsDebug()
{
	return(LogIsShowLog());
}

Boolean SerialIsIque()
{
	return(cport == IQUE_ID);
}

Boolean SerialIsRunning()
{
	return(is_running);
}

static void SerialInitDevices()
{
	UInt16 cnt;
	Char **c;
	SerialGetDevices(&cnt, &c);
}

UInt16 SerialGetDefPort()
{
	return(defdev);
}

UInt16 SerialGetDefSpeed()
{
	Int16 i = 0;

	while((SPEED[i] != NULL) && (StrCompare(SPEED[i], "4800") != 0))
		i++;

	return(i);
}

void SerialGetSpeeds(UInt16 *cnt, Char ***items)
{
	UInt16 i = 0;

	while(SPEED[i] != NULL)
		i++;
	*cnt = i;
	*items = SPEED;
}

UInt32 SerialGetMaxSpeed(UInt16 devId)
{
	if(has_new_sm)
	{
		DeviceInfoType di;
		SrmGetDeviceInfo(devId, &di);
		return(di.serDevMaxBaudRate);
	}
	else
		return(115200);
}

void SerialGetDevices(UInt16 *cnt, Char ***items)
{
	if(def_item == NULL)
	{
		if(has_new_sm)
		{
			UInt16 i, tcnt, tcnt2;

			// how many serial devices do we have?
			SrmGetDeviceCount(&tcnt);
			tcnt2 = tcnt;

			// and one for iQues
			if(UtilIsIque())
				tcnt++;

			serports = (UInt32*)UtilMalloc(tcnt * sizeof(UInt32));
			def_item = (Char**) UtilMalloc(tcnt * sizeof(Char*));
			def_cnt = 0;

			for(i = 0; i < tcnt2; i++)
			{
				char *t;
				UInt16 len;
				DeviceInfoType di;

				SrmGetDeviceInfo(i, &di);
				t   = (char*)&di.serDevCreator;
				len = StrLen(di.serDevPortInfoStr) + 10;
				def_item[def_cnt] = UtilMalloc((UInt32)len);

				StrPrintF(def_item[def_cnt], "\"%s\" [%c%c%c%c]",
							di.serDevPortInfoStr,
							t[0], t[1], t[2], t[3]);

				serports[def_cnt] = di.serDevCreator;
				if(di.serDevFtrInfo & serDevCradlePort)
					defdev = def_cnt;
				def_cnt++;
			}

			if(UtilIsIque())
			{
				UInt16 len;
				char *t;

				serports[def_cnt] = IQUE_ID;
				t   = (char*)&serports[def_cnt];
				len = StrLen(IQUE_PORT) + 10;
				def_item[def_cnt] = UtilMalloc((UInt32)len);
				StrPrintF(def_item[def_cnt], "\"%s\" [%c%c%c%c]",
							IQUE_PORT,
							t[0], t[1], t[2], t[3]);
				defdev = def_cnt;
				def_cnt++;
			}
		}
		else
		{
			def_cnt		= 1;
			def_item	= (Char**)UtilMalloc(def_cnt*sizeof(Char*));
			def_item[0]	= UtilStrDup(DEF_PORT);
		}
	}

	*cnt = def_cnt;
	*items = def_item;
}

Err SerialOpenPortByIndex(Int16 portindex, Int16 speedindex)
{
	return(SerialOpenPort(serports[portindex], StrAToI(SPEED[speedindex])));
}

Err SerialOpenPort(UInt32 port, UInt32 speed)
{
	Err err = errNone;
	UInt32 max_speed;

	if((first_open == true) && (port == 'u550') && UtilIsOS(4, 0))
	{
		// hack for OS4 on HandEra
		first_open = false;
		SerialOpenPort('u8vc', 4800);
		SerialClosePort();
	}

	cport  = port;
	cspeed = speed;

	if(cport == IQUE_ID)
	{
		err = IqueOpenGPS();
		if(err == errNone)
			is_running = true;
		return(err);
	}

	max_speed = SerialGetMaxSpeed(port);
	if(cspeed > max_speed)
		cspeed = max_speed;

	switch(has_new_sm)
	{
		case 0:
			err = SerOpen(ser_ref, 0, cspeed);
			break;
		case 1:
			err = SrmOpen(port, cspeed, &ser_ref);
			break;
		default:
			{
				SrmOpenConfigType srm_cfg;

				MemSet(&srm_cfg, sizeof(SrmOpenConfigType), 0);
				srm_cfg.function     = serFncUndefined;
				srm_cfg.baud         = cspeed;
				srm_cfg.drvrDataP    = NULL;
				srm_cfg.drvrDataSize = 0;

				if(port == sysFileCVirtRfComm)
				{
					CAppPrefs *prefs;
					BtVdOpenParams btParams;

					prefs = PrefsGetPrefs();
					MemSet(&btParams, sizeof(BtVdOpenParams), 0);
					btParams.role = btVdClient;
					btParams.u.client.remoteDevAddr.address[0] = prefs->btaddress.address[0];
					btParams.u.client.remoteDevAddr.address[1] = prefs->btaddress.address[1];
					btParams.u.client.remoteDevAddr.address[2] = prefs->btaddress.address[2];
					btParams.u.client.remoteDevAddr.address[3] = prefs->btaddress.address[3];
					btParams.u.client.remoteDevAddr.address[4] = prefs->btaddress.address[4];
					btParams.u.client.remoteDevAddr.address[5] = prefs->btaddress.address[5];

					srm_cfg.baud         = 0;
					srm_cfg.drvrDataP    = (MemPtr)&btParams;
					srm_cfg.drvrDataSize = sizeof(BtVdOpenParams);
				}

				err = SrmExtOpen(port, &srm_cfg, sizeof(SrmOpenConfigType), &ser_ref);
			}
			break;
	}

	if(err == errNone)
	{
		if(has_new_sm)
		{
			UInt16 op   = srmCtlSetFlags;
			UInt16 size = sizeof(UInt32);
			UInt32 val  =	srmSettingsFlagBitsPerChar8		|
							srmSettingsFlagStopBits1		|
							srmSettingsFlagRTSAutoM			|
							srmSettingsFlagFlowControlIn;
			SrmControl(ser_ref, op, &val, &size);
		}
		else
		{
			SerSettingsType settings;
			SerGetSettings(ser_ref, &settings);
			settings.flags =	serSettingsFlagBitsPerChar8	|
								serSettingsFlagStopBits1	|
								serSettingsFlagRTSAutoM;
			SerSetSettings(ser_ref, &settings);
		}
		is_running = true;
	
		if(has_new_sm)
			err = SrmSetReceiveBuffer(ser_ref, rcv_buffer, rcvbufmax); 
		else
			err = SerSetReceiveBuffer(ser_ref, rcv_buffer, rcvbufmax); 

		if(err == errNone)
			has_rcvbuffer = true;
	}

	return(err);
}

Err SerialClosePort()
{
	Err err = errNone;

	if(!is_running)
		return(0);

	if(cport == IQUE_ID)
	{
		is_running = false;
		return(IqueCloseGPS());
	}

	if(has_new_sm)
	{
		if(has_rcvbuffer)
			SrmSetReceiveBuffer(ser_ref, NULL, 0); 
		if(ser_ref != 0)
			err = SrmClose(ser_ref);
	}
	else
	{
		if(has_rcvbuffer)
			SerSetReceiveBuffer(ser_ref, NULL, 0); 
		if(ser_ref != 0)
			err = SerClose(ser_ref);
	}
	is_running = false;
	return(err);
}

/* never call this before SerialOpen, we need cport and cspeed first! */
Boolean SerialNextLine(char *ser_buf)
{
	UInt32 n_start, n_end;
	char *cp_start, *cp_end;

	if(is_running == false)
	{
		if(SerialOpenPort(cport, cspeed) != errNone)
			return(false);
	}

	if(LogIsShowLog() == true)
	{
		if(has_new_sm)
			SrmReceiveFlush(ser_ref, 0);
		else
			SerReceiveFlush(ser_ref, 0);

		return(LogNextLine(ser_buf));
	}

	SerialDrain();

	/* check for beginning of next sentence */
	cp_start = buffer;
	n_start  = 0;
	while((n_start < dlen) && (*cp_start != '$'))
	{
		cp_start++;
		n_start++;
	}

	if(n_start > NMEA_MAX_LENGTH)
	{
		/* buffer contains crap, throw away */
		dlen = 0;
		return(false);
	}

	/* find end of sentence */
	cp_end = cp_start;
	n_end  = n_start;
	while((n_end < dlen) && (*cp_end != chrLineFeed))
	{
		cp_end++;
		n_end++;
	}

	/* if found, make a sentence */
	if(n_end < dlen)
	{
		UInt32 len;

		len = n_end - n_start;
		if(len >= NMEA_MAX_LENGTH)
		{
			/* throw away big message and the rest */
			dlen = 0;
			return(false);
		}
		/* copy NMEA sentence */
		MemMove(ser_buf, cp_start, len-1);
		ser_buf[len-1] = '\0';
		dlen -= n_end + 1;
		MemMove(buffer, cp_end+1, dlen);
		return(true);
	}

	return(false);
}

static void SerialDrain()
{
	UInt32 nbytes;
	Err err = errNone;

	/* check how many bytes are available */
	if(has_new_sm)
		err = SrmReceiveCheck(ser_ref, &nbytes);
	else
		err = SerReceiveCheck(ser_ref, &nbytes);

	if(err == serErrLineErr)
	{
		HandleLineError();
		return;
	}

	/* receive data */
	if(nbytes > 0)
	{
		/* restrict to buffer length */
		if(dlen + nbytes > bufmax)
			nbytes = bufmax - dlen;

		if(has_new_sm)
			SrmReceive(ser_ref, &buffer[dlen], nbytes, -1, &err);
		else
			SerReceive(ser_ref, &buffer[dlen], nbytes, -1, &err);

		if(err == serErrLineErr)
		{
			HandleLineError();
			return;
		}

		dlen += nbytes;
	}
}

static void HandleLineError()
{
	dlen = 0;

	if(has_new_sm)
	{
		SrmClearErr(ser_ref);
		SrmReceiveFlush(ser_ref, 0);
	}
	else
	{
		SerClearErr(ser_ref);
		SerReceiveFlush(ser_ref, 0);
	}
}

