/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include "main.h"
#include "forms.h"
#include "util.h"
#include "nmea.h"
#include "draw.h"
#include "marker.h"
#include "prefs.h"
#include "resource.h"
#include "def.h"

static void FillTarget(FormPtr form) CODE2_SECTION;
static void SizeForm(FormPtr form) CODE2_SECTION;
static void ResizeCompass(FormPtr form) CODE2_SECTION;
static void GPSHandler(FormPtr form, nmea_type *nmea) CODE2_SECTION;

static Boolean compass_dwn;
static meter_type compass;
static Boolean form_ready = false;
static Int16 marker_index;

Boolean TargetFormHandler(EventPtr event)
{
	Boolean handled = false;
	FormPtr	form = FrmGetActiveForm();

	switch(event->eType)
	{
		case frmOpenEvent:
			form_ready = false;
			marker_index = -1;
			if(UtilIsVGA())
				VgaFormModify(form, vgaFormModify160To240);
			if(UtilIsResizeableNoDia())
				SizeForm(form);
			if(MarkerHasDestination() == false)
				UtilShowObject(form, TargetInfoID, false);
			FrmDrawForm(form);
			MeterInit(&compass, TYPE_COMPASS);
			MeterSetDoublebuffered(&compass, true);
			MeterSetFixed(&compass, PrefsGetPrefs()->compass_fixed);
			compass.wants_dval = false;
			ResizeCompass(form);
			MeterDraw(&compass);
			FillTarget(form);
			form_ready = true;
			handled = true;
			break;
		case frmCloseEvent:
			form_ready = false;
			MeterFinish(&compass);
			break;
		case frmUpdateEvent:
			FrmDrawForm(form);
			FillTarget(form);
			UtilUpdateLabel(form, TargetDist, " ");
			UtilUpdateLabel(form, TargetCourse, " ");
			MeterErase(&compass);
			MeterDraw(&compass);
			handled = true;
			break;
		case winDisplayChangedEvent:
		case displayExtentChangedEvent:
			if(form_ready == true)
			{
				SizeForm(form);
				FrmDrawForm(form);
				ResizeCompass(form);
				MeterDraw(&compass);
			}
			handled = true;
			break;
		case penDownEvent:
			if(MeterPenInside(&compass, event))
				compass_dwn = true;
			else
				compass_dwn = false;
			break;
		case penUpEvent:
			if(compass_dwn && MeterPenInside(&compass, event))
			{
				CAppPrefs *prefs = PrefsGetPrefs();
				prefs->compass_fixed = !prefs->compass_fixed;
				MeterSetFixed(&compass, prefs->compass_fixed);
			}
			compass_dwn = false;
			break;
		case ctlSelectEvent:
			switch(event->data.ctlSelect.controlID)
			{
				case TargetInfoID:
					MarkerEdit(PrefsGetPrefs()->marker_idx, false);
					handled = true;
					break;
			}
			break;
		case GpsNmeaEvent:
				GPSHandler(form, CASTNMEA(event));
			break;
		default:
			break;
	}
	return(handled);
}

static void FillTarget(FormPtr form)
{
	marker_type marker;
	CAppPrefs *prefs;

	prefs = PrefsGetPrefs();

	if(prefs->format_coord == 5) // QTH
	{
		UtilShowObject(form, TargetLabel1ID, false);
		UtilShowObject(form, TargetLabel2ID, false);
		UtilUpdateLabel(form, TargetLonTarget,  " ");
		UtilUpdateLabel(form, TargetLonCurrent, " ");
	}
	else
	{
		UtilShowObject(form, TargetLabel1ID, true);
		UtilShowObject(form, TargetLabel2ID, true);
		UtilUpdateLabel(form, TargetLonTarget,  "-");
		UtilUpdateLabel(form, TargetLonCurrent, "-");
	}
	UtilUpdateLabel(form, TargetLatTarget, "-");

	marker_index = prefs->marker_idx;
	if(MarkerGetMarker(marker_index, &marker) == true)
	{
		PrefsUpdateLabelCoord(form,
								TargetLatTarget, TargetLonTarget,
								marker.lat, marker.lon);
	}
}

static void SizeForm(FormPtr form)
{
	RectangleType curBounds, displayBounds;
	Coord diff;

	UtilGetFrmBounds(form, &curBounds);
	UtilGetDpyBounds(&displayBounds);

	diff = displayBounds.extent.y - (curBounds.topLeft.y + curBounds.extent.y);

	if(diff)
	{
		curBounds.extent.y += diff;
		WinSetBounds(FrmGetWindowHandle(form), &curBounds);
		UtilMoveObject(form, TargetDist, diff, true);
		UtilMoveObject(form, TargetCourse, diff, true);
	}
}

static void ResizeCompass(FormPtr form)
{
	RectangleType field_rect, rect;
	Coord comp_x, comp_y, comp_w, w, h;

	UtilGetFrmBounds(form, &rect);
	w = rect.extent.x;
	h = rect.extent.y;
	FrmGetObjectBounds(form, FrmGetObjectIndex(form, TargetLonCurrent),
							&field_rect);

	if(UtilIsVGA())
		field_rect.extent.y += 4;

	comp_y = field_rect.topLeft.y + field_rect.extent.y+1;
	comp_w = MIN((h - comp_y - 2), (w-1));
	comp_x = (w - comp_w);
	MeterSetSize(&compass, comp_x, comp_y, comp_w, comp_w);
}

static void GPSHandler(FormPtr form, nmea_type *nmea)
{
	MainSetGPSTitle(form);

	MeterUpdate(&compass, nmea);

	if(nmea->is_new_gga || nmea->is_new_nexrmc)
	{
		if(MarkerHasDestination() == true)
		{
			UInt32 odo, dist;

			dist = compass.dval;
			odo  = (UInt32)(dist/1000.);
			UtilUpdateLabel(form, TargetDist, "%ld.%03ld  ",
							odo, dist-(odo*1000));
			UtilUpdateSingle(form, TargetCourse, compass.aval2, "");
		}

		PrefsUpdateLabelCoord(form,
								TargetLatCurrent, TargetLonCurrent,
								nmea->lat, nmea->lon);
		nmea->is_new_gga = false;
		nmea->is_new_nexrmc = false;
	}
}

