/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <VFSMgr.h>
#include <TextMgr.h>
#include "MathLib.h"
#include "track.h"
#include "main.h"
#include "nmea.h"
#include "dist.h"
#include "prefs.h"
#include "util.h"
#include "area.h"
#include "time.h"
#include "beam.h"
#include "pdouble.h"

static const Int16 INTERVAL[] =	{ 1, 2, 10, 60 };
#define COTOGPS_DIR "/PALM/cotoGPS/"
#define COTOGPS_DOCNAME "cotoGPS track"
#define COTOGPS_COMMENT "Written by cotoGPS"

typedef struct
{
	Char track_name[MAX_TRACK_NAME_LENGTH];
	Boolean is_capturing;
	UInt32 cnt;
	double prev_x, prev_y;
	dist_type dist;
	UInt32 time;
} tracksave_type;

static tracksave_type tsave;
static trackdraw_iter iter;
Char renamer[dmDBNameLength];

static UInt16 getFra(Int32 *v, double mult);
static UInt16 getFirstVolRef();
static double toDD(double v);
static Err WriteLine(FileRef fileRef, Boolean do_conv, Char *txt);

Err TrackGetTracks(LocalID **trackids, UInt16 *cnt)
{
	return(UtilGetDBList(CGCreator, CGDBTrackType, trackids, cnt));
}

Err TrackRemoveDB(Char *name)
{
	Err err;

	UtilDBRemove(name, &err);
	if(StrCompare(tsave.track_name, name) == 0)
	{
		// reset current db to new db (== null records)
		err = TrackNewDB(name);
	}

	return(err);
}

void TrackSaveInit(Char *trackname)
{
	tsave.prev_x = 0.;
	tsave.prev_y = 0.;
	tsave.time   = 0;

	if(	(trackname == NULL) ||
		(StrLen(trackname) >= MAX_TRACK_NAME_LENGTH) ||
		(StrLen(trackname) == 0))
	{
		SndPlaySystemSound(sndError);
		return;
	}

	StrCopy(tsave.track_name, trackname);
	tsave.cnt = TrackGetCount();
	if(tsave.cnt > 0)
	{
		DmOpenRef db;
		MemHandle mrec;

		db = UtilDBOpen(tsave.track_name, CGCreator, CGDBTrackType, dmModeReadOnly);
		if(db == NULL)
		{
			tsave.cnt = 0;
			return;
		}
		mrec = DmGetRecord(db, tsave.cnt-1);
		if(mrec != NULL)
		{
			trackpoint_type *pt;

			pt = (trackpoint_type *)MemHandleLock(mrec);

			DistInit(&tsave.dist, pt->lat, pt->lon, tsave.cnt);
			tsave.prev_x = pt->x;
			tsave.prev_y = pt->y;

			MemHandleUnlock(mrec);
			DmReleaseRecord(db, tsave.cnt-1, false);
		}
		DmCloseDatabase(db);
	}
	else
		DistInit(&tsave.dist, 0., 0., 0);
}

void TrackSaveFinish()
{
	StrCopy(PrefsGetPrefs()->track_name, tsave.track_name);
}

Err TrackNewDB(Char *trackname)
{
	Err err;

	DmOpenRef db = UtilDBOpen(trackname, CGCreator, CGDBTrackType, dmModeReadOnly);
	if(db != NULL)
		DmCloseDatabase(db);

	UtilDBSetUpdateFlag(trackname, &err);
	return(err);
}

Char *TrackGetName()
{
	return(tsave.track_name);
}

void TrackSavePosition(nmea_type *nmea, Boolean force)
{
	/* do not save point when speed==0, unless forced (e.g. key press) */
	if((force == false) && (nmea->speed_f == 0.))
		return;

	if((nmea->has_gps == true) && (nmea->fix_quality > 0))
	{
		trackpoint_type pt;
		Err err;

		pt.time = TimeGetGPSSeconds();
		if(pt.time == tsave.time)
			return;
		tsave.time = pt.time;
		pt.lat = nmea->lat;
		pt.lon = nmea->lon;
		pt.alt = (Int16)nmea->alt;
		pt.hdop = 0;
		pt.vdop = 0;
		pt.pdop = 0;
		if(nmea->has_gsa)
		{
			pt.hdop = getFra(&(nmea->hdop), 0);
			pt.vdop = getFra(&(nmea->vdop), 0);
			pt.pdop = getFra(&(nmea->pdop), 0);
		}
		else if(nmea->has_gga)
		{
			pt.hdop = getFra(&(nmea->gga_hdop), 0);
		}
		else if(nmea->has_nexrmc)
		{
			pt.hdop = getFra(&(nmea->hdop), 0);
		}
		pt.sat_tracked = (UInt8)nmea->sat_tracked;

		if(nmea->warning)
			pt.fix_quality = 0;
		else if(nmea->fix_quality == 2)
			pt.fix_quality = 4;
		else
		{
			if(nmea->has_gsa)
				pt.fix_quality = nmea->fix_dim;
			else
				pt.fix_quality = 3;
		}

		pt.speed = getFra(&(nmea->speed), 0.514856);

		pt.arc  = DistGetCourse(&tsave.dist, nmea->lat, nmea->lon)+M_PI_2;
		pt.dist = DistGetDistUpd(&tsave.dist, nmea->lat, nmea->lon);
		pt.x    = tsave.prev_x + pt.dist * cos(pt.arc);
		pt.y    = tsave.prev_y + pt.dist * sin(pt.arc);

		UtilDBSaveRecord(tsave.track_name, CGCreator, CGDBTrackType,
							&pt, sizeof(trackpoint_type), &err);
		tsave.prev_x = pt.x;
		tsave.prev_y = pt.y;
		tsave.cnt++;

		TrackSendEvent();

		if(PrefsGetPrefs()->track_beep)
			SndPlaySystemSound(sndError);
	}
	else
		SndPlaySystemSound(sndError);
}

UInt32 TrackGetCount()
{
	return(TrackGetCountByName(tsave.track_name));
}

UInt32 TrackGetCountByName(Char *dbname)
{
	return(UtilDBGetCountByName(dbname));
}

UInt32 TrackGetCountById(LocalID db_id)
{
	return(UtilDBGetCountById(db_id));
}

void TrackSetCapturing(Boolean b)
{
	tsave.is_capturing = b;
	TrackSendEvent();
}

Boolean TrackIsCapturing()
{
	return(tsave.is_capturing);
}

Int16 TrackGetSaveInterval()
{
	return(INTERVAL[PrefsGetPrefs()->track_interval_index]);
}

Boolean TrackCalc(double *distP, double *areaP)
{
	UInt32 cnt, i;
	double offset, length;
	double first_lat, first_lon, last_lat, last_lon, first_x, first_y;
	DmOpenRef db;
	CAppPrefs *prefs;
	area_type area;

	cnt = TrackGetCount();
	if(cnt == 0)
	{
		*distP = 0.;
		if(areaP != NULL)
			*areaP = 0.;
		return(true);
	}

	prefs = PrefsGetPrefs();
	offset = (double)prefs->track_offset_c;
	for(i = 0; i < prefs->track_offset_c_l; i++)
		offset /= 10.;
	offset += prefs->track_offset_m;

	if(prefs->track_side_inside == true)
		offset = -offset;

	db = UtilDBOpen(tsave.track_name, CGCreator, CGDBTrackType, dmModeReadOnly);
	if(db == NULL)
		return(false);

	if(areaP != NULL)
		AreaInit(&area, offset);

	first_lat = first_lon = last_lat = last_lon = 0.;
	length = first_x = first_y = 0.;
	for(i = 0; i < cnt; i++)
	{
		MemHandle mrec;
		trackpoint_type *pt;

		mrec = DmGetRecord(db, i);
		if(mrec == NULL)
			break;
		pt = (trackpoint_type *)MemHandleLock(mrec);

		if(i == 0)
		{
			first_x   = pt->x;
			first_y   = pt->y;
			first_lon = pt->lon;
			first_lat = pt->lat;
		}
		else
		{
			length += pt->dist;

			if(areaP != NULL)
				AreaAddPoint(&area, pt->dist, pt->x, pt->y);
		}
		last_lat = pt->lat;
		last_lon = pt->lon;

		MemHandleUnlock(mrec);
		DmReleaseRecord(db, i, false);
	}
	DmCloseDatabase(db);

	if(areaP != NULL)
	{
		double ll;
		dist_type dist;

		DistInit(&dist, first_lat, first_lon, 0);
		ll  = DistGetDist(&dist, last_lat, last_lon);

		AreaAddPoint(&area, ll, first_x, first_y);
		*areaP = AreaGetArea(&area);
		AreaFinish(&area);
		length += ll;
	}
	*distP = length;

	return(true);
}

void TrackDraw()
{
	UtilPopupForm(TrackDrawForm);
}

trackdraw_iter *TrackDrawInit()
{
	UInt32 cnt;

	cnt = TrackGetCount();
	if(cnt == 0)
		return(NULL);

	iter.db = UtilDBOpen(tsave.track_name, CGCreator, CGDBTrackType, dmModeReadOnly);
	if(iter.db == NULL)
		return(NULL);

	iter.curr = 0;
	iter.max  = cnt;
	iter.is_last = (iter.max == 0);

	iter.cache = (trackpoint2_type*)UtilMallocSilent(cnt*sizeof(trackpoint2_type));
	if(iter.cache != NULL)
	{
		UInt32 i;

		for(i = 0; i < cnt; i++)
		{
			trackpoint_type *pt;
			MemHandle mrec;

			mrec = DmGetRecord(iter.db, iter.curr);
			if(mrec == NULL)
			{
				UtilFree(iter.cache);
				iter.cache = NULL;
				break;
			}

			pt = (trackpoint_type *)MemHandleLock(mrec);

			iter.cache[i].x = pt->x;
			iter.cache[i].y = pt->y;
			iter.cache[i].alt = pt->alt;
			iter.cache[i].speed = pt->speed;

			MemHandleUnlock(mrec);
			DmReleaseRecord(iter.db, iter.curr, false);
			iter.curr++;
		}
		iter.curr = 0;
		DmCloseDatabase(iter.db);
	}

	return(&iter);
}

Boolean TrackDrawNext(trackdraw_iter *iter)
{
	trackpoint_type *pt;
	MemHandle mrec;

	if(iter->curr >= iter->max)
		return(false);

	if(iter->cache == NULL)
	{
		mrec = DmGetRecord(iter->db, iter->curr);
		if(mrec == NULL)
			return(false);

		pt = (trackpoint_type *)MemHandleLock(mrec);

		iter->p.x = pt->x;
		iter->p.y = pt->y;
		iter->p.alt = pt->alt;
		iter->p.speed = pt->speed;

		MemHandleUnlock(mrec);
		DmReleaseRecord(iter->db, iter->curr, false);
	}
	else
	{
		iter->p = iter->cache[iter->curr];
	}

	iter->curr++;
	if(iter->curr >= iter->max)
	{
		iter->curr = 0;
		iter->is_last = true;
	}
	else
		iter->is_last = false;

	return(true);
}

void TrackDrawReset(trackdraw_iter *iter)
{
	iter->curr = 0;
	iter->is_last = false;
}

void TrackDrawFinish(trackdraw_iter *iter)
{
	if(iter == NULL)
		return;

	if(iter->cache != NULL)
	{
		UtilFree(iter->cache);
		iter->cache = NULL;
	}
	else
		DmCloseDatabase(iter->db);
}

void TrackSendEvent()
{
	EventType evt;

	MemSet(&evt, sizeof(evt), 0);
	evt.eType = GpsPointSaved;
	EvtAddEventToQueue(&evt);
}

UInt16 getFra(Int32 *val, double mult)
{
	double v;
	Int32 *frac;
	UInt16 *dec;
	Int16 i;

	frac = val+1;
	dec  = (UInt16*)(frac+1);

	v = (double)(*val);
	if(*dec > 0)
	{
		for(i = 0; i < *dec; i++)
			v *= 10.;
		v += *frac;
		for(i = 0; i < (*dec-1); i++)
			v /= 10.;
	}
	if(mult)
		v *= mult;

	return((UInt16)v);
}

double toDD(double v)
{
	v = v * 180. / M_PI;
	v = ((Int32)(v * 1000000)) / 1000000.;
	return(v);
}

void TrackSetRenameDB(Char *fname)
{
	if(fname != NULL)
		StrNCopy(renamer, fname, dmDBNameLength);
	else
		renamer[0] = '\0';
}

Char *TrackGetRenameDB()
{
	return(renamer);
}

Err TrackRenameTo(Char *name)
{
	Err err;

	// same name, no rename!
	if(StrCompare(renamer, name) == 0)
		return(errNone);

	err = UtilDBRename(renamer, name);
	if(err == errNone)
	{
		// did we rename the current track?
		if(StrCompare(tsave.track_name, renamer) == 0)
			StrCopy(tsave.track_name, name);
	}
	else
		UtilBeep(BEEP_ERROR);

	return(err);
}

// TODO: let prefs choose the correct storage card!
UInt16 getFirstVolRef()
{
	UInt16 volRefNum;
	UInt32 volIterator = vfsIteratorStart;

	while(volIterator != vfsIteratorStop)
	{
		Err err = VFSVolumeEnumerate(&volRefNum, &volIterator);
		if(err == errNone)
		{
			VolumeInfoType volInfo;
			VFSVolumeInfo(volRefNum, &volInfo);
			return(volRefNum);
		}
	}
	return(-1);
}

// TODO IMPLEMENT!
Err TrackSaveAsPDB(Char *trackname)
{
	Err err = ftrErrNoSuchFeature;
	
	return(err);
}

Err TrackSaveAsKML(Char *trackname)
{
	Err err = ftrErrNoSuchFeature;
	UInt32 vfsMgrVersion;

	if(UtilIsOS(4, 0) == true)
	{
		err = FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &vfsMgrVersion);
		if(err == errNone)
		{
			UInt16 volRefNum;
			Char *pathName;
			UInt16 openMode;
			FileRef fileRef;

			volRefNum = getFirstVolRef();

			// create our directory, if not there
			err = VFSDirCreate(volRefNum, COTOGPS_DIR);
			// vfsErrFileGeneric works around a bug in the emulator
			if(err == errNone || err == vfsErrFileAlreadyExists || err == vfsErrFileGeneric)
			{
				// open KML file
				pathName  = UtilMalloc((StrLen(COTOGPS_DIR)+StrLen(trackname)+5)*sizeof(Char));
				StrPrintF(pathName, "%s%s.kml", COTOGPS_DIR, trackname);
				openMode  = vfsModeCreate | vfsModeWrite | vfsModeTruncate;
				err = VFSFileOpen(volRefNum, pathName, openMode, &fileRef);
				if(err == errNone)
				{
					DmOpenRef db;
					UInt32 i, cnt;
					
					cnt = TrackGetCountByName(trackname);
					db = UtilDBOpen(trackname, CGCreator, CGDBTrackType, dmModeReadOnly);
					if(db != NULL)
					{
						// write KML file
						WriteLine(fileRef, false, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
						WriteLine(fileRef, false, "<kml xmlns=\"http://www.opengis.net/kml/2.2\">");
						WriteLine(fileRef, false, " <Document>");
						WriteLine(fileRef, false, "  <name>");
						WriteLine(fileRef, true , COTOGPS_DOCNAME);
						WriteLine(fileRef, false, "  </name>");
						WriteLine(fileRef, false, "  <description><![CDATA[");
						WriteLine(fileRef, true , COTOGPS_COMMENT);
						WriteLine(fileRef, false, "  ]]></description>");
						WriteLine(fileRef, false, "  <Style id=\"cotoGPSStyle\">");
						WriteLine(fileRef, false, "   <LineStyle>");
						WriteLine(fileRef, false, "    <color>7dff0000</color>");
						WriteLine(fileRef, false, "    <width>4</width>");
						WriteLine(fileRef, false, "   </LineStyle>");
						WriteLine(fileRef, false, "   <PolyStyle>");
						WriteLine(fileRef, false, "    <color>7dff0000</color>");
						WriteLine(fileRef, false, "   </PolyStyle>");
						WriteLine(fileRef, false, "  </Style>");
						WriteLine(fileRef, false, "  <Placemark>");
						WriteLine(fileRef, false, "   <name>");
						WriteLine(fileRef, true , trackname);
						WriteLine(fileRef, false, "   </name>");
						WriteLine(fileRef, false, "   <description></description>");
						WriteLine(fileRef, false, "   <styleUrl>");
						WriteLine(fileRef, false, "#cotoGPSStyle");
						WriteLine(fileRef, false, "   </styleUrl>");
						WriteLine(fileRef, false, "   <LineString>");
						WriteLine(fileRef, false, "    <extrude>0</extrude>");
						WriteLine(fileRef, false, "    <tessellate>1</tessellate>");
						WriteLine(fileRef, false, "    <altitudeMode>clampToGround</altitudeMode>");
						WriteLine(fileRef, false, "    <coordinates>");
						// write coordinates and altitude
						for(i = 0; i < cnt; i++)
						{
							MemHandle mrec;
							trackpoint_type *pt;
							Char buf[80], buf1[20], buf2[20], buf3[20];

							mrec = DmGetRecord(db, i);
							if(mrec == NULL)
								break;
							pt = (trackpoint_type *)MemHandleLock(mrec);

							printDouble(toDD(-pt->lon), buf1);
							printDouble(toDD( pt->lat), buf2);
							printDouble(pt->alt*3.2808, buf3);
							StrPrintF(buf, "%s, %s, %s", buf1, buf2, buf3);
							WriteLine(fileRef, false, buf);

							MemHandleUnlock(mrec);
							DmReleaseRecord(db, i, false);
						}

						WriteLine(fileRef, false, "    </coordinates>");
						WriteLine(fileRef, false, "   </LineString>");
						WriteLine(fileRef, false, "  </Placemark>");
						WriteLine(fileRef, false, " </Document>");
						WriteLine(fileRef, false, "</kml>");
					}
					DmCloseDatabase(db);

					// close KML file
					VFSFileClose(fileRef);
				}
				UtilFree(pathName);
			}
		}
	}
	UtilBeep(BEEP_OK);
	return(err);
}

static Err WriteLine(FileRef fileRef, Boolean conv_utf8, Char *txt)
{
	Err err = errNone;

	// convert to UTF-8 before writing
	if(conv_utf8 == true)
	{
		UInt16 len, dstBytes;
		UInt32 srcEncoding;
		Char *buf;

		// get current encoding
		FtrGet(sysFtrCreator, sysFtrNumEncoding, &srcEncoding);

		// encode string to UTF-8
		// UTF-8-Character can be max. 4 bytes in size, so we are safe with 4*len!
		len = StrLen(txt);
		dstBytes = 4*len;
		buf = UtilMalloc(dstBytes);
		err = TxtConvertEncoding(true, NULL, txt, &len, srcEncoding, buf, &dstBytes, charEncodingUTF8, "?", 1);
		err = VFSFileWrite(fileRef, dstBytes, buf, NULL);
		UtilFree(buf);
	}
	else
	{
		// write directly to VFS (no conversion)
		err = VFSFileWrite(fileRef, StrLen(txt), txt, NULL);
	}
	VFSFileWrite(fileRef, 1, "\n", NULL);
	return(err);
}
