/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include "track.h"
#include "draw.h"
#include "main.h"
#include "util.h"
#include "nmea.h"
#include "forms.h"
#include "resource.h"
#include "def.h"

static void GPSHandler(FormPtr form, nmea_type *nmea) CODE3_SECTION;
static Boolean SizeForm(FormPtr form) CODE3_SECTION;
static void DrawTrack(FormPtr form) CODE3_SECTION;

static Boolean form_ready = false;
static UInt16 title_height;
static Boolean is_busy = false;
static Boolean is_mode_alt;
static trackdraw_iter *iter;
UInt8 extreme_mode = 0;
RectangleType frmBounds;

typedef struct
{
	double min_x, min_y, max_x, max_y;
	double scale;
	Coord off_x, off_y;
} trackdrawdata_type;

typedef struct
{
	Boolean is_inited;
	UInt32 is_extreme;
	UInt32 index_max_speed;
	UInt32 index_max_alt;
	UInt32 index_min_alt;
	trackdrawdata_type t;
	trackdrawdata_type a;
	trackdrawdata_type *c;
} trackdraw_type;
static trackdraw_type ddata;

#define EXTREME_NONE     0
#define EXTREME_MAXSPEED 1
#define EXTREME_MAXALT   2
#define EXTREME_MINALT   3

Boolean TrackDrawFormHandler(EventPtr event)
{
	Boolean handled = false;
	FormPtr	form;

	form = FrmGetActiveForm();
	switch(event->eType)
	{
		case frmOpenEvent:
			form_ready = false;
			is_mode_alt = false;
			is_busy = false;
			title_height = 16;
			if(UtilIsVGA())
			{
				title_height = VgaGetFrmTitleHeight()+1;
				VgaFormModify(form, vgaFormModify160To240);
			}
			ddata.is_inited = false;
			iter = TrackDrawInit();
			SizeForm(form);
			FrmDrawForm(form);
			DrawTrack(form);
			form_ready = true;
			handled = true;
			break;
		case frmCloseEvent:
			TrackDrawFinish(iter);
			form_ready = false;
			break;
		case frmUpdateEvent:
			FrmDrawForm(form);
			DrawTrack(form);
			handled = true;
			break;
		case menuEvent:
			switch(event->data.menu.itemID)
			{
				case TrackDrawNoSelectID:
					extreme_mode = EXTREME_NONE;
					DrawTrack(form);
					break;
				case TrackDrawMaxSpeedID:
					extreme_mode = EXTREME_MAXSPEED;
					DrawTrack(form);
					break;
				case TrackDrawMinAltID:
					extreme_mode = EXTREME_MINALT;
					DrawTrack(form);
					break;
				case TrackDrawMaxAltID:
					extreme_mode = EXTREME_MAXALT;
					DrawTrack(form);
					break;
			}
			break;
		case winDisplayChangedEvent:
		case displayExtentChangedEvent:
			if(SizeForm(form))
				DrawTrack(form);
			handled = true;
			break;
		case ctlSelectEvent:
			switch(event->data.ctlSelect.controlID)
			{
				case TrackDrawReturnID:
					UtilReturnToLastForm();
					handled = true;
					break;
				case TrackDrawModeID:
					is_mode_alt = !is_mode_alt;
					CtlSetLabel(UtilGetObjectPtr(form, TrackDrawModeID),
						is_mode_alt ? "A":"T");
					DrawTrack(form);
					handled = true;
					break;
			}
			break;
		case GpsNmeaEvent:
				GPSHandler(form, CASTNMEA(event));
			break;
		case keyDownEvent:
			if(UtilIsHardBtn(event) && UtilIsArrowBtn(event))
				handled = true;
			break;
		default:
			break;
	}
	return(handled);
}

static Boolean SizeForm(FormPtr form)
{
	RectangleType displayBounds;
	Coord diff, diff2;

	UtilGetFrmBounds(form, &frmBounds);
	UtilGetDpyBounds(&displayBounds);

	diff = displayBounds.extent.y - (frmBounds.topLeft.y + frmBounds.extent.y);
	diff2 = diff >> 1;

	if(diff)
	{
		ddata.is_inited = false;

		frmBounds.extent.y += diff;

		WinEraseRectangle(&displayBounds, 0);
		WinSetBounds(FrmGetWindowHandle(form), &frmBounds);
		UtilMoveObject(form, TrackDrawReturnID, diff, true);
		UtilMoveObject(form, TrackDrawModeID, diff, true);
		return(true);
	}
	return(false);
}

static void GPSHandler(FormPtr form, nmea_type *nmea)
{
	if(!is_busy)
		MainSetGPSTitle(form);
}

static void DrawTrack(FormPtr form)
{
	FrmEraseForm(form);
	FrmDrawForm(form);

	is_busy = true;
	SetTitleRes(form, TitleCALC);
	is_busy = false;

	if(iter != NULL)
	{
		Coord dw, dh, dh2, dmax;
		Coord fy, fw, fh, th, cb;
		RectangleType field_rect;

		FrmGetObjectBounds(form, FrmGetObjectIndex(form, TrackDrawModeID), &field_rect);

		cb = 4;
		dw = frmBounds.extent.x;
		dh = frmBounds.extent.y;
		fy = field_rect.topLeft.y;
		fw = field_rect.topLeft.x-cb;
		fh = field_rect.extent.y;
		th = title_height;

		if(UtilIsHires())
		{
			dw <<= 1;
			dh <<= 1;
			fy <<= 1;
			fw <<= 1;
			fh <<= 1;
			th <<= 1;
			cb <<= 1;
		}

		dh2   = dh-fh-th-cb;
		dmax  = MIN(dw, dh2);

		if(ddata.is_inited == false)
		{
			UInt32 max_speed, min_alt, max_alt;

			ddata.is_extreme = -1;
			ddata.index_max_speed  = -1;
			ddata.index_max_alt    = -1;
			ddata.index_min_alt    = -1;
			ddata.t.min_x = ddata.t.min_y = ddata.t.max_x = ddata.t.max_y = 0.;
			ddata.a.min_x = ddata.a.min_y = ddata.a.max_x = ddata.a.max_y = 0.;
			max_speed = min_alt = max_alt = 0;

			if((TrackDrawNext(iter)) && (iter->is_last == false))
			{
				max_speed = iter->p.speed;
				max_alt   = iter->p.alt;
				min_alt   = iter->p.alt;

				// for mode altitude
				ddata.a.min_x = 0;
				ddata.a.max_x = iter->max;
				ddata.a.min_y = ddata.a.max_y = -iter->p.alt;

				// for mode track
				ddata.t.min_x = ddata.t.max_x = iter->p.x;
				ddata.t.min_y = ddata.t.max_y = iter->p.y;

				do
				{
					if(TrackDrawNext(iter) == false)
						break;

					if(iter->p.speed > max_speed)
					{
						max_speed  = iter->p.speed;
						ddata.index_max_speed = iter->curr;
					}
					if(iter->p.alt > max_alt)
					{
						max_alt    = iter->p.alt;
						ddata.index_max_alt = iter->curr;
					}
					if(iter->p.alt < min_alt)
					{
						min_alt    = iter->p.alt;
						ddata.index_min_alt = iter->curr;
					}

					// for mode altitude
					ddata.a.min_y = MIN(-iter->p.alt, ddata.a.min_y);
					ddata.a.max_y = MAX(-iter->p.alt, ddata.a.max_y);

					// for mode track
					ddata.t.min_x = MIN(iter->p.x, ddata.t.min_x);
					ddata.t.max_x = MAX(iter->p.x, ddata.t.max_x);
					ddata.t.min_y = MIN(iter->p.y, ddata.t.min_y);
					ddata.t.max_y = MAX(iter->p.y, ddata.t.max_y);
				}
				while(iter->is_last == false);
			}

			ddata.a.max_y = MAX(0, ddata.a.max_y);
			ddata.a.max_x -= ddata.a.min_x;
			ddata.a.max_y -= ddata.a.min_y;
			ddata.t.max_x -= ddata.t.min_x;
			ddata.t.max_y -= ddata.t.min_y;
			ddata.a.scale = (double)MAX(ddata.a.max_x, ddata.a.max_y) / (double)dmax;
			ddata.t.scale = (double)MAX(ddata.t.max_x, ddata.t.max_y) / (double)dmax;
			ddata.a.off_x = (Coord)(ddata.a.max_x / ddata.a.scale);
			ddata.a.off_y = (Coord)(ddata.a.max_y / ddata.a.scale);
			ddata.t.off_x = (Coord)(ddata.t.max_x / ddata.t.scale);
			ddata.t.off_y = (Coord)(ddata.t.max_y / ddata.t.scale);

			ddata.a.off_x = (dw - ddata.a.off_x) >> 1;
			ddata.a.off_y = ((dh2 - ddata.a.off_y) >> 1) + th + (cb>>1);
			ddata.t.off_x = (dw - ddata.t.off_x) >> 1;
			ddata.t.off_y = ((dh2 - ddata.t.off_y) >> 1) + th + (cb>>1);

			ddata.is_inited = true;
			TrackDrawReset(iter);
		}

		SetTitleRes(form, TitleDRAWING);

		switch(extreme_mode)
		{
			case EXTREME_NONE:
				ddata.is_extreme = -1;
				break;
			case EXTREME_MAXSPEED:
				ddata.is_extreme = ddata.index_max_speed;
				break;
			case EXTREME_MAXALT:
				ddata.is_extreme = ddata.index_max_alt;
				break;
			case EXTREME_MINALT:
				ddata.is_extreme = ddata.index_min_alt;
				break;
		}

		if((TrackDrawNext(iter)) && (iter->is_last == false))
		{
			Coord x1, y1, x2, y2;
			Coord lx1, ly1, lx2, ly2;
			double vx, vy;
			CSYS_INIT

			CSYS_DOUBLE

			if(is_mode_alt)
			{
				ddata.c = &(ddata.a);

				vx = 0;
				vy = -iter->p.alt;
			}
			else
			{
				ddata.c = &(ddata.t);

				vx = iter->p.x;
				vy = iter->p.y;
			}

			lx1 = ly1 = lx2 = ly2 = -1;
			x1 = ddata.c->off_x + (Coord)((vx - ddata.c->min_x) / ddata.c->scale);
			y1 = ddata.c->off_y + (Coord)((vy - ddata.c->min_y) / ddata.c->scale);
			do
			{
				if(TrackDrawNext(iter) == false)
					break;

				if(is_mode_alt)
				{
					vx++;
					vy = -iter->p.alt;
				}
				else
				{
					vx = iter->p.x;
					vy = iter->p.y;
				}
				x2 = ddata.c->off_x + (Coord)((vx - ddata.c->min_x) / ddata.c->scale);
				y2 = ddata.c->off_y + (Coord)((vy - ddata.c->min_y) / ddata.c->scale);
				if(x1 != lx1 || x2 != lx2 || y1 != ly1 || y2 != ly2)
				{
					WinDrawLine(x1, y1, x2, y2);
					lx1 = x1;
					lx2 = x2;
					ly1 = y1;
					ly2 = y2;
					x1 = x2;
					y1 = y2;
				}
				if(iter->curr == ddata.is_extreme)
				{
					WinDrawLine(x2-5, y2-5, x2+5, y2+5);
					WinDrawLine(x2-5, y2+5, x2+5, y2-5);
				}
			}
			while(iter->is_last == false);

			if(is_mode_alt)
			{
				Coord scale_width = ddata.c->off_x-4;

				x1 = ddata.c->off_x;
				x2 = dw - ddata.c->off_x;
				y1 = ddata.c->off_y + (Coord)(-ddata.c->min_y / ddata.c->scale);
				WinDrawLine(x1, y1, x2, y1);

				if(scale_width < 8)
					scale_width = 8;
				RctSetRectangle(&field_rect, 0, ddata.c->off_y, scale_width, y1-ddata.c->off_y);
			}
			else
			{
				RctSetRectangle(&field_rect, 0, fy, fw, fh);
			}
			DrawScale(&field_rect, ddata.c->scale, is_mode_alt);

			CSYS_RESET
		}
		// if cached, draw a dot at the right bottom as a "secret" hint
		if(iter->cache != NULL)
		{
			WinDrawPixel(	frmBounds.topLeft.x+frmBounds.extent.x-1,
							frmBounds.topLeft.y+frmBounds.extent.y-1);
		}
	}

	SetTitleRes(form, is_mode_alt ? TitleALTDRAW : TitleTRACKDRAW);
}

