/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include "track.h"
#include "main.h"
#include "util.h"
#include "nmea.h"
#include "prefs.h"
#include "forms.h"
#include "beam.h"
#include "resource.h"
#include "def.h"

static void GPSHandler(FormPtr form, nmea_type *nmea) CODE2_SECTION;
static void SizeForm(FormPtr form) CODE2_SECTION;

static void SaveOffset(FormPtr form) CODE2_SECTION;
static void EnableItems(FormPtr form, Boolean b) CODE2_SECTION;
static Boolean UpdateTrackPoints(FormPtr form) CODE2_SECTION;
static void CalcData(FormPtr form, Boolean is_area) CODE2_SECTION;
static void UpdateData(FormPtr form) CODE2_SECTION;

static Boolean form_ready = false;
static UInt32 last_check;
static Boolean was_tracking;
static Char *string_pts;

Boolean TrackFormHandler(EventPtr event)
{
	CAppPrefs *prefs;
	Boolean handled = false;
	char buf[8];
	char fbuf[8];
	FormPtr	form;

	form = FrmGetActiveForm();
	switch(event->eType)
	{
		case frmOpenEvent:
			form_ready = false;
			last_check = 0L;
			string_pts = UtilGetResourceString(StringPOINTS);
			if(UtilIsVGA())
				VgaFormModify(form, vgaFormModify160To240);
			if(UtilIsResizeableNoDia())
				SizeForm(form);
			FrmDrawForm(form);
			was_tracking = TrackIsCapturing();
			EnableItems(form, was_tracking);
			prefs = PrefsGetPrefs();
			if(prefs->track_side_inside == true)
				UtilSetControlValue(form, TrackPushIID, 1);
			else
				UtilSetControlValue(form, TrackPushOID, 1);
			StrPrintF(buf, "%d", prefs->track_offset_m);
			UtilSetFieldText(form, TrackOffsetMID, buf);
			StrPrintF(fbuf, "%%0%ud", prefs->track_offset_c_l);
			StrPrintF(buf, fbuf, prefs->track_offset_c);
			UtilSetFieldText(form, TrackOffsetCID, buf);
			PrefsUpdateLabelOffset(form, TrackOffsetUnitID);
			UtilSetListSelection(form, DistListID, DistTriggerID,
									prefs->track_format_dist);
			UtilSetListSelection(form, AreaListID, AreaTriggerID,
									prefs->track_format_area);
			UtilSetListSelection(form, TrackIntervalListID, TrackIntervalTID,
									prefs->track_interval_index);
			UtilSetControlValue(form, TrackDoBeepID, prefs->track_beep);
			UtilUpdateLabel(form, TrackNameID, TrackGetName());
			UpdateTrackPoints(form);
			UpdateData(form);
			form_ready = true;
			handled = true;
			break;
		case frmCloseEvent:
			form_ready = false;
			SaveOffset(form);
			UtilFree(string_pts);
			break;
		case frmUpdateEvent:
			UtilUpdateLabel(form, TrackNameID, TrackGetName());
			FrmDrawForm(form);
			break;
		case winDisplayChangedEvent:
		case displayExtentChangedEvent:
			if(form_ready == true)
			{
				SizeForm(form);
				FrmDrawForm(form);
			}
			handled = true;
			break;
		case ctlSelectEvent:
			switch(event->data.ctlSelect.controlID)
			{
				case TrackDrawID:
						TrackDraw();
					break;
				case TrackStartID:
						TrackSetCapturing(true);
					break;
				case TrackStopID:
						TrackSetCapturing(false);
					break;
				case TrackDoBeepID:
						prefs = PrefsGetPrefs();
						prefs->track_beep =
							UtilGetControlValue(form, TrackDoBeepID);
					break;
				case TrackCalcDistID:
						CalcData(form, false);
					break;
				case TrackCalcAreaID:
						CalcData(form, true);
					break;
				case TrackCalcNullID:
						UtilSetFieldText(form, TrackOffsetMID, "0");
						UtilSetFieldText(form, TrackOffsetCID, "0");
					break;
				case TrackManagerID:
						UtilPopupForm(TrackManagerForm);
					break;
			}
			break;
		case popSelectEvent:
				handled = false;
				if(event->data.popSelect.listID == TrackIntervalListID)
				{
					PrefsGetPrefs()->track_interval_index =
							event->data.popSelect.selection;
				}
				if(event->data.popSelect.listID == DistListID)
				{
					PrefsGetPrefs()->track_format_dist =
							event->data.popSelect.selection;
					UpdateData(form);
				}
				if(event->data.popSelect.listID == AreaListID)
				{
					PrefsGetPrefs()->track_format_area =
							event->data.popSelect.selection;
					UpdateData(form);
				}
			break;
		case GpsNmeaEvent:
				GPSHandler(form, CASTNMEA(event));
			break;
		case GpsPointSaved:
				handled = UpdateTrackPoints(form);
			break;
		default:
			break;
	}
	return(handled);
}

static void SizeForm(FormPtr form)
{
	RectangleType curBounds, displayBounds;
	Coord diff, diff2;

	UtilGetFrmBounds(form, &curBounds);
	UtilGetDpyBounds(&displayBounds);

	diff = displayBounds.extent.y - (curBounds.topLeft.y + curBounds.extent.y);

	if(diff)
	{
		curBounds.extent.y += diff;
		WinSetBounds(FrmGetWindowHandle(form), &curBounds);

		diff2 = diff >> 1;

		UtilMoveObject(form, TrackPushLabelID, diff, true);
		UtilMoveObject(form, TrackPushIID, diff, true);
		UtilMoveObject(form, TrackPushOID, diff, true);
		UtilMoveObject(form, TrackOffsetMID, diff, true);
		UtilMoveObject(form, TrackOffsetPointID, diff, true);
		UtilMoveObject(form, TrackOffsetCID, diff, true);
		UtilMoveObject(form, TrackOffsetUnitID, diff, true);
		UtilMoveObject(form, TrackCalcNullID, diff, true);
		UtilMoveObject(form, TrackSaveLabelID, diff, true);
		UtilMoveObject(form, TrackIntervalTID, diff, true);
		UtilMoveObject(form, TrackIntervalListID, diff, false);
		UtilMoveObject(form, TrackDoBeepID, diff, true);
		UtilMoveObject(form, TrackStartID, diff, !TrackIsCapturing());
		UtilMoveObject(form, TrackStopID, diff, TrackIsCapturing());
		UtilMoveObject(form, TrackManagerID, diff, true);
		UtilMoveObject(form, TrackDrawID, diff, true);

		UtilMoveObject(form, TrackDistLabelID, diff2, true);
		UtilMoveObject(form, DistanceID, diff2, true);
		UtilMoveObject(form, DistTriggerID, diff2, true);
		UtilMoveObject(form, DistListID, diff2, false);
		UtilMoveObject(form, TrackCalcDistID, diff2, true);
		UtilMoveObject(form, TrackAreaLabelID, diff2, true);
		UtilMoveObject(form, AreaID, diff2, true);
		UtilMoveObject(form, AreaTriggerID, diff2, true);
		UtilMoveObject(form, AreaListID, diff2, false);
		UtilMoveObject(form, TrackCalcAreaID, diff2, true);
	}
}

static void GPSHandler(FormPtr form, nmea_type *nmea)
{
	MainSetGPSTitle(form);
}

static void SaveOffset(FormPtr form)
{
	Int32 val;
	Int16 val16;
	CAppPrefs *prefs;

	prefs = PrefsGetPrefs();

	if(UtilGetControlValue(form, TrackPushIID) != 0)
		prefs->track_side_inside = true;
	else
		prefs->track_side_inside = false;

	if(UtilGetInt(form, TrackOffsetMID, &val, NULL) == true)
		prefs->track_offset_m = (Int16)val;
	if(UtilGetInt(form, TrackOffsetCID, &val, &val16) == true)
	{
		prefs->track_offset_c = (Int16)val;
		prefs->track_offset_c_l = val16;
	}
}

static void EnableItems(FormPtr form, Boolean b)
{
	UtilShowObject(form, TrackStartID, !b);
	UtilShowObject(form, TrackStopID, b);
	UtilShowObject(form, TrackManagerID, !b);
	UtilShowObject(form, TrackDrawID, !b);
}

static Boolean UpdateTrackPoints(FormPtr form)
{
	UInt32 cnt, mem;
	Boolean is_tracking;
	double bf;
	UInt8 dec;
	static Char s[4];

	dec = 2;
	cnt = TrackGetCount();
	MemCardInfo(0, 0, 0, 0, 0, 0, 0, &mem);
	bf = (double)mem;
    if(mem > 1152921504606846976L)
    {
        bf /= 1152921504606846976L;
        StrCopy(s, "EiB");
    }
    else if(mem > 1125899906842624L)
    {
        bf /= 1125899906842624L;
        StrCopy(s, "PiB");
    }
    else if(mem > 1099511627776L)
    {
        bf /= 1099511627776L;
        StrCopy(s, "TiB");
    }
    else if(mem > 1073741824L)
    {
        bf /= 1073741824L;
        StrCopy(s, "GiB");
    }
    else if(mem > 1048576L)
    {
        bf /= 1048576L;
        StrCopy(s, "MiB");
    }
    else if(mem > 1024L)
    {
        bf /= 1024L;
        StrCopy(s, "KiB");
    }
    else
	{
        StrCopy(s, "B");
		dec = 0;
	}

	if(dec == 0)
		UtilUpdateSingle(form, MemoryLeftID, (Int32)bf, s);
	else
		UtilUpdateDoubleUnit(form, MemoryLeftID, bf, dec, s);

	UtilUpdateSingle(form, TrackPointsID, cnt, string_pts);

	is_tracking = TrackIsCapturing();
	if(is_tracking != was_tracking)
	{
		EnableItems(form, is_tracking);
		was_tracking = is_tracking;
	}
	return(true);
}

static void CalcData(FormPtr form, Boolean is_area)
{
	double dist, area;
	CAppPrefs *prefs;

	SaveOffset(form);

	dist = 0.;
	area = 0.;
	prefs = PrefsGetPrefs();

	if(is_area)
	{
		TrackCalc(&dist, &area);
		prefs->track_has_last_area = true;
		prefs->track_last_area = area;
	}
	else
	{
		TrackCalc(&dist, NULL);
		prefs->track_has_last_area = false;
		prefs->track_last_area = 0.;
	}

	prefs->track_last_dist = dist;
	prefs->track_has_last_dist = true;

	UpdateData(form);
}

static void UpdateData(FormPtr form)
{
	CAppPrefs *prefs;

	prefs = PrefsGetPrefs();

	if(prefs->track_has_last_dist)
	{
		UtilUpdateDouble(form, DistanceID,
							PrefsTrackDistToPref(prefs->track_last_dist), 2);
	}
	else
		UtilUpdateString(form, DistanceID, "---");

	if(prefs->track_has_last_area)
	{
		UtilUpdateDouble(form, AreaID,
							PrefsTrackAreaToPref(prefs->track_last_area), 2);
	}
	else
		UtilUpdateString(form, AreaID, "---");
}

