/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <handera/Vga.h>
#include <VFSMgr.h>
#include "main.h"
#include "forms.h"
#include "util.h"
#include "serial.h"
#include "prefs.h"
#include "dist.h"
#include "marker.h"
#include "track.h"
#include "beam.h"
#include "resource.h"

#define MAXVOLLABELLEN 255

static void UpdateTrackList(FormPtr form) CODE3_SECTION;
static void FreeTracks() CODE3_SECTION;
static void NewTrack() CODE3_SECTION;
static void RemoveDB(FormPtr form) CODE3_SECTION;
static void SaveAsPDB(FormPtr form) CODE3_SECTION;
static void SaveAsKML(FormPtr form) CODE3_SECTION;
static void ButtonHandler(FormPtr form, WChar direction) CODE3_SECTION;
static void SizeForm(FormPtr form) CODE3_SECTION;
static void ListDrawFunc(Int16 item, RectangleType *bounds, Char **data) CODE3_SECTION;
static Char *GetSelectedName(FormPtr form) CODE3_SECTION;
static Char *GetName(LocalID id) CODE3_SECTION;
static void InitCategories(FormPtr form) CODE3_SECTION;
static void FreeCategories() CODE3_SECTION;

static Boolean form_ready = false;
static UInt16 trackcnt;
static LocalID *tracks;
static UInt32 *tracksize;

static Int16 fs_cnt = 0, fs_selected = 0;
//static UInt16 *fs_volrefnum;
static Char **fs_items = NULL;

Boolean TrackManagerFormHandler(EventPtr event)
{
	FormPtr	form;
	Boolean handled = false;

	form = FrmGetActiveForm();

	switch(event->eType)
	{
		case frmOpenEvent:
			form_ready = false;
			tracks = NULL;
			tracksize = NULL;
			trackcnt = 0;
			if(UtilIsVGA())
				VgaFormModify(form, vgaFormModify160To240);
			SizeForm(form);
			FrmDrawForm(form);
			SetTitleRes(form, TitleTRACKMANAGER);
			InitCategories(form);
			LstSetDrawFunction(UtilGetObjectPtr(form, TrackManagerListID), ListDrawFunc);
			UpdateTrackList(form);
			form_ready = true;
			handled = true;
			break;
		case frmCloseEvent:
			FreeTracks();
			FreeCategories();
			form_ready = false;
			break;
		case frmUpdateEvent:
			switch(event->data.frmUpdate.updateCode)
			{
				case frmRedrawUpdateCode:
					UpdateTrackList(form);
					FrmDrawForm(form);
					handled = true;
					break;
			}
			break;
		case menuEvent:
			switch(event->data.menu.itemID)
			{
				case TrackManagerNewID:
					NewTrack();
					UpdateTrackList(form);
					break;
				case TrackManagerRenameID:
					TrackSetRenameDB(GetSelectedName(form));
					FrmPopupForm(RenameForm);
					break;
				case TrackManagerSendID:
					BeamDB(GetSelectedName(form), BEAM_TRACK);
					break;
				case TrackManagerDeleteID:
					// TODO ask again!
					RemoveDB(form);
					break;
				case TrackSaveAsPDBID:
					SaveAsPDB(form);
					break;
				case TrackSaveAsKMLID:
					SaveAsKML(form);
					break;
			}
			break;
		case ctlSelectEvent:
			switch(event->data.ctlSelect.controlID)
			{
				case TrackManagerUseID:
					TrackSaveInit(GetSelectedName(form));
					LstDrawList(UtilGetObjectPtr(form, TrackManagerListID));
					UtilReturnToLastForm();
					handled = true;
					break;
				case TrackManagerCloseID:
					UtilReturnToLastForm();
					handled = true;
					break;
			}
			break;
		case popSelectEvent:
			if(event->data.popSelect.listID == TrackManagerCatListID)
			{
				fs_selected = event->data.popSelect.selection;
				UpdateTrackList(form);
			}
			break;
		case winDisplayChangedEvent:
		case displayExtentChangedEvent:
			if(form_ready == true)
			{
				SizeForm(form);
				FrmDrawForm(form);
			}
			handled = true;
			break;
		case keyDownEvent:
			if(UtilIsHardBtn(event) && UtilIsArrowBtn(event))
			{
				ButtonHandler(form, UtilBtnDirection(event));
				handled = true;
			}
			break;
		default:
			break;
	}
	return(handled);
}

static void UpdateTrackList(FormPtr form)
{
	Err err;

	if(trackcnt > 0)
		FreeTracks();

	//TODO add VFS support!
	err = TrackGetTracks(&tracks, &trackcnt);
	if(err == errNone)
	{
		ListType *list;
		Int16 i;
		Char *cur;

		/* tracksize is unsigned! hack alert! */
		tracksize = (UInt32*)UtilMalloc((trackcnt)*sizeof(UInt32));
		for(i = 0; i < trackcnt; i++)
			tracksize[i] = -1;

		list = UtilGetObjectPtr(form, TrackManagerListID);
		LstSetListChoices(list, NULL, (Int16)trackcnt);

		cur = TrackGetName();
		for(i = 0; i < trackcnt; i++)
		{
			if(StrCompare(cur, GetName(tracks[i])) == 0)
			{
				LstSetSelection(list, i);
				break;
			}
		}
		LstDrawList(list);
	}
}

static void FreeTracks()
{
	if(tracks != NULL)
	{
		UtilFree(tracks);
		tracks = NULL;
	}
	if(tracksize != NULL)
	{
		UtilFree(tracksize);
		tracksize = NULL;
	}
}

static void NewTrack()
{
	Char fname[dmDBNameLength];
	UInt32 i = 0;

	while(true)
	{
		StrPrintF(fname, "%s%ld", CGDBTrack, i);
		if(UtilDBExists(fname) == false)
			break;
		i++;
	}

	TrackNewDB(fname);
}

static void RemoveDB(FormPtr form)
{
	if(TrackRemoveDB(GetSelectedName(form)) == errNone)
		UpdateTrackList(form);
	else
		UtilBeep(BEEP_ERROR);
}

static void SaveAsPDB(FormPtr form)
{
	if(TrackSaveAsPDB(GetSelectedName(form)) != errNone)
		UtilBeep(BEEP_ERROR);	
}

static void SaveAsKML(FormPtr form)
{
	if(TrackSaveAsKML(GetSelectedName(form)) != errNone)
		UtilBeep(BEEP_ERROR);	
}

static void ButtonHandler(FormPtr form, WChar direction)
{
	ListType *list;
	Int16 vis;

	list = UtilGetObjectPtr(form, TrackManagerListID);
	vis = LstGetVisibleItems(list);
	LstScrollList(list, (direction == vchrPageUp ? winUp : winDown), vis-1);
}

static void SizeForm(FormPtr form)
{
	RectangleType curBounds, displayBounds;
	Coord diff;

	UtilGetFrmBounds(form, &curBounds);
	UtilGetDpyBounds(&displayBounds);

	diff = displayBounds.extent.y - (curBounds.topLeft.y + curBounds.extent.y);

	if(diff)
	{
		curBounds.extent.y += diff;
		WinSetBounds(FrmGetWindowHandle(form), &curBounds);

		UtilResizeObject(form, TrackManagerListID,  diff, true);
		UtilMoveObject(form, TrackManagerPopupTID,  diff, true);
		UtilMoveObject(form, TrackManagerCatListID, diff, true);
		UtilMoveObject(form, TrackManagerUseID,     diff, true);
		UtilMoveObject(form, TrackManagerCloseID,   diff, true);
	}
}

static void ListDrawFunc(Int16 index, RectangleType *bounds, Char **data)
{
	Char *name;
	Coord w, x, y;
	Int16 slen, cw, refw = 0;
	Boolean fits;
	Char buf[32];
	UInt32 cnt=0;

	name = GetName(tracks[index]);

	x    = bounds->topLeft.x;
	y    = bounds->topLeft.y;
	w    = bounds->extent.x;
	slen = StrLen(name);

	FntCharsInWidth(name, &w, &slen, &fits);
	WinDrawChars(name, slen, x, y);

	if(StrCompare(name, TrackGetName()) == 0)
	{
		FontID old_font;
		Char isref = 0x16;

		if(UtilIsVGA())
			old_font = FntSetFont(VgaBaseToVgaFont(symbolFont));
		else
			old_font = FntSetFont(symbolFont);
		refw = FntCharsWidth(&isref, 1);
		WinDrawChars(&isref, 1, bounds->extent.x-refw, y);
		FntSetFont(old_font);
	}

	/* tracksize is unsigned! hack alert! */
	if(tracksize != NULL && tracksize[index] != -1)
		cnt = tracksize[index];
	else
	{
		cnt = TrackGetCountById(tracks[index]);
		if(tracksize != NULL)
			tracksize[index] = cnt;
	}
	StrPrintF(buf, "%ld", cnt);
	slen = StrLen(buf);
	cw = FntCharsWidth(buf, slen);
	WinDrawChars(buf, slen, bounds->extent.x-cw-refw, y);
}

static Char *GetSelectedName(FormPtr form)
{
	ListType *list;
	Int16 index;

	list = UtilGetObjectPtr(form, TrackManagerListID);
	index = LstGetSelection(list);
	if(index != noListSelection)
		return(GetName(tracks[index]));

	return(NULL);
}

static Char *GetName(LocalID id)
{
	static Char name[dmDBNameLength];

	UtilDBGetName(id, name);
	return(name);
}

static void InitCategories(FormPtr form)
{
/*
	UInt32 vfsMgrVersion;
	Err err;

	FreeCategories();

	err = FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &vfsMgrVersion);
	if(!err)
	{
		UInt16 volRefNum;
		UInt32 volIterator = vfsIteratorStart;

		fs_cnt = 0;
		while(volIterator != vfsIteratorStop)
		{
			err = VFSVolumeEnumerate(&volRefNum, &volIterator);
			if(err == errNone)
			{
				VolumeInfoType volInfo;
				err = VFSVolumeInfo(volRefNum, &volInfo);
				if(err == errNone && !(volInfo.attributes & vfsVolumeAttrHidden))
				{
					fs_cnt++;
				}
			}
		}

		if(fs_cnt > 0)
		{
			RectangleType rect;
			UInt16 n = 0;

			fs_cnt++; // add one for "internal memory"
			fs_volrefnum = (UInt16*) UtilMalloc(fs_cnt * sizeof(UInt16));
			fs_items = (Char**) UtilMalloc(fs_cnt * sizeof(Char*));

			fs_volrefnum[n] = 0;
			fs_items[n] = UtilGetResourceString(StringInternalMemory);
			n++;

			volIterator = vfsIteratorStart;
			while(volIterator != vfsIteratorStop)
			{
				err = VFSVolumeEnumerate(&volRefNum, &volIterator);
				if(err == errNone)
				{
					VolumeInfoType volInfo;
					err = VFSVolumeInfo(volRefNum, &volInfo);
					if(err == errNone && !(volInfo.attributes & vfsVolumeAttrHidden))
					{
						Char label[MAXVOLLABELLEN+1];
						err = VFSVolumeGetLabel(volRefNum, label, MAXVOLLABELLEN);
						if(err == errNone)
						{
							fs_volrefnum[n] = volRefNum;
							fs_items[n] = UtilMalloc(StrLen(label)+1);
							StrCopy(fs_items[n], label);
							n++;
						}
					}
				}
			}

			UtilSetList(form, TrackManagerCatListID, TrackManagerPopupTID, fs_items, fs_cnt, fs_selected);
			FrmGetObjectBounds(form, FrmGetObjectIndex(form, TrackManagerPopupTID), &rect);
			n = rect.extent.y + 4;
			UtilResizeObject(form, TrackManagerListID,  -n, false);
			UtilMoveObject(form, TrackManagerListID,  n, true);
			UtilShowObject(form, TrackManagerPopupTID, true);
			return;
		}
	}

	UtilShowObject(form, TrackManagerPopupTID, false);
*/
}

void FreeCategories()
{
	if(fs_cnt > 0)
	{
		UInt16 i;

		for(i = 0; i < fs_cnt; i++)
			UtilFree(fs_items[i]);
		UtilFree(fs_items);
	}
	fs_items = NULL;
	fs_cnt = 0;
}

