/*
 * Copyright (c) 2010, Thomas Runge (trunge@truwo.de)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <PalmOS.h>
#include <DLServer.h>
#include <handera/Vga.h>
#include "main.h"
#include "util.h"
#include "nmea.h"
#include "time.h"
#include "draw.h"
#include "resource.h"
#include "prefs.h"
#include "forms.h"
#include "def.h"

static void GPSHandler(FormPtr form, nmea_type *nmea) CODE2_SECTION;
static void SizeForm(FormPtr form) CODE2_SECTION;
static void HandleTap(FormPtr form, Boolean is_down, Coord x, Coord y) CODE2_SECTION;

static Boolean form_ready = false;
static Boolean was_down = false;
static Int16 x, y, w, h;

Boolean WorldFormHandler(EventPtr event)
{
	char syncname[dlkUserNameBufSize+1];
	RectangleType rect;
	Boolean handled = false;
	FormPtr	form = FrmGetActiveForm();

	switch(event->eType)
	{
		case frmOpenEvent:
			form_ready = false;
			if(UtilIsVGA())
				VgaFormModify(form, vgaFormModify160To240);
			if(UtilIsResizeableNoDia())
				SizeForm(form);
			FrmDrawForm(form);
			PrefsUpdateLabelLatLon(form, LatLabelID, LonLabelID, false);
			DlkGetSyncInfo(0, 0, 0, syncname, 0, 0);
			FrmCopyTitle(form, syncname);
			FrmGetObjectBounds(form, FrmGetObjectIndex(form, WorldBitmapID), &rect);
			x = rect.topLeft.x;
			y = rect.topLeft.y;
			w = rect.extent.x;
			h = rect.extent.y;
			form_ready = true;
			handled = true;
			break;
		case frmCloseEvent:
			form_ready = false;
			break;
		case penDownEvent:
			HandleTap(form, true, event->screenX, event->screenY);
			break;
		case penUpEvent:
			HandleTap(form, false, event->screenX, event->screenY);
			break;
		case winDisplayChangedEvent:
		case displayExtentChangedEvent:
			if(form_ready == true)
			{
				SizeForm(form);
				FrmDrawForm(form);
			}
			handled = true;
			break;
		case GpsNmeaEvent:
				GPSHandler(form, CASTNMEA(event));
			break;
		default:
			break;
	}
	return(handled);
}

static void SizeForm(FormPtr form)
{
	RectangleType curBounds, displayBounds;
	Coord diff;

	UtilGetFrmBounds(form, &curBounds);
	UtilGetDpyBounds(&displayBounds);

	diff  = displayBounds.extent.y - (curBounds.topLeft.y + curBounds.extent.y);

	if(diff || UtilIsVGA())
	{
		UInt16 objindex;
		RectangleType frect;

		curBounds.extent.y += diff;
		WinSetBounds(FrmGetWindowHandle(form), &curBounds);
		WinEraseRectangle(&curBounds, 0);

		objindex = FrmGetObjectIndex(form, WorldBitmapID);
		FrmGetObjectBounds(form, objindex, &frect);
		frect.topLeft.x = (curBounds.extent.x - frect.extent.x) >> 1;
		frect.topLeft.y = curBounds.extent.y - frect.extent.y;
		if(diff > 0)
			frect.topLeft.y -= diff/2;
		FrmSetObjectBounds(form, objindex, &frect);
		x = frect.topLeft.x;
		y = frect.topLeft.y;
		w = frect.extent.x;
		h = frect.extent.y;
	}
}

static void HandleTap(FormPtr form, Boolean is_down, Coord x, Coord y)
{
	UInt16 oindex;
	RectangleType frect;

	oindex = FrmGetObjectIndex(form, WorldBitmapID);
	FrmGetObjectBounds(form, oindex, &frect);

	if(RctPtInRectangle(x, y, &frect))
	{
		if(is_down)
			was_down = true;
		else
			if(was_down)
				UtilReturnToLastForm();
	}
	else
		was_down = false;
}

static void GPSHandler(FormPtr form, nmea_type *nmea)
{
	if(nmea->is_new_gga || nmea->is_new_nexrmc)
	{
		Int16 x1, y1, x2, y2;
		Int32 deg;

		FrmDrawForm(form);

		deg = (Int32) (nmea->lat * 180. / M_PI);
		x1 = x;
		y1 = y + (h>>1) - deg * h/180;
		x2 = x1+w;
		y2 = y1;
		WinDrawGrayLine(x1, y1, x2, y2);

		deg = (Int32) (nmea->lon * 180. / M_PI);
		x1 = x + (w>>1) - deg * w/360;
		y1 = y;
		x2 = x1;
		y2 = y1+h;
		WinDrawGrayLine(x1, y1, x2, y2);

		PrefsUpdateLabelAlt(form, AltID, nmea->alt, nmea->alt_unit,
								nmea->alt_WGS84, nmea->alt_WGS84_unit,
								nmea->lat, nmea->lon);
		PrefsUpdateLabelCoord(form,
								LatitudeID, LongitudeID,
								nmea->lat, nmea->lon);
	}

	if(nmea->is_new_rmc || nmea->is_new_gga || nmea->is_new_nexrmc)
	{
		UtilUpdateString(form, DateID, TimeGetDateString());
		UtilUpdateString(form, TimeID, TimeGetTimeString());

		nmea->is_new_rmc = false;
		nmea->is_new_gga = false;
		nmea->is_new_nexrmc = false;
	}
}

